implement Testxts;

include "sys.m";
	sys: Sys;
	sprint: import sys;
include "draw.m";
include "arg.m";
include "keyring.m";
	kr: Keyring;
include "../lib/aesxts.m";
	aesxts: Aesxts;

Xts: adt {
	seq:	big;
	key:	array of byte;
	plain, cipher:	array of byte;
};

Testxts: module {
	init:	fn(nil: ref Draw->Context, args: list of string);
};

init(nil: ref Draw->Context, args: list of string)
{
	sys = load Sys Sys->PATH;
	kr = load Keyring Keyring->PATH;
	aesxts = load Aesxts Aesxts->PATH;
	aesxts->init();

	arg := load Arg Arg->PATH;
	arg->init(args);
	arg->setusage("testxts");
	while((c := arg->opt()) != 0)
		case c {
		* =>	arg->usage();
		}

	if(arg->argv() != nil)
		arg->usage();

	for(i := 0; i < len tests; i++)
		test(tests[i]);
}

test(xts: Xts)
{
	k1 := xts.key[:len xts.key/2];
	k2 := xts.key[len xts.key/2:];
	warn(sprint("test, xts, seq %bx, keys (%d, %d, %d) %s %s, len data %d %d",
		xts.seq, len xts.key, len k1, len k2, hex(k1), hex(k2), len xts.plain, len xts.cipher));

	st1 := kr->aessetup(k1, nil);
	st2 := kr->aessetup(k2, nil);
	cipher := array[len xts.plain] of byte;
	cipher[:] = xts.plain;
	aesxts->crypt(st1, st2, xts.seq, cipher, len cipher, kr->Encrypt);
	if(!eq(xts.cipher, cipher)) {
		warn("encryption failed");
		say(sprint("expect (%d):", len xts.cipher));
		say(hex(xts.cipher));
		say(sprint("saw: (%d)", len cipher));
		say(hex(cipher));
	} else
		warn("encryption okay");

	plain := array[len xts.cipher] of byte;
	plain[:] = xts.cipher;
	aesxts->crypt(st1, st2, xts.seq, plain, len plain, kr->Decrypt);
	if(!eq(xts.plain, plain))
		warn("decryption failed");
	else
		warn("decryption okay");
}

eq(a, b: array of byte): int
{
	if(len a != len b)
		return 0;
	for(i := 0; i < len a; i++)
		if(a[i] != b[i])
			return 0;
	return 1;
}

hex(d: array of byte): string
{
	s := "";
	for(i := 0; i < len d; i++)
		s += sprint("%02x", int d[i]);
	return s;
}

warn(s: string)
{
	say(s);
}

say(s: string)
{
	sys->fprint(sys->fildes(2), "%s\n", s);
}

fail(s: string)
{
	sys->fprint(sys->fildes(2), "%s\n", s);
	raise "fail:"+s;
}


tests := array[] of {
Xts (
	big 16r00000000,
	array[] of {
		byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, 
		byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, 
		byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, 
		byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, 
	},
	array[] of {
		byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, 
		byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, 
		byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, 
		byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, byte 16r00, 
	},
	array[] of {
		byte 16r91, byte 16r7c, byte 16rf6, byte 16r9e, byte 16rbd, byte 16r68, byte 16rb2, byte 16rec, 
		byte 16r9b, byte 16r9f, byte 16re9, byte 16ra3, byte 16rea, byte 16rdd, byte 16ra6, byte 16r92, 
		byte 16rcd, byte 16r43, byte 16rd2, byte 16rf5, byte 16r95, byte 16r98, byte 16red, byte 16r85, 
		byte 16r8c, byte 16r02, byte 16rc2, byte 16r65, byte 16r2f, byte 16rbf, byte 16r92, byte 16r2e, 
	}
),
Xts (
	big 16r3333333333,
	array[] of {
		byte 16r11, byte 16r11, byte 16r11, byte 16r11, byte 16r11, byte 16r11, byte 16r11, byte 16r11, 
		byte 16r11, byte 16r11, byte 16r11, byte 16r11, byte 16r11, byte 16r11, byte 16r11, byte 16r11, 
		byte 16r22, byte 16r22, byte 16r22, byte 16r22, byte 16r22, byte 16r22, byte 16r22, byte 16r22, 
		byte 16r22, byte 16r22, byte 16r22, byte 16r22, byte 16r22, byte 16r22, byte 16r22, byte 16r22, 
	},
	array[] of {
		byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, 
		byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, 
		byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, 
		byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, 
	},
	array[] of {
		byte 16rc4, byte 16r54, byte 16r18, byte 16r5e, byte 16r6a, byte 16r16, byte 16r93, byte 16r6e, 
		byte 16r39, byte 16r33, byte 16r40, byte 16r38, byte 16rac, byte 16ref, byte 16r83, byte 16r8b, 
		byte 16rfb, byte 16r18, byte 16r6f, byte 16rff, byte 16r74, byte 16r80, byte 16rad, byte 16rc4, 
		byte 16r28, byte 16r93, byte 16r82, byte 16rec, byte 16rd6, byte 16rd3, byte 16r94, byte 16rf0, 
	}
),
Xts (
	big 16r3333333333,
	array[] of {
		byte 16rff, byte 16rfe, byte 16rfd, byte 16rfc, byte 16rfb, byte 16rfa, byte 16rf9, byte 16rf8, 
		byte 16rf7, byte 16rf6, byte 16rf5, byte 16rf4, byte 16rf3, byte 16rf2, byte 16rf1, byte 16rf0, 
		byte 16r22, byte 16r22, byte 16r22, byte 16r22, byte 16r22, byte 16r22, byte 16r22, byte 16r22, 
		byte 16r22, byte 16r22, byte 16r22, byte 16r22, byte 16r22, byte 16r22, byte 16r22, byte 16r22, 
	},
	array[] of {
		byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, 
		byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, 
		byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, 
		byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, byte 16r44, 
	},
	array[] of {
		byte 16raf, byte 16r85, byte 16r33, byte 16r6b, byte 16r59, byte 16r7a, byte 16rfc, byte 16r1a, 
		byte 16r90, byte 16r0b, byte 16r2e, byte 16rb2, byte 16r1e, byte 16rc9, byte 16r49, byte 16rd2, 
		byte 16r92, byte 16rdf, byte 16r4c, byte 16r04, byte 16r7e, byte 16r0b, byte 16r21, byte 16r53, 
		byte 16r21, byte 16r86, byte 16ra5, byte 16r97, byte 16r1a, byte 16r22, byte 16r7a, byte 16r89, 
	}
),
Xts (
	big 16r00000000,
	array[] of {
		byte 16r27, byte 16r18, byte 16r28, byte 16r18, byte 16r28, byte 16r45, byte 16r90, byte 16r45, 
		byte 16r23, byte 16r53, byte 16r60, byte 16r28, byte 16r74, byte 16r71, byte 16r35, byte 16r26, 
		byte 16r31, byte 16r41, byte 16r59, byte 16r26, byte 16r53, byte 16r58, byte 16r97, byte 16r93, 
		byte 16r23, byte 16r84, byte 16r62, byte 16r64, byte 16r33, byte 16r83, byte 16r27, byte 16r95, 
	},
	array[] of {
		byte 16r00, byte 16r01, byte 16r02, byte 16r03, byte 16r04, byte 16r05, byte 16r06, byte 16r07, 
		byte 16r08, byte 16r09, byte 16r0a, byte 16r0b, byte 16r0c, byte 16r0d, byte 16r0e, byte 16r0f, 
		byte 16r10, byte 16r11, byte 16r12, byte 16r13, byte 16r14, byte 16r15, byte 16r16, byte 16r17, 
		byte 16r18, byte 16r19, byte 16r1a, byte 16r1b, byte 16r1c, byte 16r1d, byte 16r1e, byte 16r1f, 
		byte 16r20, byte 16r21, byte 16r22, byte 16r23, byte 16r24, byte 16r25, byte 16r26, byte 16r27, 
		byte 16r28, byte 16r29, byte 16r2a, byte 16r2b, byte 16r2c, byte 16r2d, byte 16r2e, byte 16r2f, 
		byte 16r30, byte 16r31, byte 16r32, byte 16r33, byte 16r34, byte 16r35, byte 16r36, byte 16r37, 
		byte 16r38, byte 16r39, byte 16r3a, byte 16r3b, byte 16r3c, byte 16r3d, byte 16r3e, byte 16r3f, 
		byte 16r40, byte 16r41, byte 16r42, byte 16r43, byte 16r44, byte 16r45, byte 16r46, byte 16r47, 
		byte 16r48, byte 16r49, byte 16r4a, byte 16r4b, byte 16r4c, byte 16r4d, byte 16r4e, byte 16r4f, 
		byte 16r50, byte 16r51, byte 16r52, byte 16r53, byte 16r54, byte 16r55, byte 16r56, byte 16r57, 
		byte 16r58, byte 16r59, byte 16r5a, byte 16r5b, byte 16r5c, byte 16r5d, byte 16r5e, byte 16r5f, 
		byte 16r60, byte 16r61, byte 16r62, byte 16r63, byte 16r64, byte 16r65, byte 16r66, byte 16r67, 
		byte 16r68, byte 16r69, byte 16r6a, byte 16r6b, byte 16r6c, byte 16r6d, byte 16r6e, byte 16r6f, 
		byte 16r70, byte 16r71, byte 16r72, byte 16r73, byte 16r74, byte 16r75, byte 16r76, byte 16r77, 
		byte 16r78, byte 16r79, byte 16r7a, byte 16r7b, byte 16r7c, byte 16r7d, byte 16r7e, byte 16r7f, 
		byte 16r80, byte 16r81, byte 16r82, byte 16r83, byte 16r84, byte 16r85, byte 16r86, byte 16r87, 
		byte 16r88, byte 16r89, byte 16r8a, byte 16r8b, byte 16r8c, byte 16r8d, byte 16r8e, byte 16r8f, 
		byte 16r90, byte 16r91, byte 16r92, byte 16r93, byte 16r94, byte 16r95, byte 16r96, byte 16r97, 
		byte 16r98, byte 16r99, byte 16r9a, byte 16r9b, byte 16r9c, byte 16r9d, byte 16r9e, byte 16r9f, 
		byte 16ra0, byte 16ra1, byte 16ra2, byte 16ra3, byte 16ra4, byte 16ra5, byte 16ra6, byte 16ra7, 
		byte 16ra8, byte 16ra9, byte 16raa, byte 16rab, byte 16rac, byte 16rad, byte 16rae, byte 16raf, 
		byte 16rb0, byte 16rb1, byte 16rb2, byte 16rb3, byte 16rb4, byte 16rb5, byte 16rb6, byte 16rb7, 
		byte 16rb8, byte 16rb9, byte 16rba, byte 16rbb, byte 16rbc, byte 16rbd, byte 16rbe, byte 16rbf, 
		byte 16rc0, byte 16rc1, byte 16rc2, byte 16rc3, byte 16rc4, byte 16rc5, byte 16rc6, byte 16rc7, 
		byte 16rc8, byte 16rc9, byte 16rca, byte 16rcb, byte 16rcc, byte 16rcd, byte 16rce, byte 16rcf, 
		byte 16rd0, byte 16rd1, byte 16rd2, byte 16rd3, byte 16rd4, byte 16rd5, byte 16rd6, byte 16rd7, 
		byte 16rd8, byte 16rd9, byte 16rda, byte 16rdb, byte 16rdc, byte 16rdd, byte 16rde, byte 16rdf, 
		byte 16re0, byte 16re1, byte 16re2, byte 16re3, byte 16re4, byte 16re5, byte 16re6, byte 16re7, 
		byte 16re8, byte 16re9, byte 16rea, byte 16reb, byte 16rec, byte 16red, byte 16ree, byte 16ref, 
		byte 16rf0, byte 16rf1, byte 16rf2, byte 16rf3, byte 16rf4, byte 16rf5, byte 16rf6, byte 16rf7, 
		byte 16rf8, byte 16rf9, byte 16rfa, byte 16rfb, byte 16rfc, byte 16rfd, byte 16rfe, byte 16rff, 
		byte 16r00, byte 16r01, byte 16r02, byte 16r03, byte 16r04, byte 16r05, byte 16r06, byte 16r07, 
		byte 16r08, byte 16r09, byte 16r0a, byte 16r0b, byte 16r0c, byte 16r0d, byte 16r0e, byte 16r0f, 
		byte 16r10, byte 16r11, byte 16r12, byte 16r13, byte 16r14, byte 16r15, byte 16r16, byte 16r17, 
		byte 16r18, byte 16r19, byte 16r1a, byte 16r1b, byte 16r1c, byte 16r1d, byte 16r1e, byte 16r1f, 
		byte 16r20, byte 16r21, byte 16r22, byte 16r23, byte 16r24, byte 16r25, byte 16r26, byte 16r27, 
		byte 16r28, byte 16r29, byte 16r2a, byte 16r2b, byte 16r2c, byte 16r2d, byte 16r2e, byte 16r2f, 
		byte 16r30, byte 16r31, byte 16r32, byte 16r33, byte 16r34, byte 16r35, byte 16r36, byte 16r37, 
		byte 16r38, byte 16r39, byte 16r3a, byte 16r3b, byte 16r3c, byte 16r3d, byte 16r3e, byte 16r3f, 
		byte 16r40, byte 16r41, byte 16r42, byte 16r43, byte 16r44, byte 16r45, byte 16r46, byte 16r47, 
		byte 16r48, byte 16r49, byte 16r4a, byte 16r4b, byte 16r4c, byte 16r4d, byte 16r4e, byte 16r4f, 
		byte 16r50, byte 16r51, byte 16r52, byte 16r53, byte 16r54, byte 16r55, byte 16r56, byte 16r57, 
		byte 16r58, byte 16r59, byte 16r5a, byte 16r5b, byte 16r5c, byte 16r5d, byte 16r5e, byte 16r5f, 
		byte 16r60, byte 16r61, byte 16r62, byte 16r63, byte 16r64, byte 16r65, byte 16r66, byte 16r67, 
		byte 16r68, byte 16r69, byte 16r6a, byte 16r6b, byte 16r6c, byte 16r6d, byte 16r6e, byte 16r6f, 
		byte 16r70, byte 16r71, byte 16r72, byte 16r73, byte 16r74, byte 16r75, byte 16r76, byte 16r77, 
		byte 16r78, byte 16r79, byte 16r7a, byte 16r7b, byte 16r7c, byte 16r7d, byte 16r7e, byte 16r7f, 
		byte 16r80, byte 16r81, byte 16r82, byte 16r83, byte 16r84, byte 16r85, byte 16r86, byte 16r87, 
		byte 16r88, byte 16r89, byte 16r8a, byte 16r8b, byte 16r8c, byte 16r8d, byte 16r8e, byte 16r8f, 
		byte 16r90, byte 16r91, byte 16r92, byte 16r93, byte 16r94, byte 16r95, byte 16r96, byte 16r97, 
		byte 16r98, byte 16r99, byte 16r9a, byte 16r9b, byte 16r9c, byte 16r9d, byte 16r9e, byte 16r9f, 
		byte 16ra0, byte 16ra1, byte 16ra2, byte 16ra3, byte 16ra4, byte 16ra5, byte 16ra6, byte 16ra7, 
		byte 16ra8, byte 16ra9, byte 16raa, byte 16rab, byte 16rac, byte 16rad, byte 16rae, byte 16raf, 
		byte 16rb0, byte 16rb1, byte 16rb2, byte 16rb3, byte 16rb4, byte 16rb5, byte 16rb6, byte 16rb7, 
		byte 16rb8, byte 16rb9, byte 16rba, byte 16rbb, byte 16rbc, byte 16rbd, byte 16rbe, byte 16rbf, 
		byte 16rc0, byte 16rc1, byte 16rc2, byte 16rc3, byte 16rc4, byte 16rc5, byte 16rc6, byte 16rc7, 
		byte 16rc8, byte 16rc9, byte 16rca, byte 16rcb, byte 16rcc, byte 16rcd, byte 16rce, byte 16rcf, 
		byte 16rd0, byte 16rd1, byte 16rd2, byte 16rd3, byte 16rd4, byte 16rd5, byte 16rd6, byte 16rd7, 
		byte 16rd8, byte 16rd9, byte 16rda, byte 16rdb, byte 16rdc, byte 16rdd, byte 16rde, byte 16rdf, 
		byte 16re0, byte 16re1, byte 16re2, byte 16re3, byte 16re4, byte 16re5, byte 16re6, byte 16re7, 
		byte 16re8, byte 16re9, byte 16rea, byte 16reb, byte 16rec, byte 16red, byte 16ree, byte 16ref, 
		byte 16rf0, byte 16rf1, byte 16rf2, byte 16rf3, byte 16rf4, byte 16rf5, byte 16rf6, byte 16rf7, 
		byte 16rf8, byte 16rf9, byte 16rfa, byte 16rfb, byte 16rfc, byte 16rfd, byte 16rfe, byte 16rff, 
	},
	array[] of {
		byte 16r27, byte 16ra7, byte 16r47, byte 16r9b, byte 16ref, byte 16ra1, byte 16rd4, byte 16r76, 
		byte 16r48, byte 16r9f, byte 16r30, byte 16r8c, byte 16rd4, byte 16rcf, byte 16ra6, byte 16re2, 
		byte 16ra9, byte 16r6e, byte 16r4b, byte 16rbe, byte 16r32, byte 16r08, byte 16rff, byte 16r25, 
		byte 16r28, byte 16r7d, byte 16rd3, byte 16r81, byte 16r96, byte 16r16, byte 16re8, byte 16r9c, 
		byte 16rc7, byte 16r8c, byte 16rf7, byte 16rf5, byte 16re5, byte 16r43, byte 16r44, byte 16r5f, 
		byte 16r83, byte 16r33, byte 16rd8, byte 16rfa, byte 16r7f, byte 16r56, byte 16r00, byte 16r00, 
		byte 16r05, byte 16r27, byte 16r9f, byte 16ra5, byte 16rd8, byte 16rb5, byte 16re4, byte 16rad, 
		byte 16r40, byte 16re7, byte 16r36, byte 16rdd, byte 16rb4, byte 16rd3, byte 16r54, byte 16r12, 
		byte 16r32, byte 16r80, byte 16r63, byte 16rfd, byte 16r2a, byte 16rab, byte 16r53, byte 16re5, 
		byte 16rea, byte 16r1e, byte 16r0a, byte 16r9f, byte 16r33, byte 16r25, byte 16r00, byte 16ra5, 
		byte 16rdf, byte 16r94, byte 16r87, byte 16rd0, byte 16r7a, byte 16r5c, byte 16r92, byte 16rcc, 
		byte 16r51, byte 16r2c, byte 16r88, byte 16r66, byte 16rc7, byte 16re8, byte 16r60, byte 16rce, 
		byte 16r93, byte 16rfd, byte 16rf1, byte 16r66, byte 16ra2, byte 16r49, byte 16r12, byte 16rb4, 
		byte 16r22, byte 16r97, byte 16r61, byte 16r46, byte 16rae, byte 16r20, byte 16rce, byte 16r84, 
		byte 16r6b, byte 16rb7, byte 16rdc, byte 16r9b, byte 16ra9, byte 16r4a, byte 16r76, byte 16r7a, 
		byte 16rae, byte 16rf2, byte 16r0c, byte 16r0d, byte 16r61, byte 16rad, byte 16r02, byte 16r65, 
		byte 16r5e, byte 16ra9, byte 16r2d, byte 16rc4, byte 16rc4, byte 16re4, byte 16r1a, byte 16r89, 
		byte 16r52, byte 16rc6, byte 16r51, byte 16rd3, byte 16r31, byte 16r74, byte 16rbe, byte 16r51, 
		byte 16ra1, byte 16r0c, byte 16r42, byte 16r11, byte 16r10, byte 16re6, byte 16rd8, byte 16r15, 
		byte 16r88, byte 16red, byte 16re8, byte 16r21, byte 16r03, byte 16ra2, byte 16r52, byte 16rd8, 
		byte 16ra7, byte 16r50, byte 16re8, byte 16r76, byte 16r8d, byte 16ref, byte 16rff, byte 16red, 
		byte 16r91, byte 16r22, byte 16r81, byte 16r0a, byte 16rae, byte 16rb9, byte 16r9f, byte 16r91, 
		byte 16r72, byte 16raf, byte 16r82, byte 16rb6, byte 16r04, byte 16rdc, byte 16r4b, byte 16r8e, 
		byte 16r51, byte 16rbc, byte 16rb0, byte 16r82, byte 16r35, byte 16ra6, byte 16rf4, byte 16r34, 
		byte 16r13, byte 16r32, byte 16re4, byte 16rca, byte 16r60, byte 16r48, byte 16r2a, byte 16r4b, 
		byte 16ra1, byte 16ra0, byte 16r3b, byte 16r3e, byte 16r65, byte 16r00, byte 16r8f, byte 16rc5, 
		byte 16rda, byte 16r76, byte 16rb7, byte 16r0b, byte 16rf1, byte 16r69, byte 16r0d, byte 16rb4, 
		byte 16rea, byte 16re2, byte 16r9c, byte 16r5f, byte 16r1b, byte 16rad, byte 16rd0, byte 16r3c, 
		byte 16r5c, byte 16rcf, byte 16r2a, byte 16r55, byte 16rd7, byte 16r05, byte 16rdd, byte 16rcd, 
		byte 16r86, byte 16rd4, byte 16r49, byte 16r51, byte 16r1c, byte 16reb, byte 16r7e, byte 16rc3, 
		byte 16r0b, byte 16rf1, byte 16r2b, byte 16r1f, byte 16ra3, byte 16r5b, byte 16r91, byte 16r3f, 
		byte 16r9f, byte 16r74, byte 16r7a, byte 16r8a, byte 16rfd, byte 16r1b, byte 16r13, byte 16r0e, 
		byte 16r94, byte 16rbf, byte 16rf9, byte 16r4e, byte 16rff, byte 16rd0, byte 16r1a, byte 16r91, 
		byte 16r73, byte 16r5c, byte 16ra1, byte 16r72, byte 16r6a, byte 16rcd, byte 16r0b, byte 16r19, 
		byte 16r7c, byte 16r4e, byte 16r5b, byte 16r03, byte 16r39, byte 16r36, byte 16r97, byte 16re1, 
		byte 16r26, byte 16r82, byte 16r6f, byte 16rb6, byte 16rbb, byte 16rde, byte 16r8e, byte 16rcc, 
		byte 16r1e, byte 16r08, byte 16r29, byte 16r85, byte 16r16, byte 16re2, byte 16rc9, byte 16red, 
		byte 16r03, byte 16rff, byte 16r3c, byte 16r1b, byte 16r78, byte 16r60, byte 16rf6, byte 16rde, 
		byte 16r76, byte 16rd4, byte 16rce, byte 16rcd, byte 16r94, byte 16rc8, byte 16r11, byte 16r98, 
		byte 16r55, byte 16ref, byte 16r52, byte 16r97, byte 16rca, byte 16r67, byte 16re9, byte 16rf3, 
		byte 16re7, byte 16rff, byte 16r72, byte 16rb1, byte 16re9, byte 16r97, byte 16r85, byte 16rca, 
		byte 16r0a, byte 16r7e, byte 16r77, byte 16r20, byte 16rc5, byte 16rb3, byte 16r6d, byte 16rc6, 
		byte 16rd7, byte 16r2c, byte 16rac, byte 16r95, byte 16r74, byte 16rc8, byte 16rcb, byte 16rbc, 
		byte 16r2f, byte 16r80, byte 16r1e, byte 16r23, byte 16re5, byte 16r6f, byte 16rd3, byte 16r44, 
		byte 16rb0, byte 16r7f, byte 16r22, byte 16r15, byte 16r4b, byte 16reb, byte 16ra0, byte 16rf0, 
		byte 16r8c, byte 16re8, byte 16r89, byte 16r1e, byte 16r64, byte 16r3e, byte 16rd9, byte 16r95, 
		byte 16rc9, byte 16r4d, byte 16r9a, byte 16r69, byte 16rc9, byte 16rf1, byte 16rb5, byte 16rf4, 
		byte 16r99, byte 16r02, byte 16r7a, byte 16r78, byte 16r57, byte 16r2a, byte 16ree, byte 16rbd, 
		byte 16r74, byte 16rd2, byte 16r0c, byte 16rc3, byte 16r98, byte 16r81, byte 16rc2, byte 16r13, 
		byte 16ree, byte 16r77, byte 16r0b, byte 16r10, byte 16r10, byte 16re4, byte 16rbe, byte 16ra7, 
		byte 16r18, byte 16r84, byte 16r69, byte 16r77, byte 16rae, byte 16r11, byte 16r9f, byte 16r7a, 
		byte 16r02, byte 16r3a, byte 16rb5, byte 16r8c, byte 16rca, byte 16r0a, byte 16rd7, byte 16r52, 
		byte 16raf, byte 16re6, byte 16r56, byte 16rbb, byte 16r3c, byte 16r17, byte 16r25, byte 16r6a, 
		byte 16r9f, byte 16r6e, byte 16r9b, byte 16rf1, byte 16r9f, byte 16rdd, byte 16r5a, byte 16r38, 
		byte 16rfc, byte 16r82, byte 16rbb, byte 16re8, byte 16r72, byte 16rc5, byte 16r53, byte 16r9e, 
		byte 16rdb, byte 16r60, byte 16r9e, byte 16rf4, byte 16rf7, byte 16r9c, byte 16r20, byte 16r3e, 
		byte 16rbb, byte 16r14, byte 16r0f, byte 16r2e, byte 16r58, byte 16r3c, byte 16rb2, byte 16rad, 
		byte 16r15, byte 16rb4, byte 16raa, byte 16r5b, byte 16r65, byte 16r50, byte 16r16, byte 16ra8, 
		byte 16r44, byte 16r92, byte 16r77, byte 16rdb, byte 16rd4, byte 16r77, byte 16ref, byte 16r2c, 
		byte 16r8d, byte 16r6c, byte 16r01, byte 16r7d, byte 16rb7, byte 16r38, byte 16rb1, byte 16r8d, 
		byte 16reb, byte 16r4a, byte 16r42, byte 16r7d, byte 16r19, byte 16r23, byte 16rce, byte 16r3f, 
		byte 16rf2, byte 16r62, byte 16r73, byte 16r57, byte 16r79, byte 16ra4, byte 16r18, byte 16rf2, 
		byte 16r0a, byte 16r28, byte 16r2d, byte 16rf9, byte 16r20, byte 16r14, byte 16r7b, byte 16rea, 
		byte 16rbe, byte 16r42, byte 16r1e, byte 16re5, byte 16r31, byte 16r9d, byte 16r05, byte 16r68, 
	}
),
Xts (
	big 16r00000001,
	array[] of {
		byte 16r27, byte 16r18, byte 16r28, byte 16r18, byte 16r28, byte 16r45, byte 16r90, byte 16r45, 
		byte 16r23, byte 16r53, byte 16r60, byte 16r28, byte 16r74, byte 16r71, byte 16r35, byte 16r26, 
		byte 16r31, byte 16r41, byte 16r59, byte 16r26, byte 16r53, byte 16r58, byte 16r97, byte 16r93, 
		byte 16r23, byte 16r84, byte 16r62, byte 16r64, byte 16r33, byte 16r83, byte 16r27, byte 16r95, 
	},
	array[] of {
		byte 16r27, byte 16ra7, byte 16r47, byte 16r9b, byte 16ref, byte 16ra1, byte 16rd4, byte 16r76, 
		byte 16r48, byte 16r9f, byte 16r30, byte 16r8c, byte 16rd4, byte 16rcf, byte 16ra6, byte 16re2, 
		byte 16ra9, byte 16r6e, byte 16r4b, byte 16rbe, byte 16r32, byte 16r08, byte 16rff, byte 16r25, 
		byte 16r28, byte 16r7d, byte 16rd3, byte 16r81, byte 16r96, byte 16r16, byte 16re8, byte 16r9c, 
		byte 16rc7, byte 16r8c, byte 16rf7, byte 16rf5, byte 16re5, byte 16r43, byte 16r44, byte 16r5f, 
		byte 16r83, byte 16r33, byte 16rd8, byte 16rfa, byte 16r7f, byte 16r56, byte 16r00, byte 16r00, 
		byte 16r05, byte 16r27, byte 16r9f, byte 16ra5, byte 16rd8, byte 16rb5, byte 16re4, byte 16rad, 
		byte 16r40, byte 16re7, byte 16r36, byte 16rdd, byte 16rb4, byte 16rd3, byte 16r54, byte 16r12, 
		byte 16r32, byte 16r80, byte 16r63, byte 16rfd, byte 16r2a, byte 16rab, byte 16r53, byte 16re5, 
		byte 16rea, byte 16r1e, byte 16r0a, byte 16r9f, byte 16r33, byte 16r25, byte 16r00, byte 16ra5, 
		byte 16rdf, byte 16r94, byte 16r87, byte 16rd0, byte 16r7a, byte 16r5c, byte 16r92, byte 16rcc, 
		byte 16r51, byte 16r2c, byte 16r88, byte 16r66, byte 16rc7, byte 16re8, byte 16r60, byte 16rce, 
		byte 16r93, byte 16rfd, byte 16rf1, byte 16r66, byte 16ra2, byte 16r49, byte 16r12, byte 16rb4, 
		byte 16r22, byte 16r97, byte 16r61, byte 16r46, byte 16rae, byte 16r20, byte 16rce, byte 16r84, 
		byte 16r6b, byte 16rb7, byte 16rdc, byte 16r9b, byte 16ra9, byte 16r4a, byte 16r76, byte 16r7a, 
		byte 16rae, byte 16rf2, byte 16r0c, byte 16r0d, byte 16r61, byte 16rad, byte 16r02, byte 16r65, 
		byte 16r5e, byte 16ra9, byte 16r2d, byte 16rc4, byte 16rc4, byte 16re4, byte 16r1a, byte 16r89, 
		byte 16r52, byte 16rc6, byte 16r51, byte 16rd3, byte 16r31, byte 16r74, byte 16rbe, byte 16r51, 
		byte 16ra1, byte 16r0c, byte 16r42, byte 16r11, byte 16r10, byte 16re6, byte 16rd8, byte 16r15, 
		byte 16r88, byte 16red, byte 16re8, byte 16r21, byte 16r03, byte 16ra2, byte 16r52, byte 16rd8, 
		byte 16ra7, byte 16r50, byte 16re8, byte 16r76, byte 16r8d, byte 16ref, byte 16rff, byte 16red, 
		byte 16r91, byte 16r22, byte 16r81, byte 16r0a, byte 16rae, byte 16rb9, byte 16r9f, byte 16r91, 
		byte 16r72, byte 16raf, byte 16r82, byte 16rb6, byte 16r04, byte 16rdc, byte 16r4b, byte 16r8e, 
		byte 16r51, byte 16rbc, byte 16rb0, byte 16r82, byte 16r35, byte 16ra6, byte 16rf4, byte 16r34, 
		byte 16r13, byte 16r32, byte 16re4, byte 16rca, byte 16r60, byte 16r48, byte 16r2a, byte 16r4b, 
		byte 16ra1, byte 16ra0, byte 16r3b, byte 16r3e, byte 16r65, byte 16r00, byte 16r8f, byte 16rc5, 
		byte 16rda, byte 16r76, byte 16rb7, byte 16r0b, byte 16rf1, byte 16r69, byte 16r0d, byte 16rb4, 
		byte 16rea, byte 16re2, byte 16r9c, byte 16r5f, byte 16r1b, byte 16rad, byte 16rd0, byte 16r3c, 
		byte 16r5c, byte 16rcf, byte 16r2a, byte 16r55, byte 16rd7, byte 16r05, byte 16rdd, byte 16rcd, 
		byte 16r86, byte 16rd4, byte 16r49, byte 16r51, byte 16r1c, byte 16reb, byte 16r7e, byte 16rc3, 
		byte 16r0b, byte 16rf1, byte 16r2b, byte 16r1f, byte 16ra3, byte 16r5b, byte 16r91, byte 16r3f, 
		byte 16r9f, byte 16r74, byte 16r7a, byte 16r8a, byte 16rfd, byte 16r1b, byte 16r13, byte 16r0e, 
		byte 16r94, byte 16rbf, byte 16rf9, byte 16r4e, byte 16rff, byte 16rd0, byte 16r1a, byte 16r91, 
		byte 16r73, byte 16r5c, byte 16ra1, byte 16r72, byte 16r6a, byte 16rcd, byte 16r0b, byte 16r19, 
		byte 16r7c, byte 16r4e, byte 16r5b, byte 16r03, byte 16r39, byte 16r36, byte 16r97, byte 16re1, 
		byte 16r26, byte 16r82, byte 16r6f, byte 16rb6, byte 16rbb, byte 16rde, byte 16r8e, byte 16rcc, 
		byte 16r1e, byte 16r08, byte 16r29, byte 16r85, byte 16r16, byte 16re2, byte 16rc9, byte 16red, 
		byte 16r03, byte 16rff, byte 16r3c, byte 16r1b, byte 16r78, byte 16r60, byte 16rf6, byte 16rde, 
		byte 16r76, byte 16rd4, byte 16rce, byte 16rcd, byte 16r94, byte 16rc8, byte 16r11, byte 16r98, 
		byte 16r55, byte 16ref, byte 16r52, byte 16r97, byte 16rca, byte 16r67, byte 16re9, byte 16rf3, 
		byte 16re7, byte 16rff, byte 16r72, byte 16rb1, byte 16re9, byte 16r97, byte 16r85, byte 16rca, 
		byte 16r0a, byte 16r7e, byte 16r77, byte 16r20, byte 16rc5, byte 16rb3, byte 16r6d, byte 16rc6, 
		byte 16rd7, byte 16r2c, byte 16rac, byte 16r95, byte 16r74, byte 16rc8, byte 16rcb, byte 16rbc, 
		byte 16r2f, byte 16r80, byte 16r1e, byte 16r23, byte 16re5, byte 16r6f, byte 16rd3, byte 16r44, 
		byte 16rb0, byte 16r7f, byte 16r22, byte 16r15, byte 16r4b, byte 16reb, byte 16ra0, byte 16rf0, 
		byte 16r8c, byte 16re8, byte 16r89, byte 16r1e, byte 16r64, byte 16r3e, byte 16rd9, byte 16r95, 
		byte 16rc9, byte 16r4d, byte 16r9a, byte 16r69, byte 16rc9, byte 16rf1, byte 16rb5, byte 16rf4, 
		byte 16r99, byte 16r02, byte 16r7a, byte 16r78, byte 16r57, byte 16r2a, byte 16ree, byte 16rbd, 
		byte 16r74, byte 16rd2, byte 16r0c, byte 16rc3, byte 16r98, byte 16r81, byte 16rc2, byte 16r13, 
		byte 16ree, byte 16r77, byte 16r0b, byte 16r10, byte 16r10, byte 16re4, byte 16rbe, byte 16ra7, 
		byte 16r18, byte 16r84, byte 16r69, byte 16r77, byte 16rae, byte 16r11, byte 16r9f, byte 16r7a, 
		byte 16r02, byte 16r3a, byte 16rb5, byte 16r8c, byte 16rca, byte 16r0a, byte 16rd7, byte 16r52, 
		byte 16raf, byte 16re6, byte 16r56, byte 16rbb, byte 16r3c, byte 16r17, byte 16r25, byte 16r6a, 
		byte 16r9f, byte 16r6e, byte 16r9b, byte 16rf1, byte 16r9f, byte 16rdd, byte 16r5a, byte 16r38, 
		byte 16rfc, byte 16r82, byte 16rbb, byte 16re8, byte 16r72, byte 16rc5, byte 16r53, byte 16r9e, 
		byte 16rdb, byte 16r60, byte 16r9e, byte 16rf4, byte 16rf7, byte 16r9c, byte 16r20, byte 16r3e, 
		byte 16rbb, byte 16r14, byte 16r0f, byte 16r2e, byte 16r58, byte 16r3c, byte 16rb2, byte 16rad, 
		byte 16r15, byte 16rb4, byte 16raa, byte 16r5b, byte 16r65, byte 16r50, byte 16r16, byte 16ra8, 
		byte 16r44, byte 16r92, byte 16r77, byte 16rdb, byte 16rd4, byte 16r77, byte 16ref, byte 16r2c, 
		byte 16r8d, byte 16r6c, byte 16r01, byte 16r7d, byte 16rb7, byte 16r38, byte 16rb1, byte 16r8d, 
		byte 16reb, byte 16r4a, byte 16r42, byte 16r7d, byte 16r19, byte 16r23, byte 16rce, byte 16r3f, 
		byte 16rf2, byte 16r62, byte 16r73, byte 16r57, byte 16r79, byte 16ra4, byte 16r18, byte 16rf2, 
		byte 16r0a, byte 16r28, byte 16r2d, byte 16rf9, byte 16r20, byte 16r14, byte 16r7b, byte 16rea, 
		byte 16rbe, byte 16r42, byte 16r1e, byte 16re5, byte 16r31, byte 16r9d, byte 16r05, byte 16r68, 
	},
	array[] of {
		byte 16r26, byte 16r4d, byte 16r3c, byte 16ra8, byte 16r51, byte 16r21, byte 16r94, byte 16rfe, 
		byte 16rc3, byte 16r12, byte 16rc8, byte 16rc9, byte 16r89, byte 16r1f, byte 16r27, byte 16r9f, 
		byte 16ref, byte 16rdd, byte 16r60, byte 16r8d, byte 16r0c, byte 16r02, byte 16r7b, byte 16r60, 
		byte 16r48, byte 16r3a, byte 16r3f, byte 16ra8, byte 16r11, byte 16rd6, byte 16r5e, byte 16re5, 
		byte 16r9d, byte 16r52, byte 16rd9, byte 16re4, byte 16r0e, byte 16rc5, byte 16r67, byte 16r2d, 
		byte 16r81, byte 16r53, byte 16r2b, byte 16r38, byte 16rb6, byte 16rb0, byte 16r89, byte 16rce, 
		byte 16r95, byte 16r1f, byte 16r0f, byte 16r9c, byte 16r35, byte 16r59, byte 16r0b, byte 16r8b, 
		byte 16r97, byte 16r8d, byte 16r17, byte 16r52, byte 16r13, byte 16rf3, byte 16r29, byte 16rbb, 
		byte 16r1c, byte 16r2f, byte 16rd3, byte 16r0f, byte 16r2f, byte 16r7f, byte 16r30, byte 16r49, 
		byte 16r2a, byte 16r61, byte 16ra5, byte 16r32, byte 16ra7, byte 16r9f, byte 16r51, byte 16rd3, 
		byte 16r6f, byte 16r5e, byte 16r31, byte 16ra7, byte 16rc9, byte 16ra1, byte 16r2c, byte 16r28, 
		byte 16r60, byte 16r82, byte 16rff, byte 16r7d, byte 16r23, byte 16r94, byte 16rd1, byte 16r8f, 
		byte 16r78, byte 16r3e, byte 16r1a, byte 16r8e, byte 16r72, byte 16rc7, byte 16r22, byte 16rca, 
		byte 16raa, byte 16ra5, byte 16r2d, byte 16r8f, byte 16r06, byte 16r56, byte 16r57, byte 16rd2, 
		byte 16r63, byte 16r1f, byte 16rd2, byte 16r5b, byte 16rfd, byte 16r8e, byte 16r5b, byte 16raa, 
		byte 16rd6, byte 16re5, byte 16r27, byte 16rd7, byte 16r63, byte 16r51, byte 16r75, byte 16r01, 
		byte 16rc6, byte 16r8c, byte 16r5e, byte 16rdc, byte 16r3c, byte 16rdd, byte 16r55, byte 16r43, 
		byte 16r5c, byte 16r53, byte 16r2d, byte 16r71, byte 16r25, byte 16rc8, byte 16r61, byte 16r4d, 
		byte 16ree, byte 16rd9, byte 16rad, byte 16raa, byte 16r3a, byte 16rca, byte 16rde, byte 16r58, 
		byte 16r88, byte 16rb8, byte 16r7b, byte 16ref, byte 16r64, byte 16r1c, byte 16r4c, byte 16r99, 
		byte 16r4c, byte 16r80, byte 16r91, byte 16rb5, byte 16rbc, byte 16rd3, byte 16r87, byte 16rf3, 
		byte 16r96, byte 16r3f, byte 16rb5, byte 16rbc, byte 16r37, byte 16raa, byte 16r92, byte 16r2f, 
		byte 16rbf, byte 16re3, byte 16rdf, byte 16r4e, byte 16r5b, byte 16r91, byte 16r5e, byte 16r6e, 
		byte 16rb5, byte 16r14, byte 16r71, byte 16r7b, byte 16rdd, byte 16r2a, byte 16r74, byte 16r07, 
		byte 16r9a, byte 16r50, byte 16r73, byte 16rf5, byte 16rc4, byte 16rbf, byte 16rd4, byte 16r6a, 
		byte 16rdf, byte 16r7d, byte 16r28, byte 16r2e, byte 16r7a, byte 16r39, byte 16r3a, byte 16r52, 
		byte 16r57, byte 16r9d, byte 16r11, byte 16ra0, byte 16r28, byte 16rda, byte 16r4d, byte 16r9c, 
		byte 16rd9, byte 16rc7, byte 16r71, byte 16r24, byte 16rf9, byte 16r64, byte 16r8e, byte 16re3, 
		byte 16r83, byte 16rb1, byte 16rac, byte 16r76, byte 16r39, byte 16r30, byte 16re7, byte 16r16, 
		byte 16r2a, byte 16r8d, byte 16r37, byte 16rf3, byte 16r50, byte 16rb2, byte 16rf7, byte 16r4b, 
		byte 16r84, byte 16r72, byte 16rcf, byte 16r09, byte 16r90, byte 16r20, byte 16r63, byte 16rc6, 
		byte 16rb3, byte 16r2e, byte 16r8c, byte 16r2d, byte 16r92, byte 16r90, byte 16rce, byte 16rfb, 
		byte 16rd7, byte 16r34, byte 16r6d, byte 16r1c, byte 16r77, byte 16r9a, byte 16r0d, byte 16rf5, 
		byte 16r0e, byte 16rdc, byte 16rde, byte 16r45, byte 16r31, byte 16rda, byte 16r07, byte 16rb0, 
		byte 16r99, byte 16rc6, byte 16r38, byte 16re8, byte 16r3a, byte 16r75, byte 16r59, byte 16r44, 
		byte 16rdf, byte 16r2a, byte 16ref, byte 16r1a, byte 16ra3, byte 16r17, byte 16r52, byte 16rfd, 
		byte 16r32, byte 16r3d, byte 16rcb, byte 16r71, byte 16r0f, byte 16rb4, byte 16rbf, byte 16rbb, 
		byte 16r9d, byte 16r22, byte 16rb9, byte 16r25, byte 16rbc, byte 16r35, byte 16r77, byte 16re1, 
		byte 16rb8, byte 16r94, byte 16r9e, byte 16r72, byte 16r9a, byte 16r90, byte 16rbb, byte 16raf, 
		byte 16rea, byte 16rcf, byte 16r7f, byte 16r78, byte 16r79, byte 16re7, byte 16rb1, byte 16r14, 
		byte 16r7e, byte 16r28, byte 16rba, byte 16r0b, byte 16rae, byte 16r94, byte 16r0d, byte 16rb7, 
		byte 16r95, byte 16ra6, byte 16r1b, byte 16r15, byte 16rec, byte 16rf4, byte 16rdf, byte 16r8d, 
		byte 16rb0, byte 16r7b, byte 16r82, byte 16r4b, byte 16rb0, byte 16r62, byte 16r80, byte 16r2c, 
		byte 16rc9, byte 16r8a, byte 16r95, byte 16r45, byte 16rbb, byte 16r2a, byte 16rae, byte 16red, 
		byte 16r77, byte 16rcb, byte 16r3f, byte 16rc6, byte 16rdb, byte 16r15, byte 16rdc, byte 16rd7, 
		byte 16rd8, byte 16r0d, byte 16r7d, byte 16r5b, byte 16rc4, byte 16r06, byte 16rc4, byte 16r97, 
		byte 16r0a, byte 16r34, byte 16r78, byte 16rad, byte 16ra8, byte 16r89, byte 16r9b, byte 16r32, 
		byte 16r91, byte 16r98, byte 16reb, byte 16r61, byte 16rc1, byte 16r93, byte 16rfb, byte 16r62, 
		byte 16r75, byte 16raa, byte 16r8c, byte 16ra3, byte 16r40, byte 16r34, byte 16r4a, byte 16r75, 
		byte 16ra8, byte 16r62, byte 16rae, byte 16rbe, byte 16r92, byte 16ree, byte 16re1, byte 16rce, 
		byte 16r03, byte 16r2f, byte 16rd9, byte 16r50, byte 16rb4, byte 16r7d, byte 16r77, byte 16r04, 
		byte 16ra3, byte 16r87, byte 16r69, byte 16r23, byte 16rb4, byte 16rad, byte 16r62, byte 16r84, 
		byte 16r4b, byte 16rf4, byte 16ra0, byte 16r9c, byte 16r4d, byte 16rbe, byte 16r8b, byte 16r43, 
		byte 16r97, byte 16r18, byte 16r4b, byte 16r74, byte 16r71, byte 16r36, byte 16r0c, byte 16r95, 
		byte 16r64, byte 16r88, byte 16r0a, byte 16red, byte 16rdd, byte 16rb9, byte 16rba, byte 16ra4, 
		byte 16raf, byte 16r2e, byte 16r75, byte 16r39, byte 16r4b, byte 16r08, byte 16rcd, byte 16r32, 
		byte 16rff, byte 16r47, byte 16r9c, byte 16r57, byte 16ra0, byte 16r7d, byte 16r3e, byte 16rab, 
		byte 16r5d, byte 16r54, byte 16rde, byte 16r5f, byte 16r97, byte 16r38, byte 16rb8, byte 16rd2, 
		byte 16r7f, byte 16r27, byte 16ra9, byte 16rf0, byte 16rab, byte 16r11, byte 16r79, byte 16r9d, 
		byte 16r7b, byte 16r7f, byte 16rfe, byte 16rfb, byte 16r27, byte 16r04, byte 16rc9, byte 16r5c, 
		byte 16r6a, byte 16rd1, byte 16r2c, byte 16r39, byte 16rf1, byte 16re8, byte 16r67, byte 16ra4, 
		byte 16rb7, byte 16rb1, byte 16rd7, byte 16r81, byte 16r8a, byte 16r4b, byte 16r75, byte 16r3d, 
		byte 16rfd, byte 16r2a, byte 16r89, byte 16rcc, byte 16rb4, byte 16r5e, byte 16r00, byte 16r1a, 
		byte 16r03, byte 16ra8, byte 16r67, byte 16rb1, byte 16r87, byte 16rf2, byte 16r25, byte 16rdd, 
	}
),
Xts (
	big 16r00000002,
	array[] of {
		byte 16r27, byte 16r18, byte 16r28, byte 16r18, byte 16r28, byte 16r45, byte 16r90, byte 16r45, 
		byte 16r23, byte 16r53, byte 16r60, byte 16r28, byte 16r74, byte 16r71, byte 16r35, byte 16r26, 
		byte 16r31, byte 16r41, byte 16r59, byte 16r26, byte 16r53, byte 16r58, byte 16r97, byte 16r93, 
		byte 16r23, byte 16r84, byte 16r62, byte 16r64, byte 16r33, byte 16r83, byte 16r27, byte 16r95, 
	},
	array[] of {
		byte 16r26, byte 16r4d, byte 16r3c, byte 16ra8, byte 16r51, byte 16r21, byte 16r94, byte 16rfe, 
		byte 16rc3, byte 16r12, byte 16rc8, byte 16rc9, byte 16r89, byte 16r1f, byte 16r27, byte 16r9f, 
		byte 16ref, byte 16rdd, byte 16r60, byte 16r8d, byte 16r0c, byte 16r02, byte 16r7b, byte 16r60, 
		byte 16r48, byte 16r3a, byte 16r3f, byte 16ra8, byte 16r11, byte 16rd6, byte 16r5e, byte 16re5, 
		byte 16r9d, byte 16r52, byte 16rd9, byte 16re4, byte 16r0e, byte 16rc5, byte 16r67, byte 16r2d, 
		byte 16r81, byte 16r53, byte 16r2b, byte 16r38, byte 16rb6, byte 16rb0, byte 16r89, byte 16rce, 
		byte 16r95, byte 16r1f, byte 16r0f, byte 16r9c, byte 16r35, byte 16r59, byte 16r0b, byte 16r8b, 
		byte 16r97, byte 16r8d, byte 16r17, byte 16r52, byte 16r13, byte 16rf3, byte 16r29, byte 16rbb, 
		byte 16r1c, byte 16r2f, byte 16rd3, byte 16r0f, byte 16r2f, byte 16r7f, byte 16r30, byte 16r49, 
		byte 16r2a, byte 16r61, byte 16ra5, byte 16r32, byte 16ra7, byte 16r9f, byte 16r51, byte 16rd3, 
		byte 16r6f, byte 16r5e, byte 16r31, byte 16ra7, byte 16rc9, byte 16ra1, byte 16r2c, byte 16r28, 
		byte 16r60, byte 16r82, byte 16rff, byte 16r7d, byte 16r23, byte 16r94, byte 16rd1, byte 16r8f, 
		byte 16r78, byte 16r3e, byte 16r1a, byte 16r8e, byte 16r72, byte 16rc7, byte 16r22, byte 16rca, 
		byte 16raa, byte 16ra5, byte 16r2d, byte 16r8f, byte 16r06, byte 16r56, byte 16r57, byte 16rd2, 
		byte 16r63, byte 16r1f, byte 16rd2, byte 16r5b, byte 16rfd, byte 16r8e, byte 16r5b, byte 16raa, 
		byte 16rd6, byte 16re5, byte 16r27, byte 16rd7, byte 16r63, byte 16r51, byte 16r75, byte 16r01, 
		byte 16rc6, byte 16r8c, byte 16r5e, byte 16rdc, byte 16r3c, byte 16rdd, byte 16r55, byte 16r43, 
		byte 16r5c, byte 16r53, byte 16r2d, byte 16r71, byte 16r25, byte 16rc8, byte 16r61, byte 16r4d, 
		byte 16ree, byte 16rd9, byte 16rad, byte 16raa, byte 16r3a, byte 16rca, byte 16rde, byte 16r58, 
		byte 16r88, byte 16rb8, byte 16r7b, byte 16ref, byte 16r64, byte 16r1c, byte 16r4c, byte 16r99, 
		byte 16r4c, byte 16r80, byte 16r91, byte 16rb5, byte 16rbc, byte 16rd3, byte 16r87, byte 16rf3, 
		byte 16r96, byte 16r3f, byte 16rb5, byte 16rbc, byte 16r37, byte 16raa, byte 16r92, byte 16r2f, 
		byte 16rbf, byte 16re3, byte 16rdf, byte 16r4e, byte 16r5b, byte 16r91, byte 16r5e, byte 16r6e, 
		byte 16rb5, byte 16r14, byte 16r71, byte 16r7b, byte 16rdd, byte 16r2a, byte 16r74, byte 16r07, 
		byte 16r9a, byte 16r50, byte 16r73, byte 16rf5, byte 16rc4, byte 16rbf, byte 16rd4, byte 16r6a, 
		byte 16rdf, byte 16r7d, byte 16r28, byte 16r2e, byte 16r7a, byte 16r39, byte 16r3a, byte 16r52, 
		byte 16r57, byte 16r9d, byte 16r11, byte 16ra0, byte 16r28, byte 16rda, byte 16r4d, byte 16r9c, 
		byte 16rd9, byte 16rc7, byte 16r71, byte 16r24, byte 16rf9, byte 16r64, byte 16r8e, byte 16re3, 
		byte 16r83, byte 16rb1, byte 16rac, byte 16r76, byte 16r39, byte 16r30, byte 16re7, byte 16r16, 
		byte 16r2a, byte 16r8d, byte 16r37, byte 16rf3, byte 16r50, byte 16rb2, byte 16rf7, byte 16r4b, 
		byte 16r84, byte 16r72, byte 16rcf, byte 16r09, byte 16r90, byte 16r20, byte 16r63, byte 16rc6, 
		byte 16rb3, byte 16r2e, byte 16r8c, byte 16r2d, byte 16r92, byte 16r90, byte 16rce, byte 16rfb, 
		byte 16rd7, byte 16r34, byte 16r6d, byte 16r1c, byte 16r77, byte 16r9a, byte 16r0d, byte 16rf5, 
		byte 16r0e, byte 16rdc, byte 16rde, byte 16r45, byte 16r31, byte 16rda, byte 16r07, byte 16rb0, 
		byte 16r99, byte 16rc6, byte 16r38, byte 16re8, byte 16r3a, byte 16r75, byte 16r59, byte 16r44, 
		byte 16rdf, byte 16r2a, byte 16ref, byte 16r1a, byte 16ra3, byte 16r17, byte 16r52, byte 16rfd, 
		byte 16r32, byte 16r3d, byte 16rcb, byte 16r71, byte 16r0f, byte 16rb4, byte 16rbf, byte 16rbb, 
		byte 16r9d, byte 16r22, byte 16rb9, byte 16r25, byte 16rbc, byte 16r35, byte 16r77, byte 16re1, 
		byte 16rb8, byte 16r94, byte 16r9e, byte 16r72, byte 16r9a, byte 16r90, byte 16rbb, byte 16raf, 
		byte 16rea, byte 16rcf, byte 16r7f, byte 16r78, byte 16r79, byte 16re7, byte 16rb1, byte 16r14, 
		byte 16r7e, byte 16r28, byte 16rba, byte 16r0b, byte 16rae, byte 16r94, byte 16r0d, byte 16rb7, 
		byte 16r95, byte 16ra6, byte 16r1b, byte 16r15, byte 16rec, byte 16rf4, byte 16rdf, byte 16r8d, 
		byte 16rb0, byte 16r7b, byte 16r82, byte 16r4b, byte 16rb0, byte 16r62, byte 16r80, byte 16r2c, 
		byte 16rc9, byte 16r8a, byte 16r95, byte 16r45, byte 16rbb, byte 16r2a, byte 16rae, byte 16red, 
		byte 16r77, byte 16rcb, byte 16r3f, byte 16rc6, byte 16rdb, byte 16r15, byte 16rdc, byte 16rd7, 
		byte 16rd8, byte 16r0d, byte 16r7d, byte 16r5b, byte 16rc4, byte 16r06, byte 16rc4, byte 16r97, 
		byte 16r0a, byte 16r34, byte 16r78, byte 16rad, byte 16ra8, byte 16r89, byte 16r9b, byte 16r32, 
		byte 16r91, byte 16r98, byte 16reb, byte 16r61, byte 16rc1, byte 16r93, byte 16rfb, byte 16r62, 
		byte 16r75, byte 16raa, byte 16r8c, byte 16ra3, byte 16r40, byte 16r34, byte 16r4a, byte 16r75, 
		byte 16ra8, byte 16r62, byte 16rae, byte 16rbe, byte 16r92, byte 16ree, byte 16re1, byte 16rce, 
		byte 16r03, byte 16r2f, byte 16rd9, byte 16r50, byte 16rb4, byte 16r7d, byte 16r77, byte 16r04, 
		byte 16ra3, byte 16r87, byte 16r69, byte 16r23, byte 16rb4, byte 16rad, byte 16r62, byte 16r84, 
		byte 16r4b, byte 16rf4, byte 16ra0, byte 16r9c, byte 16r4d, byte 16rbe, byte 16r8b, byte 16r43, 
		byte 16r97, byte 16r18, byte 16r4b, byte 16r74, byte 16r71, byte 16r36, byte 16r0c, byte 16r95, 
		byte 16r64, byte 16r88, byte 16r0a, byte 16red, byte 16rdd, byte 16rb9, byte 16rba, byte 16ra4, 
		byte 16raf, byte 16r2e, byte 16r75, byte 16r39, byte 16r4b, byte 16r08, byte 16rcd, byte 16r32, 
		byte 16rff, byte 16r47, byte 16r9c, byte 16r57, byte 16ra0, byte 16r7d, byte 16r3e, byte 16rab, 
		byte 16r5d, byte 16r54, byte 16rde, byte 16r5f, byte 16r97, byte 16r38, byte 16rb8, byte 16rd2, 
		byte 16r7f, byte 16r27, byte 16ra9, byte 16rf0, byte 16rab, byte 16r11, byte 16r79, byte 16r9d, 
		byte 16r7b, byte 16r7f, byte 16rfe, byte 16rfb, byte 16r27, byte 16r04, byte 16rc9, byte 16r5c, 
		byte 16r6a, byte 16rd1, byte 16r2c, byte 16r39, byte 16rf1, byte 16re8, byte 16r67, byte 16ra4, 
		byte 16rb7, byte 16rb1, byte 16rd7, byte 16r81, byte 16r8a, byte 16r4b, byte 16r75, byte 16r3d, 
		byte 16rfd, byte 16r2a, byte 16r89, byte 16rcc, byte 16rb4, byte 16r5e, byte 16r00, byte 16r1a, 
		byte 16r03, byte 16ra8, byte 16r67, byte 16rb1, byte 16r87, byte 16rf2, byte 16r25, byte 16rdd, 
	},
	array[] of {
		byte 16rfa, byte 16r76, byte 16r2a, byte 16r36, byte 16r80, byte 16rb7, byte 16r60, byte 16r07, 
		byte 16r92, byte 16r8e, byte 16rd4, byte 16ra4, byte 16rf4, byte 16r9a, byte 16r94, byte 16r56, 
		byte 16r03, byte 16r1b, byte 16r70, byte 16r47, byte 16r82, byte 16re6, byte 16r5e, byte 16r16, 
		byte 16rce, byte 16rcb, byte 16r54, byte 16red, byte 16r7d, byte 16r01, byte 16r7b, byte 16r5e, 
		byte 16r18, byte 16rab, byte 16rd6, byte 16r7b, byte 16r33, byte 16r8e, byte 16r81, byte 16r07, 
		byte 16r8f, byte 16r21, byte 16red, byte 16rb7, byte 16r86, byte 16r8d, byte 16r90, byte 16r1e, 
		byte 16rbe, byte 16r9c, byte 16r73, byte 16r1a, byte 16r7c, byte 16r18, byte 16rb5, byte 16re6, 
		byte 16rde, byte 16rc1, byte 16rd6, byte 16ra7, byte 16r2e, byte 16r07, byte 16r8a, byte 16rc9, 
		byte 16ra4, byte 16r26, byte 16r2f, byte 16r86, byte 16r0b, byte 16ree, byte 16rfa, byte 16r14, 
		byte 16rf4, byte 16re8, byte 16r21, byte 16r01, byte 16r82, byte 16r72, byte 16re4, byte 16r11, 
		byte 16ra9, byte 16r51, byte 16r50, byte 16r2b, byte 16r6e, byte 16r79, byte 16r06, byte 16r6e, 
		byte 16r84, byte 16r25, byte 16r2c, byte 16r33, byte 16r46, byte 16rf3, byte 16raa, byte 16r62, 
		byte 16r34, byte 16r43, byte 16r51, byte 16ra2, byte 16r91, byte 16rd4, byte 16rbe, byte 16rdc, 
		byte 16r7a, byte 16r07, byte 16r61, byte 16r8b, byte 16rde, byte 16ra2, byte 16raf, byte 16r63, 
		byte 16r14, byte 16r5c, byte 16rc7, byte 16ra4, byte 16rb8, byte 16rd4, byte 16r07, byte 16r06, 
		byte 16r91, byte 16rae, byte 16r89, byte 16r0c, byte 16rd6, byte 16r57, byte 16r33, byte 16re7, 
		byte 16r94, byte 16r6e, byte 16r90, byte 16r21, byte 16ra1, byte 16rdf, byte 16rfc, byte 16r4c, 
		byte 16r59, byte 16rf1, byte 16r59, byte 16r42, byte 16r5e, byte 16re6, byte 16rd5, byte 16r0c, 
		byte 16ra9, byte 16rb1, byte 16r35, byte 16rfa, byte 16r61, byte 16r62, byte 16rce, byte 16ra1, 
		byte 16r8a, byte 16r93, byte 16r98, byte 16r38, byte 16rdc, byte 16r00, byte 16r0f, byte 16rb3, 
		byte 16r86, byte 16rfa, byte 16rd0, byte 16r86, byte 16rac, byte 16rce, byte 16r5a, byte 16rc0, 
		byte 16r7c, byte 16rb2, byte 16rec, byte 16re7, byte 16rfd, byte 16r58, byte 16r0b, byte 16r00, 
		byte 16rcf, byte 16ra5, byte 16re9, byte 16r85, byte 16r89, byte 16r63, byte 16r1d, byte 16rc2, 
		byte 16r5e, byte 16r8e, byte 16r2a, byte 16r3d, byte 16raf, byte 16r2f, byte 16rfd, byte 16rec, 
		byte 16r26, byte 16r53, byte 16r16, byte 16r59, byte 16r91, byte 16r2c, byte 16r9d, byte 16r8f, 
		byte 16r7a, byte 16r15, byte 16re5, byte 16r86, byte 16r5e, byte 16ra8, byte 16rfb, byte 16r58, 
		byte 16r16, byte 16rd6, byte 16r20, byte 16r70, byte 16r52, byte 16rbd, byte 16r71, byte 16r28, 
		byte 16rcd, byte 16r74, byte 16r3c, byte 16r12, byte 16rc8, byte 16r11, byte 16r87, byte 16r91, 
		byte 16ra4, byte 16r73, byte 16r68, byte 16r11, byte 16r93, byte 16r5e, byte 16rb9, byte 16r82, 
		byte 16ra5, byte 16r32, byte 16r34, byte 16r9e, byte 16r31, byte 16rdd, byte 16r40, byte 16r1e, 
		byte 16r0b, byte 16r66, byte 16r0a, byte 16r56, byte 16r8c, byte 16rb1, byte 16ra4, byte 16r71, 
		byte 16r1f, byte 16r55, byte 16r2f, byte 16r55, byte 16rde, byte 16rd5, byte 16r9f, byte 16r1f, 
		byte 16r15, byte 16rbf, byte 16r71, byte 16r96, byte 16rb3, byte 16rca, byte 16r12, byte 16ra9, 
		byte 16r1e, byte 16r48, byte 16r8e, byte 16rf5, byte 16r9d, byte 16r64, byte 16rf3, byte 16ra0, 
		byte 16r2b, byte 16rf4, byte 16r52, byte 16r39, byte 16r49, byte 16r9a, byte 16rc6, byte 16r17, 
		byte 16r6a, byte 16re3, byte 16r21, byte 16rc4, byte 16ra2, byte 16r11, byte 16rec, byte 16r54, 
		byte 16r53, byte 16r65, byte 16r97, byte 16r1c, byte 16r5d, byte 16r3f, byte 16r4f, byte 16r09, 
		byte 16rd4, byte 16reb, byte 16r13, byte 16r9b, byte 16rfd, byte 16rf2, byte 16r07, byte 16r3d, 
		byte 16r33, byte 16r18, byte 16r0b, byte 16r21, byte 16r00, byte 16r2b, byte 16r65, byte 16rcc, 
		byte 16r98, byte 16r65, byte 16re7, byte 16r6c, byte 16rb2, byte 16r4c, byte 16rd9, byte 16r2c, 
		byte 16r87, byte 16r4c, byte 16r24, byte 16rc1, byte 16r83, byte 16r50, byte 16r39, byte 16r9a, 
		byte 16r93, byte 16r6a, byte 16rb3, byte 16r63, byte 16r70, byte 16r79, byte 16r29, byte 16r5d, 
		byte 16r76, byte 16rc4, byte 16r17, byte 16r77, byte 16r6b, byte 16r94, byte 16ref, byte 16rce, 
		byte 16r3a, byte 16r0e, byte 16rf7, byte 16r20, byte 16r6b, byte 16r15, byte 16r11, byte 16r05, 
		byte 16r19, byte 16r65, byte 16r5c, byte 16r95, byte 16r6c, byte 16rbd, byte 16r8b, byte 16r24, 
		byte 16r89, byte 16r40, byte 16r5e, byte 16re2, byte 16rb0, byte 16r9a, byte 16r6b, byte 16r6e, 
		byte 16reb, byte 16re0, byte 16rc5, byte 16r37, byte 16r90, byte 16ra1, byte 16r2a, byte 16r89, 
		byte 16r98, byte 16r37, byte 16r8b, byte 16r33, byte 16ra5, byte 16rb7, byte 16r11, byte 16r59, 
		byte 16r62, byte 16r5f, byte 16r4b, byte 16ra4, byte 16r9d, byte 16r2a, byte 16r2f, byte 16rdb, 
		byte 16ra5, byte 16r9f, byte 16rbf, byte 16r08, byte 16r97, byte 16rbc, byte 16r7a, byte 16rab, 
		byte 16rd8, byte 16rd7, byte 16r07, byte 16rdc, byte 16r14, byte 16r0a, byte 16r80, byte 16rf0, 
		byte 16rf3, byte 16r09, byte 16rf8, byte 16r35, byte 16rd3, byte 16rda, byte 16r54, byte 16rab, 
		byte 16r58, byte 16r4e, byte 16r50, byte 16r1d, byte 16rfa, byte 16r0e, byte 16re9, byte 16r77, 
		byte 16rfe, byte 16rc5, byte 16r43, byte 16rf7, byte 16r41, byte 16r86, byte 16ra8, byte 16r02, 
		byte 16rb9, byte 16ra3, byte 16r7a, byte 16rdb, byte 16r3e, byte 16r82, byte 16r91, byte 16rec, 
		byte 16ra0, byte 16r4d, byte 16r66, byte 16r52, byte 16r0d, byte 16r22, byte 16r9e, byte 16r60, 
		byte 16r40, byte 16r1e, byte 16r72, byte 16r82, byte 16rbe, byte 16rf4, byte 16r86, byte 16rae, 
		byte 16r05, byte 16r9a, byte 16ra7, byte 16r06, byte 16r96, byte 16re0, byte 16re3, byte 16r05, 
		byte 16rd7, byte 16r77, byte 16r14, byte 16r0a, byte 16r7a, byte 16r88, byte 16r3e, byte 16rcd, 
		byte 16rcb, byte 16r69, byte 16rb9, byte 16rff, byte 16r93, byte 16r8e, byte 16r8a, byte 16r42, 
		byte 16r31, byte 16r86, byte 16r4c, byte 16r69, byte 16rca, byte 16r2c, byte 16r20, byte 16r43, 
		byte 16rbe, byte 16rd0, byte 16r07, byte 16rff, byte 16r3e, byte 16r60, byte 16r5e, byte 16r01, 
		byte 16r4b, byte 16rcf, byte 16r51, byte 16r81, byte 16r38, byte 16rdc, byte 16r3a, byte 16r25, 
		byte 16rc5, byte 16re2, byte 16r36, byte 16r17, byte 16r1a, byte 16r2d, byte 16r01, byte 16rd6, 
	}
),
Xts (
	big 16r000000fd,
	array[] of {
		byte 16r27, byte 16r18, byte 16r28, byte 16r18, byte 16r28, byte 16r45, byte 16r90, byte 16r45, 
		byte 16r23, byte 16r53, byte 16r60, byte 16r28, byte 16r74, byte 16r71, byte 16r35, byte 16r26, 
		byte 16r31, byte 16r41, byte 16r59, byte 16r26, byte 16r53, byte 16r58, byte 16r97, byte 16r93, 
		byte 16r23, byte 16r84, byte 16r62, byte 16r64, byte 16r33, byte 16r83, byte 16r27, byte 16r95, 
	},
	array[] of {
		byte 16r8e, byte 16r41, byte 16rb7, byte 16r8c, byte 16r39, byte 16r0b, byte 16r5a, byte 16rf9, 
		byte 16rd7, byte 16r58, byte 16rbb, byte 16r21, byte 16r4a, byte 16r67, byte 16re9, byte 16rf6, 
		byte 16rbf, byte 16r77, byte 16r27, byte 16rb0, byte 16r9a, byte 16rc6, byte 16r12, byte 16r40, 
		byte 16r84, byte 16rc3, byte 16r76, byte 16r11, byte 16r39, byte 16r8f, byte 16ra4, byte 16r5d, 
		byte 16raa, byte 16rd9, byte 16r48, byte 16r68, byte 16r60, byte 16r0e, byte 16rd3, byte 16r91, 
		byte 16rfb, byte 16r1a, byte 16rcd, byte 16r48, byte 16r57, byte 16ra9, byte 16r5b, byte 16r46, 
		byte 16r6e, byte 16r62, byte 16ref, byte 16r9f, byte 16r4b, byte 16r37, byte 16r72, byte 16r44, 
		byte 16rd1, byte 16rc1, byte 16r52, byte 16re7, byte 16rb3, byte 16r0d, byte 16r73, byte 16r1a, 
		byte 16rad, byte 16r30, byte 16rc7, byte 16r16, byte 16rd2, byte 16r14, byte 16rb7, byte 16r07, 
		byte 16rae, byte 16rd9, byte 16r9e, byte 16rb5, byte 16rb5, byte 16re5, byte 16r80, byte 16rb3, 
		byte 16re8, byte 16r87, byte 16rcf, byte 16r74, byte 16r97, byte 16r46, byte 16r56, byte 16r51, 
		byte 16rd4, byte 16rb6, byte 16r0e, byte 16r60, byte 16r42, byte 16r05, byte 16r1d, byte 16ra3, 
		byte 16r69, byte 16r3c, byte 16r3b, byte 16r78, byte 16rc1, byte 16r44, byte 16r89, byte 16r54, 
		byte 16r3b, byte 16re8, byte 16rb6, byte 16rad, byte 16r0b, byte 16ra6, byte 16r29, byte 16r56, 
		byte 16r5b, byte 16rba, byte 16r20, byte 16r23, byte 16r13, byte 16rba, byte 16r7b, byte 16r0d, 
		byte 16r0c, byte 16r94, byte 16ra3, byte 16r25, byte 16r2b, byte 16r67, byte 16r6f, byte 16r46, 
		byte 16rcc, byte 16r02, byte 16rce, byte 16r0f, byte 16r8a, byte 16r7d, byte 16r34, byte 16rc0, 
		byte 16red, byte 16r22, byte 16r91, byte 16r29, byte 16r67, byte 16r3c, byte 16r1f, byte 16r61, 
		byte 16rae, byte 16rd5, byte 16r79, byte 16rd0, byte 16r8a, byte 16r92, byte 16r03, byte 16ra2, 
		byte 16r5a, byte 16rac, byte 16r3a, byte 16r77, byte 16re9, byte 16rdb, byte 16r60, byte 16r26, 
		byte 16r79, byte 16r96, byte 16rdb, byte 16r38, byte 16rdf, byte 16r63, byte 16r73, byte 16r56, 
		byte 16rd9, byte 16rdc, byte 16rd1, byte 16r63, byte 16r2e, byte 16r36, byte 16r99, byte 16r39, 
		byte 16rf2, byte 16ra2, byte 16r9d, byte 16r89, byte 16r34, byte 16r5c, byte 16r66, byte 16re0, 
		byte 16r50, byte 16r66, byte 16rf1, byte 16ra3, byte 16r67, byte 16r7a, byte 16ref, byte 16r18, 
		byte 16rde, byte 16ra4, byte 16r11, byte 16r3f, byte 16rae, byte 16rb6, byte 16r29, byte 16re4, 
		byte 16r67, byte 16r21, byte 16ra6, byte 16r6d, byte 16r0a, byte 16r7e, byte 16r78, byte 16r5d, 
		byte 16r3e, byte 16r29, byte 16raf, byte 16r25, byte 16r94, byte 16reb, byte 16r67, byte 16rdf, 
		byte 16ra9, byte 16r82, byte 16raf, byte 16rfe, byte 16r0a, byte 16rac, byte 16r05, byte 16r8f, 
		byte 16r6e, byte 16r15, byte 16r86, byte 16r42, byte 16r69, byte 16rb1, byte 16r35, byte 16r41, 
		byte 16r82, byte 16r61, byte 16rfc, byte 16r3a, byte 16rfb, byte 16r08, byte 16r94, byte 16r72, 
		byte 16rcf, byte 16r68, byte 16rc4, byte 16r5d, byte 16rd7, byte 16rf2, byte 16r31, byte 16rc6, 
		byte 16r24, byte 16r9b, byte 16ra0, byte 16r25, byte 16r5e, byte 16r1e, byte 16r03, byte 16r38, 
		byte 16r33, byte 16rfc, byte 16r4d, byte 16r00, byte 16ra3, byte 16rfe, byte 16r02, byte 16r13, 
		byte 16r2d, byte 16r7b, byte 16rc3, byte 16r87, byte 16r36, byte 16r14, byte 16rb8, byte 16rae, 
		byte 16re3, byte 16r42, byte 16r73, byte 16r58, byte 16r1e, byte 16ra0, byte 16r32, byte 16r5c, 
		byte 16r81, byte 16rf0, byte 16r27, byte 16r0a, byte 16rff, byte 16ra1, byte 16r36, byte 16r41, 
		byte 16rd0, byte 16r52, byte 16rd3, byte 16r6f, byte 16r07, byte 16r57, byte 16rd4, byte 16r84, 
		byte 16r01, byte 16r43, byte 16r54, byte 16rd0, byte 16r2d, byte 16r68, byte 16r83, byte 16rca, 
		byte 16r15, byte 16rc2, byte 16r4d, byte 16r8c, byte 16r39, byte 16r56, byte 16rb1, byte 16rbd, 
		byte 16r02, byte 16r7b, byte 16rcf, byte 16r41, byte 16rf1, byte 16r51, byte 16rfd, byte 16r80, 
		byte 16r23, byte 16rc5, byte 16r34, byte 16r0e, byte 16r56, byte 16r06, byte 16rf3, byte 16r7e, 
		byte 16r90, byte 16rfd, byte 16rb8, byte 16r7c, byte 16r86, byte 16rfb, byte 16r4f, byte 16ra6, 
		byte 16r34, byte 16rb3, byte 16r71, byte 16r8a, byte 16r30, byte 16rba, byte 16rce, byte 16r06, 
		byte 16ra6, byte 16r6e, byte 16raf, byte 16r8f, byte 16r63, byte 16rc4, byte 16raa, byte 16r3b, 
		byte 16r63, byte 16r78, byte 16r26, byte 16ra8, byte 16r7f, byte 16re8, byte 16rcf, byte 16ra4, 
		byte 16r42, byte 16r82, byte 16re9, byte 16r2c, byte 16rb1, byte 16r61, byte 16r5a, byte 16rf3, 
		byte 16ra2, byte 16r8e, byte 16r53, byte 16rbc, byte 16r74, byte 16rc7, byte 16rcb, byte 16ra1, 
		byte 16ra0, byte 16r97, byte 16r7b, byte 16re9, byte 16r06, byte 16r5d, byte 16r0c, byte 16r1a, 
		byte 16r5d, byte 16rec, byte 16r6c, byte 16r54, byte 16rae, byte 16r38, byte 16rd3, byte 16r7f, 
		byte 16r37, byte 16raa, byte 16r35, byte 16r28, byte 16r3e, byte 16r04, byte 16r8e, byte 16r55, 
		byte 16r30, byte 16ra8, byte 16r5c, byte 16r4e, byte 16r7a, byte 16r29, byte 16rd7, byte 16rb9, 
		byte 16r2e, byte 16rc0, byte 16rc3, byte 16r16, byte 16r9c, byte 16rdf, byte 16r2a, byte 16r80, 
		byte 16r5c, byte 16r76, byte 16r04, byte 16rbc, byte 16re6, byte 16r00, byte 16r49, byte 16rb9, 
		byte 16rfb, byte 16r7b, byte 16r8e, byte 16raa, byte 16rc1, byte 16r0f, byte 16r51, byte 16rae, 
		byte 16r23, byte 16r79, byte 16r4c, byte 16reb, byte 16ra6, byte 16r8b, byte 16rb5, byte 16r81, 
		byte 16r12, byte 16re2, byte 16r93, byte 16rb9, byte 16rb6, byte 16r92, byte 16rca, byte 16r72, 
		byte 16r1b, byte 16r37, byte 16rc6, byte 16r62, byte 16rf8, byte 16r57, byte 16r4e, byte 16rd4, 
		byte 16rdb, byte 16ra6, byte 16rf8, byte 16r8e, byte 16r17, byte 16r08, byte 16r81, byte 16rc8, 
		byte 16r2c, byte 16rdd, byte 16rc1, byte 16r03, byte 16r4a, byte 16r0c, byte 16ra7, byte 16re2, 
		byte 16r84, byte 16rbf, byte 16r09, byte 16r62, byte 16rb6, byte 16rb2, byte 16r62, byte 16r92, 
		byte 16rd8, byte 16r36, byte 16rfa, byte 16r9f, byte 16r73, byte 16rc1, byte 16rac, byte 16r77, 
		byte 16r0e, byte 16ref, byte 16r0f, byte 16r2d, byte 16r3a, byte 16r1e, byte 16raf, byte 16r61, 
		byte 16rd3, byte 16re0, byte 16r35, byte 16r55, byte 16rfd, byte 16r42, byte 16r4e, byte 16red, 
		byte 16rd6, byte 16r7e, byte 16r18, byte 16ra1, byte 16r80, byte 16r94, byte 16rf8, byte 16r88, 
	},
	array[] of {
		byte 16rd5, byte 16r5f, byte 16r68, byte 16r4f, byte 16r81, byte 16rf4, byte 16r42, byte 16r6e, 
		byte 16r9f, byte 16rde, byte 16r92, byte 16ra5, byte 16rff, byte 16r02, byte 16rdf, byte 16r2a, 
		byte 16rc8, byte 16r96, byte 16raf, byte 16r63, byte 16r96, byte 16r28, byte 16r88, byte 16ra9, 
		byte 16r79, byte 16r10, byte 16rc1, byte 16r37, byte 16r9e, byte 16r20, byte 16rb0, byte 16ra3, 
		byte 16rb1, byte 16rdb, byte 16r61, byte 16r3f, byte 16rb7, byte 16rfe, byte 16r2e, byte 16r07, 
		byte 16r00, byte 16r43, byte 16r29, byte 16rea, byte 16r5c, byte 16r22, byte 16rbf, byte 16rd3, 
		byte 16r3e, byte 16r3d, byte 16rbe, byte 16r4c, byte 16rf5, byte 16r8c, byte 16rc6, byte 16r08, 
		byte 16rc2, byte 16rc2, byte 16r6c, byte 16r19, byte 16ra2, byte 16re2, byte 16rfe, byte 16r22, 
		byte 16rf9, byte 16r87, byte 16r32, byte 16rc2, byte 16rb5, byte 16rcb, byte 16r84, byte 16r4c, 
		byte 16rc6, byte 16rc0, byte 16r70, byte 16r2d, byte 16r91, byte 16re1, byte 16rd5, byte 16r0f, 
		byte 16rc4, byte 16r38, byte 16r2a, byte 16r7e, byte 16rba, byte 16r56, byte 16r35, byte 16rcd, 
		byte 16r60, byte 16r24, byte 16r32, byte 16ra2, byte 16r30, byte 16r6a, byte 16rc4, byte 16rce, 
		byte 16r82, byte 16rf8, byte 16rd7, byte 16r0c, byte 16r8d, byte 16r9b, byte 16rc1, byte 16r5f, 
		byte 16r91, byte 16r8f, byte 16re7, byte 16r1e, byte 16r74, byte 16rc6, byte 16r22, byte 16rd5, 
		byte 16rcf, byte 16r71, byte 16r17, byte 16r8b, byte 16rf6, byte 16re0, byte 16rb9, byte 16rcc, 
		byte 16r9f, byte 16r2b, byte 16r41, byte 16rdd, byte 16r8d, byte 16rbe, byte 16r44, byte 16r1c, 
		byte 16r41, byte 16rcd, byte 16r0c, byte 16r73, byte 16ra6, byte 16rdc, byte 16r47, byte 16ra3, 
		byte 16r48, byte 16rf6, byte 16r70, byte 16r2f, byte 16r9d, byte 16r0e, byte 16r9b, byte 16r1b, 
		byte 16r14, byte 16r31, byte 16re9, byte 16r48, byte 16re2, byte 16r99, byte 16rb9, byte 16rec, 
		byte 16r22, byte 16r72, byte 16rab, byte 16r2c, byte 16r5f, byte 16r0c, byte 16r7b, byte 16re8, 
		byte 16r6a, byte 16rff, byte 16ra5, byte 16rde, byte 16rc8, byte 16r7a, byte 16r0b, byte 16ree, 
		byte 16r81, byte 16rd3, byte 16rd5, byte 16r00, byte 16r07, byte 16red, byte 16raa, byte 16r2b, 
		byte 16rcf, byte 16rcc, byte 16rb3, byte 16r56, byte 16r05, byte 16r15, byte 16r5f, byte 16rf3, 
		byte 16r6e, byte 16rd8, byte 16red, byte 16rd4, byte 16ra4, byte 16r0d, byte 16rcd, byte 16r4b, 
		byte 16r24, byte 16r3a, byte 16rcd, byte 16r11, byte 16rb2, byte 16rb9, byte 16r87, byte 16rbd, 
		byte 16rbf, byte 16raf, byte 16r91, byte 16ra7, byte 16rca, byte 16rc2, byte 16r7e, byte 16r9c, 
		byte 16r5a, byte 16rea, byte 16r52, byte 16r5e, byte 16re5, byte 16r3d, byte 16re7, byte 16rb2, 
		byte 16rd3, byte 16r33, byte 16r2c, byte 16r86, byte 16r44, byte 16r40, byte 16r2b, byte 16r82, 
		byte 16r3e, byte 16r94, byte 16ra7, byte 16rdb, byte 16r26, byte 16r27, byte 16r6d, byte 16r2d, 
		byte 16r23, byte 16raa, byte 16r07, byte 16r18, byte 16r0f, byte 16r76, byte 16rb4, byte 16rfd, 
		byte 16r29, byte 16rb9, byte 16rc0, byte 16r82, byte 16r30, byte 16r99, byte 16rc9, byte 16rd6, 
		byte 16r2c, byte 16r51, byte 16r98, byte 16r80, byte 16rae, byte 16re7, byte 16re9, byte 16r69, 
		byte 16r76, byte 16r17, byte 16rc1, byte 16r49, byte 16r7d, byte 16r47, byte 16rbf, byte 16r3e, 
		byte 16r57, byte 16r19, byte 16r50, byte 16r31, byte 16r14, byte 16r21, byte 16rb6, byte 16rb7, 
		byte 16r34, byte 16rd3, byte 16r8b, byte 16r0d, byte 16rb9, byte 16r1e, byte 16rb8, byte 16r53, 
		byte 16r31, byte 16rb9, byte 16r1e, byte 16ra9, byte 16rf6, byte 16r15, byte 16r30, byte 16rf5, 
		byte 16r45, byte 16r12, byte 16ra5, byte 16ra5, byte 16r2a, byte 16r4b, byte 16rad, byte 16r58, 
		byte 16r9e, byte 16rb6, byte 16r97, byte 16r81, byte 16rd5, byte 16r37, byte 16rf2, byte 16r32, 
		byte 16r97, byte 16rbb, byte 16r45, byte 16r9b, byte 16rda, byte 16rd2, byte 16r94, byte 16r8a, 
		byte 16r29, byte 16re1, byte 16r55, byte 16r0b, byte 16rf4, byte 16r78, byte 16r7e, byte 16r0b, 
		byte 16re9, byte 16r5b, byte 16rb1, byte 16r73, byte 16rcf, byte 16r5f, byte 16rab, byte 16r17, 
		byte 16rda, byte 16rb7, byte 16ra1, byte 16r3a, byte 16r05, byte 16r2a, byte 16r63, byte 16r45, 
		byte 16r3d, byte 16r97, byte 16rcc, byte 16rec, byte 16r1a, byte 16r32, byte 16r19, byte 16r54, 
		byte 16r88, byte 16r6b, byte 16r7a, byte 16r12, byte 16r99, byte 16rfa, byte 16rae, byte 16rec, 
		byte 16rae, byte 16r35, byte 16rc6, byte 16rea, byte 16rac, byte 16ra7, byte 16r53, byte 16rb0, 
		byte 16r41, byte 16rb5, byte 16re5, byte 16rf0, byte 16r93, byte 16rbf, byte 16r83, byte 16r39, 
		byte 16r7f, byte 16rd2, byte 16r1d, byte 16rd6, byte 16rb3, byte 16r01, byte 16r20, byte 16r66, 
		byte 16rfc, byte 16rc0, byte 16r58, byte 16rcc, byte 16r32, byte 16rc3, byte 16rb0, byte 16r9d, 
		byte 16r75, byte 16r62, byte 16rde, byte 16re2, byte 16r95, byte 16r09, byte 16rb5, byte 16r83, 
		byte 16r93, byte 16r92, byte 16rc9, byte 16rff, byte 16r05, byte 16rf5, byte 16r1f, byte 16r31, 
		byte 16r66, byte 16raa, byte 16rac, byte 16r4a, byte 16rc5, byte 16rf2, byte 16r38, byte 16r03, 
		byte 16r8a, byte 16r30, byte 16r45, byte 16re6, byte 16rf7, byte 16r2e, byte 16r48, byte 16ref, 
		byte 16r0f, byte 16re8, byte 16rbc, byte 16r67, byte 16r5e, byte 16r82, byte 16rc3, byte 16r18, 
		byte 16ra2, byte 16r68, byte 16re4, byte 16r39, byte 16r70, byte 16r27, byte 16r1b, byte 16rf1, 
		byte 16r19, byte 16rb8, byte 16r1b, byte 16rf6, byte 16ra9, byte 16r82, byte 16r74, byte 16r65, 
		byte 16r54, byte 16rf8, byte 16r4e, byte 16r72, byte 16rb9, byte 16rf0, byte 16r02, byte 16r80, 
		byte 16ra3, byte 16r20, byte 16ra0, byte 16r81, byte 16r42, byte 16r92, byte 16r3c, byte 16r23, 
		byte 16rc8, byte 16r83, byte 16r42, byte 16r3f, byte 16rf9, byte 16r49, byte 16r82, byte 16r7f, 
		byte 16r29, byte 16rbb, byte 16rac, byte 16rdc, byte 16r1c, byte 16rcd, byte 16rb0, byte 16r49, 
		byte 16r38, byte 16rce, byte 16r60, byte 16r98, byte 16rc9, byte 16r5b, byte 16ra6, byte 16rb3, 
		byte 16r25, byte 16r28, byte 16rf4, byte 16ref, byte 16r78, byte 16ree, byte 16rd7, byte 16r78, 
		byte 16rb2, byte 16re1, byte 16r22, byte 16rdd, byte 16rfd, byte 16r1c, byte 16rbd, byte 16rd1, 
		byte 16r1d, byte 16r1c, byte 16r0a, byte 16r67, byte 16r83, byte 16re0, byte 16r11, byte 16rfc, 
		byte 16r53, byte 16r6d, byte 16r63, byte 16rd0, byte 16r53, byte 16r26, byte 16r06, byte 16r37, 
	}
),
Xts (
	big 16r000000fe,
	array[] of {
		byte 16r27, byte 16r18, byte 16r28, byte 16r18, byte 16r28, byte 16r45, byte 16r90, byte 16r45, 
		byte 16r23, byte 16r53, byte 16r60, byte 16r28, byte 16r74, byte 16r71, byte 16r35, byte 16r26, 
		byte 16r31, byte 16r41, byte 16r59, byte 16r26, byte 16r53, byte 16r58, byte 16r97, byte 16r93, 
		byte 16r23, byte 16r84, byte 16r62, byte 16r64, byte 16r33, byte 16r83, byte 16r27, byte 16r95, 
	},
	array[] of {
		byte 16rd5, byte 16r5f, byte 16r68, byte 16r4f, byte 16r81, byte 16rf4, byte 16r42, byte 16r6e, 
		byte 16r9f, byte 16rde, byte 16r92, byte 16ra5, byte 16rff, byte 16r02, byte 16rdf, byte 16r2a, 
		byte 16rc8, byte 16r96, byte 16raf, byte 16r63, byte 16r96, byte 16r28, byte 16r88, byte 16ra9, 
		byte 16r79, byte 16r10, byte 16rc1, byte 16r37, byte 16r9e, byte 16r20, byte 16rb0, byte 16ra3, 
		byte 16rb1, byte 16rdb, byte 16r61, byte 16r3f, byte 16rb7, byte 16rfe, byte 16r2e, byte 16r07, 
		byte 16r00, byte 16r43, byte 16r29, byte 16rea, byte 16r5c, byte 16r22, byte 16rbf, byte 16rd3, 
		byte 16r3e, byte 16r3d, byte 16rbe, byte 16r4c, byte 16rf5, byte 16r8c, byte 16rc6, byte 16r08, 
		byte 16rc2, byte 16rc2, byte 16r6c, byte 16r19, byte 16ra2, byte 16re2, byte 16rfe, byte 16r22, 
		byte 16rf9, byte 16r87, byte 16r32, byte 16rc2, byte 16rb5, byte 16rcb, byte 16r84, byte 16r4c, 
		byte 16rc6, byte 16rc0, byte 16r70, byte 16r2d, byte 16r91, byte 16re1, byte 16rd5, byte 16r0f, 
		byte 16rc4, byte 16r38, byte 16r2a, byte 16r7e, byte 16rba, byte 16r56, byte 16r35, byte 16rcd, 
		byte 16r60, byte 16r24, byte 16r32, byte 16ra2, byte 16r30, byte 16r6a, byte 16rc4, byte 16rce, 
		byte 16r82, byte 16rf8, byte 16rd7, byte 16r0c, byte 16r8d, byte 16r9b, byte 16rc1, byte 16r5f, 
		byte 16r91, byte 16r8f, byte 16re7, byte 16r1e, byte 16r74, byte 16rc6, byte 16r22, byte 16rd5, 
		byte 16rcf, byte 16r71, byte 16r17, byte 16r8b, byte 16rf6, byte 16re0, byte 16rb9, byte 16rcc, 
		byte 16r9f, byte 16r2b, byte 16r41, byte 16rdd, byte 16r8d, byte 16rbe, byte 16r44, byte 16r1c, 
		byte 16r41, byte 16rcd, byte 16r0c, byte 16r73, byte 16ra6, byte 16rdc, byte 16r47, byte 16ra3, 
		byte 16r48, byte 16rf6, byte 16r70, byte 16r2f, byte 16r9d, byte 16r0e, byte 16r9b, byte 16r1b, 
		byte 16r14, byte 16r31, byte 16re9, byte 16r48, byte 16re2, byte 16r99, byte 16rb9, byte 16rec, 
		byte 16r22, byte 16r72, byte 16rab, byte 16r2c, byte 16r5f, byte 16r0c, byte 16r7b, byte 16re8, 
		byte 16r6a, byte 16rff, byte 16ra5, byte 16rde, byte 16rc8, byte 16r7a, byte 16r0b, byte 16ree, 
		byte 16r81, byte 16rd3, byte 16rd5, byte 16r00, byte 16r07, byte 16red, byte 16raa, byte 16r2b, 
		byte 16rcf, byte 16rcc, byte 16rb3, byte 16r56, byte 16r05, byte 16r15, byte 16r5f, byte 16rf3, 
		byte 16r6e, byte 16rd8, byte 16red, byte 16rd4, byte 16ra4, byte 16r0d, byte 16rcd, byte 16r4b, 
		byte 16r24, byte 16r3a, byte 16rcd, byte 16r11, byte 16rb2, byte 16rb9, byte 16r87, byte 16rbd, 
		byte 16rbf, byte 16raf, byte 16r91, byte 16ra7, byte 16rca, byte 16rc2, byte 16r7e, byte 16r9c, 
		byte 16r5a, byte 16rea, byte 16r52, byte 16r5e, byte 16re5, byte 16r3d, byte 16re7, byte 16rb2, 
		byte 16rd3, byte 16r33, byte 16r2c, byte 16r86, byte 16r44, byte 16r40, byte 16r2b, byte 16r82, 
		byte 16r3e, byte 16r94, byte 16ra7, byte 16rdb, byte 16r26, byte 16r27, byte 16r6d, byte 16r2d, 
		byte 16r23, byte 16raa, byte 16r07, byte 16r18, byte 16r0f, byte 16r76, byte 16rb4, byte 16rfd, 
		byte 16r29, byte 16rb9, byte 16rc0, byte 16r82, byte 16r30, byte 16r99, byte 16rc9, byte 16rd6, 
		byte 16r2c, byte 16r51, byte 16r98, byte 16r80, byte 16rae, byte 16re7, byte 16re9, byte 16r69, 
		byte 16r76, byte 16r17, byte 16rc1, byte 16r49, byte 16r7d, byte 16r47, byte 16rbf, byte 16r3e, 
		byte 16r57, byte 16r19, byte 16r50, byte 16r31, byte 16r14, byte 16r21, byte 16rb6, byte 16rb7, 
		byte 16r34, byte 16rd3, byte 16r8b, byte 16r0d, byte 16rb9, byte 16r1e, byte 16rb8, byte 16r53, 
		byte 16r31, byte 16rb9, byte 16r1e, byte 16ra9, byte 16rf6, byte 16r15, byte 16r30, byte 16rf5, 
		byte 16r45, byte 16r12, byte 16ra5, byte 16ra5, byte 16r2a, byte 16r4b, byte 16rad, byte 16r58, 
		byte 16r9e, byte 16rb6, byte 16r97, byte 16r81, byte 16rd5, byte 16r37, byte 16rf2, byte 16r32, 
		byte 16r97, byte 16rbb, byte 16r45, byte 16r9b, byte 16rda, byte 16rd2, byte 16r94, byte 16r8a, 
		byte 16r29, byte 16re1, byte 16r55, byte 16r0b, byte 16rf4, byte 16r78, byte 16r7e, byte 16r0b, 
		byte 16re9, byte 16r5b, byte 16rb1, byte 16r73, byte 16rcf, byte 16r5f, byte 16rab, byte 16r17, 
		byte 16rda, byte 16rb7, byte 16ra1, byte 16r3a, byte 16r05, byte 16r2a, byte 16r63, byte 16r45, 
		byte 16r3d, byte 16r97, byte 16rcc, byte 16rec, byte 16r1a, byte 16r32, byte 16r19, byte 16r54, 
		byte 16r88, byte 16r6b, byte 16r7a, byte 16r12, byte 16r99, byte 16rfa, byte 16rae, byte 16rec, 
		byte 16rae, byte 16r35, byte 16rc6, byte 16rea, byte 16rac, byte 16ra7, byte 16r53, byte 16rb0, 
		byte 16r41, byte 16rb5, byte 16re5, byte 16rf0, byte 16r93, byte 16rbf, byte 16r83, byte 16r39, 
		byte 16r7f, byte 16rd2, byte 16r1d, byte 16rd6, byte 16rb3, byte 16r01, byte 16r20, byte 16r66, 
		byte 16rfc, byte 16rc0, byte 16r58, byte 16rcc, byte 16r32, byte 16rc3, byte 16rb0, byte 16r9d, 
		byte 16r75, byte 16r62, byte 16rde, byte 16re2, byte 16r95, byte 16r09, byte 16rb5, byte 16r83, 
		byte 16r93, byte 16r92, byte 16rc9, byte 16rff, byte 16r05, byte 16rf5, byte 16r1f, byte 16r31, 
		byte 16r66, byte 16raa, byte 16rac, byte 16r4a, byte 16rc5, byte 16rf2, byte 16r38, byte 16r03, 
		byte 16r8a, byte 16r30, byte 16r45, byte 16re6, byte 16rf7, byte 16r2e, byte 16r48, byte 16ref, 
		byte 16r0f, byte 16re8, byte 16rbc, byte 16r67, byte 16r5e, byte 16r82, byte 16rc3, byte 16r18, 
		byte 16ra2, byte 16r68, byte 16re4, byte 16r39, byte 16r70, byte 16r27, byte 16r1b, byte 16rf1, 
		byte 16r19, byte 16rb8, byte 16r1b, byte 16rf6, byte 16ra9, byte 16r82, byte 16r74, byte 16r65, 
		byte 16r54, byte 16rf8, byte 16r4e, byte 16r72, byte 16rb9, byte 16rf0, byte 16r02, byte 16r80, 
		byte 16ra3, byte 16r20, byte 16ra0, byte 16r81, byte 16r42, byte 16r92, byte 16r3c, byte 16r23, 
		byte 16rc8, byte 16r83, byte 16r42, byte 16r3f, byte 16rf9, byte 16r49, byte 16r82, byte 16r7f, 
		byte 16r29, byte 16rbb, byte 16rac, byte 16rdc, byte 16r1c, byte 16rcd, byte 16rb0, byte 16r49, 
		byte 16r38, byte 16rce, byte 16r60, byte 16r98, byte 16rc9, byte 16r5b, byte 16ra6, byte 16rb3, 
		byte 16r25, byte 16r28, byte 16rf4, byte 16ref, byte 16r78, byte 16ree, byte 16rd7, byte 16r78, 
		byte 16rb2, byte 16re1, byte 16r22, byte 16rdd, byte 16rfd, byte 16r1c, byte 16rbd, byte 16rd1, 
		byte 16r1d, byte 16r1c, byte 16r0a, byte 16r67, byte 16r83, byte 16re0, byte 16r11, byte 16rfc, 
		byte 16r53, byte 16r6d, byte 16r63, byte 16rd0, byte 16r53, byte 16r26, byte 16r06, byte 16r37, 
	},
	array[] of {
		byte 16r72, byte 16ref, byte 16rc1, byte 16reb, byte 16rfe, byte 16r1e, byte 16re2, byte 16r59, 
		byte 16r75, byte 16ra6, byte 16reb, byte 16r3a, byte 16ra8, byte 16r58, byte 16r9d, byte 16rda, 
		byte 16r2b, byte 16r26, byte 16r1f, byte 16r1c, byte 16r85, byte 16rbd, byte 16rab, byte 16r44, 
		byte 16r2a, byte 16r9e, byte 16r5b, byte 16r2d, byte 16rd1, byte 16rd7, byte 16rc3, byte 16r95, 
		byte 16r7a, byte 16r16, byte 16rfc, byte 16r08, byte 16re5, byte 16r26, byte 16rd4, byte 16rb1, 
		byte 16r22, byte 16r3f, byte 16r1b, byte 16r12, byte 16r32, byte 16ra1, byte 16r1a, byte 16rf2, 
		byte 16r74, byte 16rc3, byte 16rd7, byte 16r0d, byte 16rac, byte 16r57, byte 16rf8, byte 16r3e, 
		byte 16r09, byte 16r83, byte 16rc4, byte 16r98, byte 16rf1, byte 16ra6, byte 16rf1, byte 16rae, 
		byte 16rcb, byte 16r02, byte 16r1c, byte 16r3e, byte 16r70, byte 16r08, byte 16r5a, byte 16r1e, 
		byte 16r52, byte 16r7f, byte 16r1c, byte 16re4, byte 16r1e, byte 16re5, byte 16r91, byte 16r1a, 
		byte 16r82, byte 16r02, byte 16r01, byte 16r61, byte 16r52, byte 16r9c, byte 16rd8, byte 16r27, 
		byte 16r73, byte 16r76, byte 16r2d, byte 16raf, byte 16r54, byte 16r59, byte 16rde, byte 16r94, 
		byte 16ra0, byte 16ra8, byte 16r2a, byte 16rda, byte 16re7, byte 16re1, byte 16r70, byte 16r3c, 
		byte 16r80, byte 16r85, byte 16r43, byte 16rc2, byte 16r9e, byte 16rd6, byte 16rfb, byte 16r32, 
		byte 16rd9, byte 16re0, byte 16r04, byte 16r32, byte 16r7c, byte 16r13, byte 16r55, byte 16r18, 
		byte 16r0c, byte 16r99, byte 16r5a, byte 16r07, byte 16r74, byte 16r14, byte 16r93, byte 16ra0, 
		byte 16r9c, byte 16r21, byte 16rba, byte 16r01, byte 16ra3, byte 16r87, byte 16r88, byte 16r2d, 
		byte 16ra4, byte 16rf6, byte 16r25, byte 16r34, byte 16rb8, byte 16r7b, byte 16rb1, byte 16r5d, 
		byte 16r60, byte 16rd1, byte 16r97, byte 16r20, byte 16r1c, byte 16r0f, byte 16rd3, byte 16rbf, 
		byte 16r30, byte 16rc1, byte 16r50, byte 16r0a, byte 16r3e, byte 16rcf, byte 16rec, byte 16rdd, 
		byte 16r66, byte 16rd8, byte 16r72, byte 16r1f, byte 16r90, byte 16rbc, byte 16rc4, byte 16rc1, 
		byte 16r7e, byte 16re9, byte 16r25, byte 16rc6, byte 16r1b, byte 16r0a, byte 16r03, byte 16r72, 
		byte 16r7a, byte 16r9c, byte 16r0d, byte 16r5f, byte 16r5c, byte 16ra4, byte 16r62, byte 16rfb, 
		byte 16rfa, byte 16r0a, byte 16rf1, byte 16rc2, byte 16r51, byte 16r3a, byte 16r9d, byte 16r9d, 
		byte 16r4b, byte 16r53, byte 16r45, byte 16rbd, byte 16r27, byte 16ra5, byte 16rf6, byte 16re6, 
		byte 16r53, byte 16rf7, byte 16r51, byte 16r69, byte 16r3e, byte 16r6b, byte 16r6a, byte 16r2b, 
		byte 16r8e, byte 16rad, byte 16r57, byte 16rd5, byte 16r11, byte 16re0, byte 16r0e, byte 16r58, 
		byte 16rc4, byte 16r5b, byte 16r7b, byte 16r8d, byte 16r00, byte 16r5a, byte 16rf7, byte 16r92, 
		byte 16r88, byte 16rf5, byte 16rc7, byte 16rc2, byte 16r2f, byte 16rd4, byte 16rf1, byte 16rbf, 
		byte 16r7a, byte 16r89, byte 16r8b, byte 16r03, byte 16ra5, byte 16r63, byte 16r4c, byte 16r6a, 
		byte 16r1a, byte 16re3, byte 16rf9, byte 16rfa, byte 16re5, byte 16rde, byte 16r4f, byte 16r29, 
		byte 16r6a, byte 16r28, byte 16r96, byte 16rb2, byte 16r3e, byte 16r7e, byte 16rd4, byte 16r3e, 
		byte 16rd1, byte 16r4f, byte 16ra5, byte 16ra2, byte 16r80, byte 16r3f, byte 16r4d, byte 16r28, 
		byte 16rf0, byte 16rd3, byte 16rff, byte 16rcf, byte 16r24, byte 16r75, byte 16r76, byte 16r77, 
		byte 16rae, byte 16rbd, byte 16rb4, byte 16r7b, byte 16rb3, byte 16r88, byte 16r37, byte 16r87, 
		byte 16r08, byte 16r94, byte 16r8a, byte 16r8d, byte 16r41, byte 16r26, byte 16red, byte 16r18, 
		byte 16r39, byte 16re0, byte 16rda, byte 16r29, byte 16ra5, byte 16r37, byte 16ra8, byte 16rc1, 
		byte 16r98, byte 16rb3, byte 16rc6, byte 16r6a, byte 16rb0, byte 16r07, byte 16r12, byte 16rdd, 
		byte 16r26, byte 16r16, byte 16r74, byte 16rbf, byte 16r45, byte 16ra7, byte 16r3d, byte 16r67, 
		byte 16rf7, byte 16r69, byte 16r14, byte 16rf8, byte 16r30, byte 16rca, byte 16r01, byte 16r4b, 
		byte 16r65, byte 16r59, byte 16r6f, byte 16r27, byte 16re4, byte 16rcf, byte 16r62, byte 16rde, 
		byte 16r66, byte 16r12, byte 16r5a, byte 16r55, byte 16r66, byte 16rdf, byte 16r99, byte 16r75, 
		byte 16r15, byte 16r56, byte 16r28, byte 16rb4, byte 16r00, byte 16rfb, byte 16rfb, byte 16r3a, 
		byte 16r29, byte 16r04, byte 16r0e, byte 16rd5, byte 16r0f, byte 16raf, byte 16rfd, byte 16rbb, 
		byte 16r18, byte 16rae, byte 16rce, byte 16r7c, byte 16r5c, byte 16r44, byte 16r69, byte 16r32, 
		byte 16r60, byte 16raa, byte 16rb3, byte 16r86, byte 16rc0, byte 16ra3, byte 16r7b, byte 16r11, 
		byte 16rb1, byte 16r14, byte 16rf1, byte 16rc4, byte 16r15, byte 16rae, byte 16rbb, byte 16r65, 
		byte 16r3b, byte 16re4, byte 16r68, byte 16r17, byte 16r94, byte 16r28, byte 16rd4, byte 16r3a, 
		byte 16r4d, byte 16r8b, byte 16rc3, byte 16rec, byte 16r38, byte 16r81, byte 16r3e, byte 16rca, 
		byte 16r30, byte 16ra1, byte 16r3c, byte 16rf1, byte 16rbb, byte 16r18, byte 16rd5, byte 16r24, 
		byte 16rf1, byte 16r99, byte 16r2d, byte 16r44, byte 16rd8, byte 16rb1, byte 16ra4, byte 16r2e, 
		byte 16ra3, byte 16r0b, byte 16r22, byte 16re6, byte 16rc9, byte 16r5b, byte 16r19, byte 16r9d, 
		byte 16r8d, byte 16r18, byte 16r2f, byte 16r88, byte 16r40, byte 16rb0, byte 16r9d, byte 16r05, 
		byte 16r95, byte 16r85, byte 16rc3, byte 16r1a, byte 16rd6, byte 16r91, byte 16rfa, byte 16r06, 
		byte 16r19, byte 16rff, byte 16r03, byte 16r8a, byte 16rca, byte 16r2c, byte 16r39, byte 16ra9, 
		byte 16r43, byte 16r42, byte 16r11, byte 16r57, byte 16r36, byte 16r17, byte 16r17, byte 16rc4, 
		byte 16r9d, byte 16r32, byte 16r20, byte 16r28, byte 16ra7, byte 16r46, byte 16r48, byte 16r11, 
		byte 16r3b, byte 16rd8, byte 16rc9, byte 16rd7, byte 16rec, byte 16r77, byte 16rcf, byte 16r3c, 
		byte 16r89, byte 16rc1, byte 16rec, byte 16r87, byte 16r18, byte 16rce, byte 16rff, byte 16r85, 
		byte 16r16, byte 16rd9, byte 16r6b, byte 16r34, byte 16rc3, byte 16rc6, byte 16r14, byte 16rf1, 
		byte 16r06, byte 16r99, byte 16rc9, byte 16rab, byte 16rc4, byte 16red, byte 16r04, byte 16r11, 
		byte 16r50, byte 16r62, byte 16r23, byte 16rbe, byte 16ra1, byte 16r6a, byte 16rf3, byte 16r5c, 
		byte 16r88, byte 16r3a, byte 16rcc, byte 16rdb, byte 16re1, byte 16r10, byte 16r4e, byte 16ref, 
		byte 16r0c, byte 16rfd, byte 16rb5, byte 16r4e, byte 16r12, byte 16rfb, byte 16r23, byte 16r0a, 
	}
),
Xts (
	big 16r000000ff,
	array[] of {
		byte 16r27, byte 16r18, byte 16r28, byte 16r18, byte 16r28, byte 16r45, byte 16r90, byte 16r45, 
		byte 16r23, byte 16r53, byte 16r60, byte 16r28, byte 16r74, byte 16r71, byte 16r35, byte 16r26, 
		byte 16r31, byte 16r41, byte 16r59, byte 16r26, byte 16r53, byte 16r58, byte 16r97, byte 16r93, 
		byte 16r23, byte 16r84, byte 16r62, byte 16r64, byte 16r33, byte 16r83, byte 16r27, byte 16r95, 
	},
	array[] of {
		byte 16r72, byte 16ref, byte 16rc1, byte 16reb, byte 16rfe, byte 16r1e, byte 16re2, byte 16r59, 
		byte 16r75, byte 16ra6, byte 16reb, byte 16r3a, byte 16ra8, byte 16r58, byte 16r9d, byte 16rda, 
		byte 16r2b, byte 16r26, byte 16r1f, byte 16r1c, byte 16r85, byte 16rbd, byte 16rab, byte 16r44, 
		byte 16r2a, byte 16r9e, byte 16r5b, byte 16r2d, byte 16rd1, byte 16rd7, byte 16rc3, byte 16r95, 
		byte 16r7a, byte 16r16, byte 16rfc, byte 16r08, byte 16re5, byte 16r26, byte 16rd4, byte 16rb1, 
		byte 16r22, byte 16r3f, byte 16r1b, byte 16r12, byte 16r32, byte 16ra1, byte 16r1a, byte 16rf2, 
		byte 16r74, byte 16rc3, byte 16rd7, byte 16r0d, byte 16rac, byte 16r57, byte 16rf8, byte 16r3e, 
		byte 16r09, byte 16r83, byte 16rc4, byte 16r98, byte 16rf1, byte 16ra6, byte 16rf1, byte 16rae, 
		byte 16rcb, byte 16r02, byte 16r1c, byte 16r3e, byte 16r70, byte 16r08, byte 16r5a, byte 16r1e, 
		byte 16r52, byte 16r7f, byte 16r1c, byte 16re4, byte 16r1e, byte 16re5, byte 16r91, byte 16r1a, 
		byte 16r82, byte 16r02, byte 16r01, byte 16r61, byte 16r52, byte 16r9c, byte 16rd8, byte 16r27, 
		byte 16r73, byte 16r76, byte 16r2d, byte 16raf, byte 16r54, byte 16r59, byte 16rde, byte 16r94, 
		byte 16ra0, byte 16ra8, byte 16r2a, byte 16rda, byte 16re7, byte 16re1, byte 16r70, byte 16r3c, 
		byte 16r80, byte 16r85, byte 16r43, byte 16rc2, byte 16r9e, byte 16rd6, byte 16rfb, byte 16r32, 
		byte 16rd9, byte 16re0, byte 16r04, byte 16r32, byte 16r7c, byte 16r13, byte 16r55, byte 16r18, 
		byte 16r0c, byte 16r99, byte 16r5a, byte 16r07, byte 16r74, byte 16r14, byte 16r93, byte 16ra0, 
		byte 16r9c, byte 16r21, byte 16rba, byte 16r01, byte 16ra3, byte 16r87, byte 16r88, byte 16r2d, 
		byte 16ra4, byte 16rf6, byte 16r25, byte 16r34, byte 16rb8, byte 16r7b, byte 16rb1, byte 16r5d, 
		byte 16r60, byte 16rd1, byte 16r97, byte 16r20, byte 16r1c, byte 16r0f, byte 16rd3, byte 16rbf, 
		byte 16r30, byte 16rc1, byte 16r50, byte 16r0a, byte 16r3e, byte 16rcf, byte 16rec, byte 16rdd, 
		byte 16r66, byte 16rd8, byte 16r72, byte 16r1f, byte 16r90, byte 16rbc, byte 16rc4, byte 16rc1, 
		byte 16r7e, byte 16re9, byte 16r25, byte 16rc6, byte 16r1b, byte 16r0a, byte 16r03, byte 16r72, 
		byte 16r7a, byte 16r9c, byte 16r0d, byte 16r5f, byte 16r5c, byte 16ra4, byte 16r62, byte 16rfb, 
		byte 16rfa, byte 16r0a, byte 16rf1, byte 16rc2, byte 16r51, byte 16r3a, byte 16r9d, byte 16r9d, 
		byte 16r4b, byte 16r53, byte 16r45, byte 16rbd, byte 16r27, byte 16ra5, byte 16rf6, byte 16re6, 
		byte 16r53, byte 16rf7, byte 16r51, byte 16r69, byte 16r3e, byte 16r6b, byte 16r6a, byte 16r2b, 
		byte 16r8e, byte 16rad, byte 16r57, byte 16rd5, byte 16r11, byte 16re0, byte 16r0e, byte 16r58, 
		byte 16rc4, byte 16r5b, byte 16r7b, byte 16r8d, byte 16r00, byte 16r5a, byte 16rf7, byte 16r92, 
		byte 16r88, byte 16rf5, byte 16rc7, byte 16rc2, byte 16r2f, byte 16rd4, byte 16rf1, byte 16rbf, 
		byte 16r7a, byte 16r89, byte 16r8b, byte 16r03, byte 16ra5, byte 16r63, byte 16r4c, byte 16r6a, 
		byte 16r1a, byte 16re3, byte 16rf9, byte 16rfa, byte 16re5, byte 16rde, byte 16r4f, byte 16r29, 
		byte 16r6a, byte 16r28, byte 16r96, byte 16rb2, byte 16r3e, byte 16r7e, byte 16rd4, byte 16r3e, 
		byte 16rd1, byte 16r4f, byte 16ra5, byte 16ra2, byte 16r80, byte 16r3f, byte 16r4d, byte 16r28, 
		byte 16rf0, byte 16rd3, byte 16rff, byte 16rcf, byte 16r24, byte 16r75, byte 16r76, byte 16r77, 
		byte 16rae, byte 16rbd, byte 16rb4, byte 16r7b, byte 16rb3, byte 16r88, byte 16r37, byte 16r87, 
		byte 16r08, byte 16r94, byte 16r8a, byte 16r8d, byte 16r41, byte 16r26, byte 16red, byte 16r18, 
		byte 16r39, byte 16re0, byte 16rda, byte 16r29, byte 16ra5, byte 16r37, byte 16ra8, byte 16rc1, 
		byte 16r98, byte 16rb3, byte 16rc6, byte 16r6a, byte 16rb0, byte 16r07, byte 16r12, byte 16rdd, 
		byte 16r26, byte 16r16, byte 16r74, byte 16rbf, byte 16r45, byte 16ra7, byte 16r3d, byte 16r67, 
		byte 16rf7, byte 16r69, byte 16r14, byte 16rf8, byte 16r30, byte 16rca, byte 16r01, byte 16r4b, 
		byte 16r65, byte 16r59, byte 16r6f, byte 16r27, byte 16re4, byte 16rcf, byte 16r62, byte 16rde, 
		byte 16r66, byte 16r12, byte 16r5a, byte 16r55, byte 16r66, byte 16rdf, byte 16r99, byte 16r75, 
		byte 16r15, byte 16r56, byte 16r28, byte 16rb4, byte 16r00, byte 16rfb, byte 16rfb, byte 16r3a, 
		byte 16r29, byte 16r04, byte 16r0e, byte 16rd5, byte 16r0f, byte 16raf, byte 16rfd, byte 16rbb, 
		byte 16r18, byte 16rae, byte 16rce, byte 16r7c, byte 16r5c, byte 16r44, byte 16r69, byte 16r32, 
		byte 16r60, byte 16raa, byte 16rb3, byte 16r86, byte 16rc0, byte 16ra3, byte 16r7b, byte 16r11, 
		byte 16rb1, byte 16r14, byte 16rf1, byte 16rc4, byte 16r15, byte 16rae, byte 16rbb, byte 16r65, 
		byte 16r3b, byte 16re4, byte 16r68, byte 16r17, byte 16r94, byte 16r28, byte 16rd4, byte 16r3a, 
		byte 16r4d, byte 16r8b, byte 16rc3, byte 16rec, byte 16r38, byte 16r81, byte 16r3e, byte 16rca, 
		byte 16r30, byte 16ra1, byte 16r3c, byte 16rf1, byte 16rbb, byte 16r18, byte 16rd5, byte 16r24, 
		byte 16rf1, byte 16r99, byte 16r2d, byte 16r44, byte 16rd8, byte 16rb1, byte 16ra4, byte 16r2e, 
		byte 16ra3, byte 16r0b, byte 16r22, byte 16re6, byte 16rc9, byte 16r5b, byte 16r19, byte 16r9d, 
		byte 16r8d, byte 16r18, byte 16r2f, byte 16r88, byte 16r40, byte 16rb0, byte 16r9d, byte 16r05, 
		byte 16r95, byte 16r85, byte 16rc3, byte 16r1a, byte 16rd6, byte 16r91, byte 16rfa, byte 16r06, 
		byte 16r19, byte 16rff, byte 16r03, byte 16r8a, byte 16rca, byte 16r2c, byte 16r39, byte 16ra9, 
		byte 16r43, byte 16r42, byte 16r11, byte 16r57, byte 16r36, byte 16r17, byte 16r17, byte 16rc4, 
		byte 16r9d, byte 16r32, byte 16r20, byte 16r28, byte 16ra7, byte 16r46, byte 16r48, byte 16r11, 
		byte 16r3b, byte 16rd8, byte 16rc9, byte 16rd7, byte 16rec, byte 16r77, byte 16rcf, byte 16r3c, 
		byte 16r89, byte 16rc1, byte 16rec, byte 16r87, byte 16r18, byte 16rce, byte 16rff, byte 16r85, 
		byte 16r16, byte 16rd9, byte 16r6b, byte 16r34, byte 16rc3, byte 16rc6, byte 16r14, byte 16rf1, 
		byte 16r06, byte 16r99, byte 16rc9, byte 16rab, byte 16rc4, byte 16red, byte 16r04, byte 16r11, 
		byte 16r50, byte 16r62, byte 16r23, byte 16rbe, byte 16ra1, byte 16r6a, byte 16rf3, byte 16r5c, 
		byte 16r88, byte 16r3a, byte 16rcc, byte 16rdb, byte 16re1, byte 16r10, byte 16r4e, byte 16ref, 
		byte 16r0c, byte 16rfd, byte 16rb5, byte 16r4e, byte 16r12, byte 16rfb, byte 16r23, byte 16r0a, 
	},
	array[] of {
		byte 16r32, byte 16r60, byte 16rae, byte 16r8d, byte 16rad, byte 16r1f, byte 16r4a, byte 16r32, 
		byte 16rc5, byte 16rca, byte 16rfe, byte 16r3a, byte 16rb0, byte 16reb, byte 16r95, byte 16r54, 
		byte 16r9d, byte 16r46, byte 16r1a, byte 16r67, byte 16rce, byte 16rb9, byte 16re5, byte 16raa, 
		byte 16r2d, byte 16r3a, byte 16rfb, byte 16r62, byte 16rde, byte 16rce, byte 16r05, byte 16r53, 
		byte 16r19, byte 16r3b, byte 16ra5, byte 16r0c, byte 16r75, byte 16rbe, byte 16r25, byte 16r1e, 
		byte 16r08, byte 16rd1, byte 16rd0, byte 16r8f, byte 16r10, byte 16r88, byte 16r57, byte 16r6c, 
		byte 16r7e, byte 16rfd, byte 16rfa, byte 16raf, byte 16r3f, byte 16r45, byte 16r95, byte 16r59, 
		byte 16r57, byte 16r1e, byte 16r12, byte 16r51, byte 16r17, byte 16r53, byte 16rb0, byte 16r7a, 
		byte 16rf0, byte 16r73, byte 16rf3, byte 16r5d, byte 16ra0, byte 16r6a, byte 16rf0, byte 16rce, 
		byte 16r0b, byte 16rbf, byte 16r6b, byte 16r8f, byte 16r5c, byte 16rcc, byte 16r5c, byte 16rea, 
		byte 16r50, byte 16r0e, byte 16rc1, byte 16rb2, byte 16r11, byte 16rbd, byte 16r51, byte 16rf6, 
		byte 16r3b, byte 16r60, byte 16r6b, byte 16rf6, byte 16r52, byte 16r87, byte 16r96, byte 16rca, 
		byte 16r12, byte 16r17, byte 16r3b, byte 16ra3, byte 16r9b, byte 16r89, byte 16r35, byte 16ree, 
		byte 16r44, byte 16rcc, byte 16rce, byte 16r64, byte 16r6f, byte 16r90, byte 16ra4, byte 16r5b, 
		byte 16rf9, byte 16rcc, byte 16rc5, byte 16r67, byte 16rf0, byte 16rac, byte 16re1, byte 16r3d, 
		byte 16rc2, byte 16rd5, byte 16r3e, byte 16rbe, byte 16red, byte 16rc8, byte 16r1f, byte 16r58, 
		byte 16rb2, byte 16re4, byte 16r11, byte 16r79, byte 16rdd, byte 16rdf, byte 16r0d, byte 16r5a, 
		byte 16r5c, byte 16r42, byte 16rf5, byte 16rd8, byte 16r50, byte 16r6c, byte 16r1a, byte 16r5d, 
		byte 16r2f, byte 16r8f, byte 16r59, byte 16rf3, byte 16rea, byte 16r87, byte 16r3c, byte 16rbc, 
		byte 16rd0, byte 16ree, byte 16rc1, byte 16r9a, byte 16rcb, byte 16rf3, byte 16r25, byte 16r42, 
		byte 16r3b, byte 16rd3, byte 16rdc, byte 16rb8, byte 16rc2, byte 16rb1, byte 16rbf, byte 16r1d, 
		byte 16r1e, byte 16rae, byte 16rd0, byte 16reb, byte 16ra7, byte 16rf0, byte 16r69, byte 16r8e, 
		byte 16r43, byte 16r14, byte 16rfb, byte 16reb, byte 16r2f, byte 16r15, byte 16r66, byte 16rd1, 
		byte 16rb9, byte 16r25, byte 16r30, byte 16r08, byte 16rcb, byte 16rcc, byte 16rf4, byte 16r5a, 
		byte 16r2b, byte 16r0d, byte 16r9c, byte 16r5c, byte 16r9c, byte 16r21, byte 16r47, byte 16r4f, 
		byte 16r40, byte 16r76, byte 16re0, byte 16r2b, byte 16re2, byte 16r60, byte 16r50, byte 16rb9, 
		byte 16r9d, byte 16ree, byte 16r4f, byte 16rd6, byte 16r8a, byte 16r4c, byte 16rf8, byte 16r90, 
		byte 16re4, byte 16r96, byte 16re4, byte 16rfc, byte 16rae, byte 16r7b, byte 16r70, byte 16rf9, 
		byte 16r4e, byte 16ra5, byte 16ra9, byte 16r06, byte 16r2d, byte 16ra0, byte 16rda, byte 16reb, 
		byte 16ra1, byte 16r99, byte 16r3d, byte 16r2c, byte 16rcd, byte 16r1d, byte 16rd3, byte 16rc2, 
		byte 16r44, byte 16rb8, byte 16r42, byte 16r88, byte 16r01, byte 16r49, byte 16r5a, byte 16r58, 
		byte 16rb2, byte 16r16, byte 16r54, byte 16r7e, byte 16r7e, byte 16r84, byte 16r7c, byte 16r46, 
		byte 16rd1, byte 16rd7, byte 16r56, byte 16r37, byte 16r7b, byte 16r62, byte 16r42, byte 16rd2, 
		byte 16re5, byte 16rfb, byte 16r83, byte 16rbf, byte 16r75, byte 16r2b, byte 16r54, byte 16re0, 
		byte 16rdf, byte 16r71, byte 16re8, byte 16r89, byte 16rf3, byte 16ra2, byte 16rbb, byte 16r0f, 
		byte 16r4c, byte 16r10, byte 16r80, byte 16r5b, byte 16rf3, byte 16rc5, byte 16r90, byte 16r37, 
		byte 16r6e, byte 16r3c, byte 16r24, byte 16re2, byte 16r2f, byte 16rf5, byte 16r7f, byte 16r7f, 
		byte 16ra9, byte 16r65, byte 16r57, byte 16r73, byte 16r75, byte 16r32, byte 16r5c, byte 16rea, 
		byte 16r5d, byte 16r92, byte 16r0d, byte 16rb9, byte 16r4b, byte 16r9c, byte 16r33, byte 16r6b, 
		byte 16r45, byte 16r5f, byte 16r6e, byte 16r89, byte 16r4c, byte 16r01, byte 16r86, byte 16r6f, 
		byte 16re9, byte 16rfb, byte 16rb8, byte 16rc8, byte 16rd3, byte 16rf7, byte 16r0a, byte 16r29, 
		byte 16r57, byte 16r28, byte 16r5f, byte 16r6d, byte 16rfb, byte 16r5d, byte 16rcd, byte 16r8c, 
		byte 16rbf, byte 16r54, byte 16r78, byte 16r2f, byte 16r8f, byte 16re7, byte 16r76, byte 16r6d, 
		byte 16r47, byte 16r23, byte 16r81, byte 16r99, byte 16r13, byte 16rac, byte 16r77, byte 16r34, 
		byte 16r21, byte 16re3, byte 16ra3, byte 16r10, byte 16r95, byte 16r86, byte 16r6b, byte 16rad, 
		byte 16r22, byte 16rc8, byte 16r6a, byte 16r60, byte 16r36, byte 16rb2, byte 16r51, byte 16r8b, 
		byte 16r20, byte 16r59, byte 16rb4, byte 16r22, byte 16r9d, byte 16r18, byte 16rc8, byte 16rc2, 
		byte 16rcc, byte 16rbd, byte 16rf9, byte 16r06, byte 16rc6, byte 16rcc, byte 16r6e, byte 16r82, 
		byte 16r46, byte 16r4e, byte 16re5, byte 16r7b, byte 16rdd, byte 16rb0, byte 16rbe, byte 16rbc, 
		byte 16rb1, byte 16rdc, byte 16r64, byte 16r53, byte 16r25, byte 16rbf, byte 16rb3, byte 16re6, 
		byte 16r65, byte 16ref, byte 16r72, byte 16r51, byte 16r08, byte 16r2c, byte 16r88, byte 16reb, 
		byte 16rb1, byte 16rcf, byte 16r20, byte 16r3b, byte 16rd7, byte 16r79, byte 16rfd, byte 16rd3, 
		byte 16r86, byte 16r75, byte 16r71, byte 16r3c, byte 16r8d, byte 16raa, byte 16rdd, byte 16r17, 
		byte 16re1, byte 16rca, byte 16rbe, byte 16re4, byte 16r32, byte 16rb0, byte 16r97, byte 16r87, 
		byte 16rb6, byte 16rdd, byte 16rf3, byte 16r30, byte 16r4e, byte 16r38, byte 16rb7, byte 16r31, 
		byte 16rb4, byte 16r5d, byte 16rf5, byte 16rdf, byte 16r51, byte 16rb7, byte 16r8f, byte 16rcf, 
		byte 16rb3, byte 16rd3, byte 16r24, byte 16r66, byte 16r02, byte 16r8d, byte 16r0b, byte 16ra3, 
		byte 16r65, byte 16r55, byte 16re7, byte 16re1, byte 16r1a, byte 16rb0, byte 16ree, byte 16r06, 
		byte 16r66, byte 16r06, byte 16r1d, byte 16r16, byte 16r45, byte 16rd9, byte 16r62, byte 16r44, 
		byte 16r4b, byte 16rc4, byte 16r7a, byte 16r38, byte 16r18, byte 16r89, byte 16r30, byte 16ra8, 
		byte 16r4b, byte 16r4d, byte 16r56, byte 16r13, byte 16r95, byte 16rc7, byte 16r3c, byte 16r08, 
		byte 16r70, byte 16r21, byte 16r92, byte 16r7c, byte 16ra6, byte 16r38, byte 16rb7, byte 16raf, 
		byte 16rc8, byte 16ra8, byte 16r67, byte 16r9c, byte 16rcb, byte 16r84, byte 16rc2, byte 16r65, 
		byte 16r55, byte 16r44, byte 16r0e, byte 16rc7, byte 16rf1, byte 16r04, byte 16r45, byte 16rcd, 
	}
),
Xts (
	big 16r000000ff,
	array[] of {
		byte 16r27, byte 16r18, byte 16r28, byte 16r18, byte 16r28, byte 16r45, byte 16r90, byte 16r45, 
		byte 16r23, byte 16r53, byte 16r60, byte 16r28, byte 16r74, byte 16r71, byte 16r35, byte 16r26, 
		byte 16r62, byte 16r49, byte 16r77, byte 16r57, byte 16r24, byte 16r70, byte 16r93, byte 16r69, 
		byte 16r99, byte 16r59, byte 16r57, byte 16r49, byte 16r66, byte 16r96, byte 16r76, byte 16r27, 
		byte 16r31, byte 16r41, byte 16r59, byte 16r26, byte 16r53, byte 16r58, byte 16r97, byte 16r93, 
		byte 16r23, byte 16r84, byte 16r62, byte 16r64, byte 16r33, byte 16r83, byte 16r27, byte 16r95, 
		byte 16r02, byte 16r88, byte 16r41, byte 16r97, byte 16r16, byte 16r93, byte 16r99, byte 16r37, 
		byte 16r51, byte 16r05, byte 16r82, byte 16r09, byte 16r74, byte 16r94, byte 16r45, byte 16r92, 
	},
	array[] of {
		byte 16r00, byte 16r01, byte 16r02, byte 16r03, byte 16r04, byte 16r05, byte 16r06, byte 16r07, 
		byte 16r08, byte 16r09, byte 16r0a, byte 16r0b, byte 16r0c, byte 16r0d, byte 16r0e, byte 16r0f, 
		byte 16r10, byte 16r11, byte 16r12, byte 16r13, byte 16r14, byte 16r15, byte 16r16, byte 16r17, 
		byte 16r18, byte 16r19, byte 16r1a, byte 16r1b, byte 16r1c, byte 16r1d, byte 16r1e, byte 16r1f, 
		byte 16r20, byte 16r21, byte 16r22, byte 16r23, byte 16r24, byte 16r25, byte 16r26, byte 16r27, 
		byte 16r28, byte 16r29, byte 16r2a, byte 16r2b, byte 16r2c, byte 16r2d, byte 16r2e, byte 16r2f, 
		byte 16r30, byte 16r31, byte 16r32, byte 16r33, byte 16r34, byte 16r35, byte 16r36, byte 16r37, 
		byte 16r38, byte 16r39, byte 16r3a, byte 16r3b, byte 16r3c, byte 16r3d, byte 16r3e, byte 16r3f, 
		byte 16r40, byte 16r41, byte 16r42, byte 16r43, byte 16r44, byte 16r45, byte 16r46, byte 16r47, 
		byte 16r48, byte 16r49, byte 16r4a, byte 16r4b, byte 16r4c, byte 16r4d, byte 16r4e, byte 16r4f, 
		byte 16r50, byte 16r51, byte 16r52, byte 16r53, byte 16r54, byte 16r55, byte 16r56, byte 16r57, 
		byte 16r58, byte 16r59, byte 16r5a, byte 16r5b, byte 16r5c, byte 16r5d, byte 16r5e, byte 16r5f, 
		byte 16r60, byte 16r61, byte 16r62, byte 16r63, byte 16r64, byte 16r65, byte 16r66, byte 16r67, 
		byte 16r68, byte 16r69, byte 16r6a, byte 16r6b, byte 16r6c, byte 16r6d, byte 16r6e, byte 16r6f, 
		byte 16r70, byte 16r71, byte 16r72, byte 16r73, byte 16r74, byte 16r75, byte 16r76, byte 16r77, 
		byte 16r78, byte 16r79, byte 16r7a, byte 16r7b, byte 16r7c, byte 16r7d, byte 16r7e, byte 16r7f, 
		byte 16r80, byte 16r81, byte 16r82, byte 16r83, byte 16r84, byte 16r85, byte 16r86, byte 16r87, 
		byte 16r88, byte 16r89, byte 16r8a, byte 16r8b, byte 16r8c, byte 16r8d, byte 16r8e, byte 16r8f, 
		byte 16r90, byte 16r91, byte 16r92, byte 16r93, byte 16r94, byte 16r95, byte 16r96, byte 16r97, 
		byte 16r98, byte 16r99, byte 16r9a, byte 16r9b, byte 16r9c, byte 16r9d, byte 16r9e, byte 16r9f, 
		byte 16ra0, byte 16ra1, byte 16ra2, byte 16ra3, byte 16ra4, byte 16ra5, byte 16ra6, byte 16ra7, 
		byte 16ra8, byte 16ra9, byte 16raa, byte 16rab, byte 16rac, byte 16rad, byte 16rae, byte 16raf, 
		byte 16rb0, byte 16rb1, byte 16rb2, byte 16rb3, byte 16rb4, byte 16rb5, byte 16rb6, byte 16rb7, 
		byte 16rb8, byte 16rb9, byte 16rba, byte 16rbb, byte 16rbc, byte 16rbd, byte 16rbe, byte 16rbf, 
		byte 16rc0, byte 16rc1, byte 16rc2, byte 16rc3, byte 16rc4, byte 16rc5, byte 16rc6, byte 16rc7, 
		byte 16rc8, byte 16rc9, byte 16rca, byte 16rcb, byte 16rcc, byte 16rcd, byte 16rce, byte 16rcf, 
		byte 16rd0, byte 16rd1, byte 16rd2, byte 16rd3, byte 16rd4, byte 16rd5, byte 16rd6, byte 16rd7, 
		byte 16rd8, byte 16rd9, byte 16rda, byte 16rdb, byte 16rdc, byte 16rdd, byte 16rde, byte 16rdf, 
		byte 16re0, byte 16re1, byte 16re2, byte 16re3, byte 16re4, byte 16re5, byte 16re6, byte 16re7, 
		byte 16re8, byte 16re9, byte 16rea, byte 16reb, byte 16rec, byte 16red, byte 16ree, byte 16ref, 
		byte 16rf0, byte 16rf1, byte 16rf2, byte 16rf3, byte 16rf4, byte 16rf5, byte 16rf6, byte 16rf7, 
		byte 16rf8, byte 16rf9, byte 16rfa, byte 16rfb, byte 16rfc, byte 16rfd, byte 16rfe, byte 16rff, 
		byte 16r00, byte 16r01, byte 16r02, byte 16r03, byte 16r04, byte 16r05, byte 16r06, byte 16r07, 
		byte 16r08, byte 16r09, byte 16r0a, byte 16r0b, byte 16r0c, byte 16r0d, byte 16r0e, byte 16r0f, 
		byte 16r10, byte 16r11, byte 16r12, byte 16r13, byte 16r14, byte 16r15, byte 16r16, byte 16r17, 
		byte 16r18, byte 16r19, byte 16r1a, byte 16r1b, byte 16r1c, byte 16r1d, byte 16r1e, byte 16r1f, 
		byte 16r20, byte 16r21, byte 16r22, byte 16r23, byte 16r24, byte 16r25, byte 16r26, byte 16r27, 
		byte 16r28, byte 16r29, byte 16r2a, byte 16r2b, byte 16r2c, byte 16r2d, byte 16r2e, byte 16r2f, 
		byte 16r30, byte 16r31, byte 16r32, byte 16r33, byte 16r34, byte 16r35, byte 16r36, byte 16r37, 
		byte 16r38, byte 16r39, byte 16r3a, byte 16r3b, byte 16r3c, byte 16r3d, byte 16r3e, byte 16r3f, 
		byte 16r40, byte 16r41, byte 16r42, byte 16r43, byte 16r44, byte 16r45, byte 16r46, byte 16r47, 
		byte 16r48, byte 16r49, byte 16r4a, byte 16r4b, byte 16r4c, byte 16r4d, byte 16r4e, byte 16r4f, 
		byte 16r50, byte 16r51, byte 16r52, byte 16r53, byte 16r54, byte 16r55, byte 16r56, byte 16r57, 
		byte 16r58, byte 16r59, byte 16r5a, byte 16r5b, byte 16r5c, byte 16r5d, byte 16r5e, byte 16r5f, 
		byte 16r60, byte 16r61, byte 16r62, byte 16r63, byte 16r64, byte 16r65, byte 16r66, byte 16r67, 
		byte 16r68, byte 16r69, byte 16r6a, byte 16r6b, byte 16r6c, byte 16r6d, byte 16r6e, byte 16r6f, 
		byte 16r70, byte 16r71, byte 16r72, byte 16r73, byte 16r74, byte 16r75, byte 16r76, byte 16r77, 
		byte 16r78, byte 16r79, byte 16r7a, byte 16r7b, byte 16r7c, byte 16r7d, byte 16r7e, byte 16r7f, 
		byte 16r80, byte 16r81, byte 16r82, byte 16r83, byte 16r84, byte 16r85, byte 16r86, byte 16r87, 
		byte 16r88, byte 16r89, byte 16r8a, byte 16r8b, byte 16r8c, byte 16r8d, byte 16r8e, byte 16r8f, 
		byte 16r90, byte 16r91, byte 16r92, byte 16r93, byte 16r94, byte 16r95, byte 16r96, byte 16r97, 
		byte 16r98, byte 16r99, byte 16r9a, byte 16r9b, byte 16r9c, byte 16r9d, byte 16r9e, byte 16r9f, 
		byte 16ra0, byte 16ra1, byte 16ra2, byte 16ra3, byte 16ra4, byte 16ra5, byte 16ra6, byte 16ra7, 
		byte 16ra8, byte 16ra9, byte 16raa, byte 16rab, byte 16rac, byte 16rad, byte 16rae, byte 16raf, 
		byte 16rb0, byte 16rb1, byte 16rb2, byte 16rb3, byte 16rb4, byte 16rb5, byte 16rb6, byte 16rb7, 
		byte 16rb8, byte 16rb9, byte 16rba, byte 16rbb, byte 16rbc, byte 16rbd, byte 16rbe, byte 16rbf, 
		byte 16rc0, byte 16rc1, byte 16rc2, byte 16rc3, byte 16rc4, byte 16rc5, byte 16rc6, byte 16rc7, 
		byte 16rc8, byte 16rc9, byte 16rca, byte 16rcb, byte 16rcc, byte 16rcd, byte 16rce, byte 16rcf, 
		byte 16rd0, byte 16rd1, byte 16rd2, byte 16rd3, byte 16rd4, byte 16rd5, byte 16rd6, byte 16rd7, 
		byte 16rd8, byte 16rd9, byte 16rda, byte 16rdb, byte 16rdc, byte 16rdd, byte 16rde, byte 16rdf, 
		byte 16re0, byte 16re1, byte 16re2, byte 16re3, byte 16re4, byte 16re5, byte 16re6, byte 16re7, 
		byte 16re8, byte 16re9, byte 16rea, byte 16reb, byte 16rec, byte 16red, byte 16ree, byte 16ref, 
		byte 16rf0, byte 16rf1, byte 16rf2, byte 16rf3, byte 16rf4, byte 16rf5, byte 16rf6, byte 16rf7, 
		byte 16rf8, byte 16rf9, byte 16rfa, byte 16rfb, byte 16rfc, byte 16rfd, byte 16rfe, byte 16rff, 
	},
	array[] of {
		byte 16r1c, byte 16r3b, byte 16r3a, byte 16r10, byte 16r2f, byte 16r77, byte 16r03, byte 16r86, 
		byte 16re4, byte 16r83, byte 16r6c, byte 16r99, byte 16re3, byte 16r70, byte 16rcf, byte 16r9b, 
		byte 16rea, byte 16r00, byte 16r80, byte 16r3f, byte 16r5e, byte 16r48, byte 16r23, byte 16r57, 
		byte 16ra4, byte 16rae, byte 16r12, byte 16rd4, byte 16r14, byte 16ra3, byte 16re6, byte 16r3b, 
		byte 16r5d, byte 16r31, byte 16re2, byte 16r76, byte 16rf8, byte 16rfe, byte 16r4a, byte 16r8d, 
		byte 16r66, byte 16rb3, byte 16r17, byte 16rf9, byte 16rac, byte 16r68, byte 16r3f, byte 16r44, 
		byte 16r68, byte 16r0a, byte 16r86, byte 16rac, byte 16r35, byte 16rad, byte 16rfc, byte 16r33, 
		byte 16r45, byte 16rbe, byte 16rfe, byte 16rcb, byte 16r4b, byte 16rb1, byte 16r88, byte 16rfd, 
		byte 16r57, byte 16r76, byte 16r92, byte 16r6c, byte 16r49, byte 16ra3, byte 16r09, byte 16r5e, 
		byte 16rb1, byte 16r08, byte 16rfd, byte 16r10, byte 16r98, byte 16rba, byte 16rec, byte 16r70, 
		byte 16raa, byte 16ra6, byte 16r69, byte 16r99, byte 16ra7, byte 16r2a, byte 16r82, byte 16rf2, 
		byte 16r7d, byte 16r84, byte 16r8b, byte 16r21, byte 16rd4, byte 16ra7, byte 16r41, byte 16rb0, 
		byte 16rc5, byte 16rcd, byte 16r4d, byte 16r5f, byte 16rff, byte 16r9d, byte 16rac, byte 16r89, 
		byte 16rae, byte 16rba, byte 16r12, byte 16r29, byte 16r61, byte 16rd0, byte 16r3a, byte 16r75, 
		byte 16r71, byte 16r23, byte 16re9, byte 16r87, byte 16r0f, byte 16r8a, byte 16rcf, byte 16r10, 
		byte 16r00, byte 16r02, byte 16r08, byte 16r87, byte 16r89, byte 16r14, byte 16r29, byte 16rca, 
		byte 16r2a, byte 16r3e, byte 16r7a, byte 16r7d, byte 16r7d, byte 16rf7, byte 16rb1, byte 16r03, 
		byte 16r55, byte 16r16, byte 16r5c, byte 16r8b, byte 16r9a, byte 16r6d, byte 16r0a, byte 16r7d, 
		byte 16re8, byte 16rb0, byte 16r62, byte 16rc4, byte 16r50, byte 16r0d, byte 16rc4, byte 16rcd, 
		byte 16r12, byte 16r0c, byte 16r0f, byte 16r74, byte 16r18, byte 16rda, byte 16re3, byte 16rd0, 
		byte 16rb5, byte 16r78, byte 16r1c, byte 16r34, byte 16r80, byte 16r3f, byte 16ra7, byte 16r54, 
		byte 16r21, byte 16rc7, byte 16r90, byte 16rdf, byte 16re1, byte 16rde, byte 16r18, byte 16r34, 
		byte 16rf2, byte 16r80, byte 16rd7, byte 16r66, byte 16r7b, byte 16r32, byte 16r7f, byte 16r6c, 
		byte 16r8c, byte 16rd7, byte 16r55, byte 16r7e, byte 16r12, byte 16rac, byte 16r3a, byte 16r0f, 
		byte 16r93, byte 16rec, byte 16r05, byte 16rc5, byte 16r2e, byte 16r04, byte 16r93, byte 16ref, 
		byte 16r31, byte 16ra1, byte 16r2d, byte 16r3d, byte 16r92, byte 16r60, byte 16rf7, byte 16r9a, 
		byte 16r28, byte 16r9d, byte 16r6a, byte 16r37, byte 16r9b, byte 16rc7, byte 16r0c, byte 16r50, 
		byte 16r84, byte 16r14, byte 16r73, byte 16rd1, byte 16ra8, byte 16rcc, byte 16r81, byte 16rec, 
		byte 16r58, byte 16r3e, byte 16r96, byte 16r45, byte 16re0, byte 16r7b, byte 16r8d, byte 16r96, 
		byte 16r70, byte 16r65, byte 16r5b, byte 16ra5, byte 16rbb, byte 16rcf, byte 16rec, byte 16rc6, 
		byte 16rdc, byte 16r39, byte 16r66, byte 16r38, byte 16r0a, byte 16rd8, byte 16rfe, byte 16rcb, 
		byte 16r17, byte 16rb6, byte 16rba, byte 16r02, byte 16r46, byte 16r9a, byte 16r02, byte 16r0a, 
		byte 16r84, byte 16re1, byte 16r8e, byte 16r8f, byte 16r84, byte 16r25, byte 16r20, byte 16r70, 
		byte 16rc1, byte 16r3e, byte 16r9f, byte 16r1f, byte 16r28, byte 16r9b, byte 16re5, byte 16r4f, 
		byte 16rbc, byte 16r48, byte 16r14, byte 16r57, byte 16r77, byte 16r8f, byte 16r61, byte 16r60, 
		byte 16r15, byte 16re1, byte 16r32, byte 16r7a, byte 16r02, byte 16rb1, byte 16r40, byte 16rf1, 
		byte 16r50, byte 16r5e, byte 16rb3, byte 16r09, byte 16r32, byte 16r6d, byte 16r68, byte 16r37, 
		byte 16r8f, byte 16r83, byte 16r74, byte 16r59, byte 16r5c, byte 16r84, byte 16r9d, byte 16r84, 
		byte 16rf4, byte 16rc3, byte 16r33, byte 16rec, byte 16r44, byte 16r23, byte 16r88, byte 16r51, 
		byte 16r43, byte 16rcb, byte 16r47, byte 16rbd, byte 16r71, byte 16rc5, byte 16red, byte 16rae, 
		byte 16r9b, byte 16re6, byte 16r9a, byte 16r2f, byte 16rfe, byte 16rce, byte 16rb1, byte 16rbe, 
		byte 16rc9, byte 16rde, byte 16r24, byte 16r4f, byte 16rbe, byte 16r15, byte 16r99, byte 16r2b, 
		byte 16r11, byte 16rb7, byte 16r7c, byte 16r04, byte 16r0f, byte 16r12, byte 16rbd, byte 16r8f, 
		byte 16r6a, byte 16r97, byte 16r5a, byte 16r44, byte 16ra0, byte 16rf9, byte 16r0c, byte 16r29, 
		byte 16ra9, byte 16rab, byte 16rc3, byte 16rd4, byte 16rd8, byte 16r93, byte 16r92, byte 16r72, 
		byte 16r84, byte 16rc5, byte 16r87, byte 16r54, byte 16rcc, byte 16re2, byte 16r94, byte 16r52, 
		byte 16r9f, byte 16r86, byte 16r14, byte 16rdc, byte 16rd2, byte 16rab, byte 16ra9, byte 16r91, 
		byte 16r92, byte 16r5f, byte 16red, byte 16rc4, byte 16rae, byte 16r74, byte 16rff, byte 16rac, 
		byte 16r6e, byte 16r33, byte 16r3b, byte 16r93, byte 16reb, byte 16r4a, byte 16rff, byte 16r04, 
		byte 16r79, byte 16rda, byte 16r9a, byte 16r41, byte 16r0e, byte 16r44, byte 16r50, byte 16re0, 
		byte 16rdd, byte 16r7a, byte 16re4, byte 16rc6, byte 16re2, byte 16r91, byte 16r09, byte 16r00, 
		byte 16r57, byte 16r5d, byte 16ra4, byte 16r01, byte 16rfc, byte 16r07, byte 16r05, byte 16r9f, 
		byte 16r64, byte 16r5e, byte 16r8b, byte 16r7e, byte 16r9b, byte 16rfd, byte 16ref, byte 16r33, 
		byte 16r94, byte 16r30, byte 16r54, byte 16rff, byte 16r84, byte 16r01, byte 16r14, byte 16r93, 
		byte 16rc2, byte 16r7b, byte 16r34, byte 16r29, byte 16rea, byte 16red, byte 16rb4, byte 16red, 
		byte 16r53, byte 16r76, byte 16r44, byte 16r1a, byte 16r77, byte 16red, byte 16r43, byte 16r85, 
		byte 16r1a, byte 16rd7, byte 16r7f, byte 16r16, byte 16rf5, byte 16r41, byte 16rdf, byte 16rd2, 
		byte 16r69, byte 16rd5, byte 16r0d, byte 16r6a, byte 16r5f, byte 16r14, byte 16rfb, byte 16r0a, 
		byte 16rab, byte 16r1c, byte 16rbb, byte 16r4c, byte 16r15, byte 16r50, byte 16rbe, byte 16r97, 
		byte 16rf7, byte 16rab, byte 16r40, byte 16r66, byte 16r19, byte 16r3c, byte 16r4c, byte 16raa, 
		byte 16r77, byte 16r3d, byte 16rad, byte 16r38, byte 16r01, byte 16r4b, byte 16rd2, byte 16r09, 
		byte 16r2f, byte 16ra7, byte 16r55, byte 16rc8, byte 16r24, byte 16rbb, byte 16r5e, byte 16r54, 
		byte 16rc4, byte 16rf3, byte 16r6f, byte 16rfd, byte 16ra9, byte 16rfc, byte 16rea, byte 16r70, 
		byte 16rb9, byte 16rc6, byte 16re6, byte 16r93, byte 16re1, byte 16r48, byte 16rc1, byte 16r51, 
	}
),
Xts (
	big 16r0000ffff,
	array[] of {
		byte 16r27, byte 16r18, byte 16r28, byte 16r18, byte 16r28, byte 16r45, byte 16r90, byte 16r45, 
		byte 16r23, byte 16r53, byte 16r60, byte 16r28, byte 16r74, byte 16r71, byte 16r35, byte 16r26, 
		byte 16r62, byte 16r49, byte 16r77, byte 16r57, byte 16r24, byte 16r70, byte 16r93, byte 16r69, 
		byte 16r99, byte 16r59, byte 16r57, byte 16r49, byte 16r66, byte 16r96, byte 16r76, byte 16r27, 
		byte 16r31, byte 16r41, byte 16r59, byte 16r26, byte 16r53, byte 16r58, byte 16r97, byte 16r93, 
		byte 16r23, byte 16r84, byte 16r62, byte 16r64, byte 16r33, byte 16r83, byte 16r27, byte 16r95, 
		byte 16r02, byte 16r88, byte 16r41, byte 16r97, byte 16r16, byte 16r93, byte 16r99, byte 16r37, 
		byte 16r51, byte 16r05, byte 16r82, byte 16r09, byte 16r74, byte 16r94, byte 16r45, byte 16r92, 
	},
	array[] of {
		byte 16r00, byte 16r01, byte 16r02, byte 16r03, byte 16r04, byte 16r05, byte 16r06, byte 16r07, 
		byte 16r08, byte 16r09, byte 16r0a, byte 16r0b, byte 16r0c, byte 16r0d, byte 16r0e, byte 16r0f, 
		byte 16r10, byte 16r11, byte 16r12, byte 16r13, byte 16r14, byte 16r15, byte 16r16, byte 16r17, 
		byte 16r18, byte 16r19, byte 16r1a, byte 16r1b, byte 16r1c, byte 16r1d, byte 16r1e, byte 16r1f, 
		byte 16r20, byte 16r21, byte 16r22, byte 16r23, byte 16r24, byte 16r25, byte 16r26, byte 16r27, 
		byte 16r28, byte 16r29, byte 16r2a, byte 16r2b, byte 16r2c, byte 16r2d, byte 16r2e, byte 16r2f, 
		byte 16r30, byte 16r31, byte 16r32, byte 16r33, byte 16r34, byte 16r35, byte 16r36, byte 16r37, 
		byte 16r38, byte 16r39, byte 16r3a, byte 16r3b, byte 16r3c, byte 16r3d, byte 16r3e, byte 16r3f, 
		byte 16r40, byte 16r41, byte 16r42, byte 16r43, byte 16r44, byte 16r45, byte 16r46, byte 16r47, 
		byte 16r48, byte 16r49, byte 16r4a, byte 16r4b, byte 16r4c, byte 16r4d, byte 16r4e, byte 16r4f, 
		byte 16r50, byte 16r51, byte 16r52, byte 16r53, byte 16r54, byte 16r55, byte 16r56, byte 16r57, 
		byte 16r58, byte 16r59, byte 16r5a, byte 16r5b, byte 16r5c, byte 16r5d, byte 16r5e, byte 16r5f, 
		byte 16r60, byte 16r61, byte 16r62, byte 16r63, byte 16r64, byte 16r65, byte 16r66, byte 16r67, 
		byte 16r68, byte 16r69, byte 16r6a, byte 16r6b, byte 16r6c, byte 16r6d, byte 16r6e, byte 16r6f, 
		byte 16r70, byte 16r71, byte 16r72, byte 16r73, byte 16r74, byte 16r75, byte 16r76, byte 16r77, 
		byte 16r78, byte 16r79, byte 16r7a, byte 16r7b, byte 16r7c, byte 16r7d, byte 16r7e, byte 16r7f, 
		byte 16r80, byte 16r81, byte 16r82, byte 16r83, byte 16r84, byte 16r85, byte 16r86, byte 16r87, 
		byte 16r88, byte 16r89, byte 16r8a, byte 16r8b, byte 16r8c, byte 16r8d, byte 16r8e, byte 16r8f, 
		byte 16r90, byte 16r91, byte 16r92, byte 16r93, byte 16r94, byte 16r95, byte 16r96, byte 16r97, 
		byte 16r98, byte 16r99, byte 16r9a, byte 16r9b, byte 16r9c, byte 16r9d, byte 16r9e, byte 16r9f, 
		byte 16ra0, byte 16ra1, byte 16ra2, byte 16ra3, byte 16ra4, byte 16ra5, byte 16ra6, byte 16ra7, 
		byte 16ra8, byte 16ra9, byte 16raa, byte 16rab, byte 16rac, byte 16rad, byte 16rae, byte 16raf, 
		byte 16rb0, byte 16rb1, byte 16rb2, byte 16rb3, byte 16rb4, byte 16rb5, byte 16rb6, byte 16rb7, 
		byte 16rb8, byte 16rb9, byte 16rba, byte 16rbb, byte 16rbc, byte 16rbd, byte 16rbe, byte 16rbf, 
		byte 16rc0, byte 16rc1, byte 16rc2, byte 16rc3, byte 16rc4, byte 16rc5, byte 16rc6, byte 16rc7, 
		byte 16rc8, byte 16rc9, byte 16rca, byte 16rcb, byte 16rcc, byte 16rcd, byte 16rce, byte 16rcf, 
		byte 16rd0, byte 16rd1, byte 16rd2, byte 16rd3, byte 16rd4, byte 16rd5, byte 16rd6, byte 16rd7, 
		byte 16rd8, byte 16rd9, byte 16rda, byte 16rdb, byte 16rdc, byte 16rdd, byte 16rde, byte 16rdf, 
		byte 16re0, byte 16re1, byte 16re2, byte 16re3, byte 16re4, byte 16re5, byte 16re6, byte 16re7, 
		byte 16re8, byte 16re9, byte 16rea, byte 16reb, byte 16rec, byte 16red, byte 16ree, byte 16ref, 
		byte 16rf0, byte 16rf1, byte 16rf2, byte 16rf3, byte 16rf4, byte 16rf5, byte 16rf6, byte 16rf7, 
		byte 16rf8, byte 16rf9, byte 16rfa, byte 16rfb, byte 16rfc, byte 16rfd, byte 16rfe, byte 16rff, 
		byte 16r00, byte 16r01, byte 16r02, byte 16r03, byte 16r04, byte 16r05, byte 16r06, byte 16r07, 
		byte 16r08, byte 16r09, byte 16r0a, byte 16r0b, byte 16r0c, byte 16r0d, byte 16r0e, byte 16r0f, 
		byte 16r10, byte 16r11, byte 16r12, byte 16r13, byte 16r14, byte 16r15, byte 16r16, byte 16r17, 
		byte 16r18, byte 16r19, byte 16r1a, byte 16r1b, byte 16r1c, byte 16r1d, byte 16r1e, byte 16r1f, 
		byte 16r20, byte 16r21, byte 16r22, byte 16r23, byte 16r24, byte 16r25, byte 16r26, byte 16r27, 
		byte 16r28, byte 16r29, byte 16r2a, byte 16r2b, byte 16r2c, byte 16r2d, byte 16r2e, byte 16r2f, 
		byte 16r30, byte 16r31, byte 16r32, byte 16r33, byte 16r34, byte 16r35, byte 16r36, byte 16r37, 
		byte 16r38, byte 16r39, byte 16r3a, byte 16r3b, byte 16r3c, byte 16r3d, byte 16r3e, byte 16r3f, 
		byte 16r40, byte 16r41, byte 16r42, byte 16r43, byte 16r44, byte 16r45, byte 16r46, byte 16r47, 
		byte 16r48, byte 16r49, byte 16r4a, byte 16r4b, byte 16r4c, byte 16r4d, byte 16r4e, byte 16r4f, 
		byte 16r50, byte 16r51, byte 16r52, byte 16r53, byte 16r54, byte 16r55, byte 16r56, byte 16r57, 
		byte 16r58, byte 16r59, byte 16r5a, byte 16r5b, byte 16r5c, byte 16r5d, byte 16r5e, byte 16r5f, 
		byte 16r60, byte 16r61, byte 16r62, byte 16r63, byte 16r64, byte 16r65, byte 16r66, byte 16r67, 
		byte 16r68, byte 16r69, byte 16r6a, byte 16r6b, byte 16r6c, byte 16r6d, byte 16r6e, byte 16r6f, 
		byte 16r70, byte 16r71, byte 16r72, byte 16r73, byte 16r74, byte 16r75, byte 16r76, byte 16r77, 
		byte 16r78, byte 16r79, byte 16r7a, byte 16r7b, byte 16r7c, byte 16r7d, byte 16r7e, byte 16r7f, 
		byte 16r80, byte 16r81, byte 16r82, byte 16r83, byte 16r84, byte 16r85, byte 16r86, byte 16r87, 
		byte 16r88, byte 16r89, byte 16r8a, byte 16r8b, byte 16r8c, byte 16r8d, byte 16r8e, byte 16r8f, 
		byte 16r90, byte 16r91, byte 16r92, byte 16r93, byte 16r94, byte 16r95, byte 16r96, byte 16r97, 
		byte 16r98, byte 16r99, byte 16r9a, byte 16r9b, byte 16r9c, byte 16r9d, byte 16r9e, byte 16r9f, 
		byte 16ra0, byte 16ra1, byte 16ra2, byte 16ra3, byte 16ra4, byte 16ra5, byte 16ra6, byte 16ra7, 
		byte 16ra8, byte 16ra9, byte 16raa, byte 16rab, byte 16rac, byte 16rad, byte 16rae, byte 16raf, 
		byte 16rb0, byte 16rb1, byte 16rb2, byte 16rb3, byte 16rb4, byte 16rb5, byte 16rb6, byte 16rb7, 
		byte 16rb8, byte 16rb9, byte 16rba, byte 16rbb, byte 16rbc, byte 16rbd, byte 16rbe, byte 16rbf, 
		byte 16rc0, byte 16rc1, byte 16rc2, byte 16rc3, byte 16rc4, byte 16rc5, byte 16rc6, byte 16rc7, 
		byte 16rc8, byte 16rc9, byte 16rca, byte 16rcb, byte 16rcc, byte 16rcd, byte 16rce, byte 16rcf, 
		byte 16rd0, byte 16rd1, byte 16rd2, byte 16rd3, byte 16rd4, byte 16rd5, byte 16rd6, byte 16rd7, 
		byte 16rd8, byte 16rd9, byte 16rda, byte 16rdb, byte 16rdc, byte 16rdd, byte 16rde, byte 16rdf, 
		byte 16re0, byte 16re1, byte 16re2, byte 16re3, byte 16re4, byte 16re5, byte 16re6, byte 16re7, 
		byte 16re8, byte 16re9, byte 16rea, byte 16reb, byte 16rec, byte 16red, byte 16ree, byte 16ref, 
		byte 16rf0, byte 16rf1, byte 16rf2, byte 16rf3, byte 16rf4, byte 16rf5, byte 16rf6, byte 16rf7, 
		byte 16rf8, byte 16rf9, byte 16rfa, byte 16rfb, byte 16rfc, byte 16rfd, byte 16rfe, byte 16rff, 
	},
	array[] of {
		byte 16r77, byte 16ra3, byte 16r12, byte 16r51, byte 16r61, byte 16r8a, byte 16r15, byte 16re6, 
		byte 16rb9, byte 16r2d, byte 16r1d, byte 16r66, byte 16rdf, byte 16rfe, byte 16r7b, byte 16r50, 
		byte 16rb5, byte 16r0b, byte 16rad, byte 16r55, byte 16r23, byte 16r05, byte 16rba, byte 16r02, 
		byte 16r17, byte 16ra6, byte 16r10, byte 16r68, byte 16r8e, byte 16rff, byte 16r7e, byte 16r11, 
		byte 16re1, byte 16rd0, byte 16r22, byte 16r54, byte 16r38, byte 16re0, byte 16r93, byte 16r24, 
		byte 16r2d, byte 16r6d, byte 16rb2, byte 16r74, byte 16rfd, byte 16re8, byte 16r01, byte 16rd4, 
		byte 16rca, byte 16re0, byte 16r6f, byte 16r20, byte 16r92, byte 16rc7, byte 16r28, byte 16rb2, 
		byte 16r47, byte 16r85, byte 16r59, byte 16rdf, byte 16r58, byte 16re8, byte 16r37, byte 16rc2, 
		byte 16r46, byte 16r9e, byte 16re4, byte 16ra4, byte 16rfa, byte 16r79, byte 16r4e, byte 16r4b, 
		byte 16rbc, byte 16r7f, byte 16r39, byte 16rbc, byte 16r02, byte 16r6e, byte 16r3c, byte 16rb7, 
		byte 16r2c, byte 16r33, byte 16rb0, byte 16r88, byte 16r8f, byte 16r25, byte 16rb4, byte 16rac, 
		byte 16rf5, byte 16r6a, byte 16r2a, byte 16r98, byte 16r04, byte 16rf1, byte 16rce, byte 16r6d, 
		byte 16r3d, byte 16r6e, byte 16r1d, byte 16rc6, byte 16rca, byte 16r18, byte 16r1d, byte 16r4b, 
		byte 16r54, byte 16r61, byte 16r79, byte 16rd5, byte 16r55, byte 16r44, byte 16raa, byte 16r77, 
		byte 16r60, byte 16rc4, byte 16r0d, byte 16r06, byte 16r74, byte 16r15, byte 16r39, byte 16rc7, 
		byte 16re3, byte 16rcd, byte 16r9d, byte 16r2f, byte 16r66, byte 16r50, byte 16rb2, byte 16r01, 
		byte 16r3f, byte 16rd0, byte 16ree, byte 16rb8, byte 16rc2, byte 16rb8, byte 16re3, byte 16rd8, 
		byte 16rd2, byte 16r40, byte 16rcc, byte 16rae, byte 16r2d, byte 16r4c, byte 16r98, byte 16r32, 
		byte 16r0a, byte 16r74, byte 16r42, byte 16re1, byte 16rc8, byte 16rd7, byte 16r5a, byte 16r42, 
		byte 16rd6, byte 16re6, byte 16rcf, byte 16ra4, byte 16rc2, byte 16rec, byte 16ra1, byte 16r79, 
		byte 16r8d, byte 16r15, byte 16r8c, byte 16r7a, byte 16rec, byte 16rdf, byte 16r82, byte 16r49, 
		byte 16r0f, byte 16r24, byte 16rbb, byte 16r9b, byte 16r38, byte 16re1, byte 16r08, byte 16rbc, 
		byte 16rda, byte 16r12, byte 16rc3, byte 16rfa, byte 16rf9, byte 16ra2, byte 16r11, byte 16r41, 
		byte 16rc3, byte 16r61, byte 16r3b, byte 16r58, byte 16r36, byte 16r7f, byte 16r92, byte 16r2a, 
		byte 16raa, byte 16r26, byte 16rcd, byte 16r22, byte 16rf2, byte 16r3d, byte 16r70, byte 16r8d, 
		byte 16rae, byte 16r69, byte 16r9a, byte 16rd7, byte 16rcb, byte 16r40, byte 16ra8, byte 16rad, 
		byte 16r0b, byte 16r6e, byte 16r27, byte 16r84, byte 16r97, byte 16r3d, byte 16rcb, byte 16r60, 
		byte 16r56, byte 16r84, byte 16rc0, byte 16r8b, byte 16r8d, byte 16r69, byte 16r98, byte 16rc6, 
		byte 16r9a, byte 16rac, byte 16r04, byte 16r99, byte 16r21, byte 16r87, byte 16r1e, byte 16rbb, 
		byte 16r65, byte 16r30, byte 16r1a, byte 16r46, byte 16r19, byte 16rca, byte 16r80, byte 16rec, 
		byte 16rb4, byte 16r85, byte 16ra3, byte 16r1d, byte 16r74, byte 16r42, byte 16r23, byte 16rce, 
		byte 16r8d, byte 16rdc, byte 16r23, byte 16r94, byte 16r82, byte 16r8d, byte 16r6a, byte 16r80, 
		byte 16r47, byte 16r0c, byte 16r09, byte 16r2f, byte 16r5b, byte 16ra4, byte 16r13, byte 16rc3, 
		byte 16r37, byte 16r8f, byte 16ra6, byte 16r05, byte 16r42, byte 16r55, byte 16rc6, byte 16rf9, 
		byte 16rdf, byte 16r44, byte 16r95, byte 16r86, byte 16r2b, byte 16rbb, byte 16r32, byte 16r87, 
		byte 16r68, byte 16r1f, byte 16r93, byte 16r1b, byte 16r68, byte 16r7c, byte 16r88, byte 16r8a, 
		byte 16rbf, byte 16r84, byte 16r4d, byte 16rfc, byte 16r8f, byte 16rc2, byte 16r83, byte 16r31, 
		byte 16re5, byte 16r79, byte 16r92, byte 16r8c, byte 16rd1, byte 16r2b, byte 16rd2, byte 16r39, 
		byte 16r0a, byte 16re1, byte 16r23, byte 16rcf, byte 16r03, byte 16r81, byte 16r8d, byte 16r14, 
		byte 16rde, byte 16rdd, byte 16re5, byte 16rc0, byte 16rc2, byte 16r4c, byte 16r8a, byte 16rb0, 
		byte 16r18, byte 16rbf, byte 16rca, byte 16r75, byte 16rca, byte 16r09, byte 16r6f, byte 16r2d, 
		byte 16r53, byte 16r1f, byte 16r3d, byte 16r16, byte 16r19, byte 16re7, byte 16r85, byte 16rf1, 
		byte 16rad, byte 16ra4, byte 16r37, byte 16rca, byte 16rb9, byte 16r2e, byte 16r98, byte 16r05, 
		byte 16r58, byte 16rb3, byte 16rdc, byte 16re1, byte 16r47, byte 16r4a, byte 16rfb, byte 16r75, 
		byte 16rbf, byte 16red, byte 16rbf, byte 16r8f, byte 16rf5, byte 16r4c, byte 16rb2, byte 16r61, 
		byte 16r8e, byte 16r02, byte 16r44, byte 16rc9, byte 16rac, byte 16r0d, byte 16r3c, byte 16r66, 
		byte 16rfb, byte 16r51, byte 16r59, byte 16r8c, byte 16rd2, byte 16rdb, byte 16r11, byte 16rf9, 
		byte 16rbe, byte 16r39, byte 16r79, byte 16r1a, byte 16rbe, byte 16r44, byte 16r7c, byte 16r63, 
		byte 16r09, byte 16r4f, byte 16r7c, byte 16r45, byte 16r3b, byte 16r7f, byte 16rf8, byte 16r7c, 
		byte 16rb5, byte 16rbb, byte 16r36, byte 16rb7, byte 16rc7, byte 16r9e, byte 16rfb, byte 16r08, 
		byte 16r72, byte 16rd1, byte 16r70, byte 16r58, byte 16rb8, byte 16r3b, byte 16r15, byte 16rab, 
		byte 16r08, byte 16r66, byte 16rad, byte 16r8a, byte 16r58, byte 16r65, byte 16r6c, byte 16r5a, 
		byte 16r7e, byte 16r20, byte 16rdb, byte 16rdf, byte 16r30, byte 16r8b, byte 16r24, byte 16r61, 
		byte 16rd9, byte 16r7c, byte 16r0e, byte 16rc0, byte 16r02, byte 16r4a, byte 16r27, byte 16r15, 
		byte 16r05, byte 16r52, byte 16r49, byte 16rcf, byte 16r3b, byte 16r47, byte 16r8d, byte 16rdd, 
		byte 16r47, byte 16r40, byte 16rde, byte 16r65, byte 16r4f, byte 16r75, byte 16rca, byte 16r68, 
		byte 16r6e, byte 16r0d, byte 16r73, byte 16r45, byte 16rc6, byte 16r9e, byte 16rd5, byte 16r0c, 
		byte 16rdc, byte 16r2a, byte 16r8b, byte 16r33, byte 16r2b, byte 16r1f, byte 16r88, byte 16r24, 
		byte 16r10, byte 16r8a, byte 16rc9, byte 16r37, byte 16reb, byte 16r05, byte 16r05, byte 16r85, 
		byte 16r60, byte 16r8e, byte 16re7, byte 16r34, byte 16r09, byte 16r7f, byte 16rc0, byte 16r90, 
		byte 16r54, byte 16rfb, byte 16rff, byte 16r89, byte 16ree, byte 16rae, byte 16rea, byte 16r79, 
		byte 16r1f, byte 16r4a, byte 16r7a, byte 16rb1, byte 16rf9, byte 16r86, byte 16r82, byte 16r94, 
		byte 16ra4, byte 16rf9, byte 16re2, byte 16r7b, byte 16r42, byte 16raf, byte 16r81, byte 16r00, 
		byte 16rcb, byte 16r9d, byte 16r59, byte 16rce, byte 16rf9, byte 16r64, byte 16r58, byte 16r03, 
	}
),
Xts (
	big 16r00ffffff,
	array[] of {
		byte 16r27, byte 16r18, byte 16r28, byte 16r18, byte 16r28, byte 16r45, byte 16r90, byte 16r45, 
		byte 16r23, byte 16r53, byte 16r60, byte 16r28, byte 16r74, byte 16r71, byte 16r35, byte 16r26, 
		byte 16r62, byte 16r49, byte 16r77, byte 16r57, byte 16r24, byte 16r70, byte 16r93, byte 16r69, 
		byte 16r99, byte 16r59, byte 16r57, byte 16r49, byte 16r66, byte 16r96, byte 16r76, byte 16r27, 
		byte 16r31, byte 16r41, byte 16r59, byte 16r26, byte 16r53, byte 16r58, byte 16r97, byte 16r93, 
		byte 16r23, byte 16r84, byte 16r62, byte 16r64, byte 16r33, byte 16r83, byte 16r27, byte 16r95, 
		byte 16r02, byte 16r88, byte 16r41, byte 16r97, byte 16r16, byte 16r93, byte 16r99, byte 16r37, 
		byte 16r51, byte 16r05, byte 16r82, byte 16r09, byte 16r74, byte 16r94, byte 16r45, byte 16r92, 
	},
	array[] of {
		byte 16r00, byte 16r01, byte 16r02, byte 16r03, byte 16r04, byte 16r05, byte 16r06, byte 16r07, 
		byte 16r08, byte 16r09, byte 16r0a, byte 16r0b, byte 16r0c, byte 16r0d, byte 16r0e, byte 16r0f, 
		byte 16r10, byte 16r11, byte 16r12, byte 16r13, byte 16r14, byte 16r15, byte 16r16, byte 16r17, 
		byte 16r18, byte 16r19, byte 16r1a, byte 16r1b, byte 16r1c, byte 16r1d, byte 16r1e, byte 16r1f, 
		byte 16r20, byte 16r21, byte 16r22, byte 16r23, byte 16r24, byte 16r25, byte 16r26, byte 16r27, 
		byte 16r28, byte 16r29, byte 16r2a, byte 16r2b, byte 16r2c, byte 16r2d, byte 16r2e, byte 16r2f, 
		byte 16r30, byte 16r31, byte 16r32, byte 16r33, byte 16r34, byte 16r35, byte 16r36, byte 16r37, 
		byte 16r38, byte 16r39, byte 16r3a, byte 16r3b, byte 16r3c, byte 16r3d, byte 16r3e, byte 16r3f, 
		byte 16r40, byte 16r41, byte 16r42, byte 16r43, byte 16r44, byte 16r45, byte 16r46, byte 16r47, 
		byte 16r48, byte 16r49, byte 16r4a, byte 16r4b, byte 16r4c, byte 16r4d, byte 16r4e, byte 16r4f, 
		byte 16r50, byte 16r51, byte 16r52, byte 16r53, byte 16r54, byte 16r55, byte 16r56, byte 16r57, 
		byte 16r58, byte 16r59, byte 16r5a, byte 16r5b, byte 16r5c, byte 16r5d, byte 16r5e, byte 16r5f, 
		byte 16r60, byte 16r61, byte 16r62, byte 16r63, byte 16r64, byte 16r65, byte 16r66, byte 16r67, 
		byte 16r68, byte 16r69, byte 16r6a, byte 16r6b, byte 16r6c, byte 16r6d, byte 16r6e, byte 16r6f, 
		byte 16r70, byte 16r71, byte 16r72, byte 16r73, byte 16r74, byte 16r75, byte 16r76, byte 16r77, 
		byte 16r78, byte 16r79, byte 16r7a, byte 16r7b, byte 16r7c, byte 16r7d, byte 16r7e, byte 16r7f, 
		byte 16r80, byte 16r81, byte 16r82, byte 16r83, byte 16r84, byte 16r85, byte 16r86, byte 16r87, 
		byte 16r88, byte 16r89, byte 16r8a, byte 16r8b, byte 16r8c, byte 16r8d, byte 16r8e, byte 16r8f, 
		byte 16r90, byte 16r91, byte 16r92, byte 16r93, byte 16r94, byte 16r95, byte 16r96, byte 16r97, 
		byte 16r98, byte 16r99, byte 16r9a, byte 16r9b, byte 16r9c, byte 16r9d, byte 16r9e, byte 16r9f, 
		byte 16ra0, byte 16ra1, byte 16ra2, byte 16ra3, byte 16ra4, byte 16ra5, byte 16ra6, byte 16ra7, 
		byte 16ra8, byte 16ra9, byte 16raa, byte 16rab, byte 16rac, byte 16rad, byte 16rae, byte 16raf, 
		byte 16rb0, byte 16rb1, byte 16rb2, byte 16rb3, byte 16rb4, byte 16rb5, byte 16rb6, byte 16rb7, 
		byte 16rb8, byte 16rb9, byte 16rba, byte 16rbb, byte 16rbc, byte 16rbd, byte 16rbe, byte 16rbf, 
		byte 16rc0, byte 16rc1, byte 16rc2, byte 16rc3, byte 16rc4, byte 16rc5, byte 16rc6, byte 16rc7, 
		byte 16rc8, byte 16rc9, byte 16rca, byte 16rcb, byte 16rcc, byte 16rcd, byte 16rce, byte 16rcf, 
		byte 16rd0, byte 16rd1, byte 16rd2, byte 16rd3, byte 16rd4, byte 16rd5, byte 16rd6, byte 16rd7, 
		byte 16rd8, byte 16rd9, byte 16rda, byte 16rdb, byte 16rdc, byte 16rdd, byte 16rde, byte 16rdf, 
		byte 16re0, byte 16re1, byte 16re2, byte 16re3, byte 16re4, byte 16re5, byte 16re6, byte 16re7, 
		byte 16re8, byte 16re9, byte 16rea, byte 16reb, byte 16rec, byte 16red, byte 16ree, byte 16ref, 
		byte 16rf0, byte 16rf1, byte 16rf2, byte 16rf3, byte 16rf4, byte 16rf5, byte 16rf6, byte 16rf7, 
		byte 16rf8, byte 16rf9, byte 16rfa, byte 16rfb, byte 16rfc, byte 16rfd, byte 16rfe, byte 16rff, 
		byte 16r00, byte 16r01, byte 16r02, byte 16r03, byte 16r04, byte 16r05, byte 16r06, byte 16r07, 
		byte 16r08, byte 16r09, byte 16r0a, byte 16r0b, byte 16r0c, byte 16r0d, byte 16r0e, byte 16r0f, 
		byte 16r10, byte 16r11, byte 16r12, byte 16r13, byte 16r14, byte 16r15, byte 16r16, byte 16r17, 
		byte 16r18, byte 16r19, byte 16r1a, byte 16r1b, byte 16r1c, byte 16r1d, byte 16r1e, byte 16r1f, 
		byte 16r20, byte 16r21, byte 16r22, byte 16r23, byte 16r24, byte 16r25, byte 16r26, byte 16r27, 
		byte 16r28, byte 16r29, byte 16r2a, byte 16r2b, byte 16r2c, byte 16r2d, byte 16r2e, byte 16r2f, 
		byte 16r30, byte 16r31, byte 16r32, byte 16r33, byte 16r34, byte 16r35, byte 16r36, byte 16r37, 
		byte 16r38, byte 16r39, byte 16r3a, byte 16r3b, byte 16r3c, byte 16r3d, byte 16r3e, byte 16r3f, 
		byte 16r40, byte 16r41, byte 16r42, byte 16r43, byte 16r44, byte 16r45, byte 16r46, byte 16r47, 
		byte 16r48, byte 16r49, byte 16r4a, byte 16r4b, byte 16r4c, byte 16r4d, byte 16r4e, byte 16r4f, 
		byte 16r50, byte 16r51, byte 16r52, byte 16r53, byte 16r54, byte 16r55, byte 16r56, byte 16r57, 
		byte 16r58, byte 16r59, byte 16r5a, byte 16r5b, byte 16r5c, byte 16r5d, byte 16r5e, byte 16r5f, 
		byte 16r60, byte 16r61, byte 16r62, byte 16r63, byte 16r64, byte 16r65, byte 16r66, byte 16r67, 
		byte 16r68, byte 16r69, byte 16r6a, byte 16r6b, byte 16r6c, byte 16r6d, byte 16r6e, byte 16r6f, 
		byte 16r70, byte 16r71, byte 16r72, byte 16r73, byte 16r74, byte 16r75, byte 16r76, byte 16r77, 
		byte 16r78, byte 16r79, byte 16r7a, byte 16r7b, byte 16r7c, byte 16r7d, byte 16r7e, byte 16r7f, 
		byte 16r80, byte 16r81, byte 16r82, byte 16r83, byte 16r84, byte 16r85, byte 16r86, byte 16r87, 
		byte 16r88, byte 16r89, byte 16r8a, byte 16r8b, byte 16r8c, byte 16r8d, byte 16r8e, byte 16r8f, 
		byte 16r90, byte 16r91, byte 16r92, byte 16r93, byte 16r94, byte 16r95, byte 16r96, byte 16r97, 
		byte 16r98, byte 16r99, byte 16r9a, byte 16r9b, byte 16r9c, byte 16r9d, byte 16r9e, byte 16r9f, 
		byte 16ra0, byte 16ra1, byte 16ra2, byte 16ra3, byte 16ra4, byte 16ra5, byte 16ra6, byte 16ra7, 
		byte 16ra8, byte 16ra9, byte 16raa, byte 16rab, byte 16rac, byte 16rad, byte 16rae, byte 16raf, 
		byte 16rb0, byte 16rb1, byte 16rb2, byte 16rb3, byte 16rb4, byte 16rb5, byte 16rb6, byte 16rb7, 
		byte 16rb8, byte 16rb9, byte 16rba, byte 16rbb, byte 16rbc, byte 16rbd, byte 16rbe, byte 16rbf, 
		byte 16rc0, byte 16rc1, byte 16rc2, byte 16rc3, byte 16rc4, byte 16rc5, byte 16rc6, byte 16rc7, 
		byte 16rc8, byte 16rc9, byte 16rca, byte 16rcb, byte 16rcc, byte 16rcd, byte 16rce, byte 16rcf, 
		byte 16rd0, byte 16rd1, byte 16rd2, byte 16rd3, byte 16rd4, byte 16rd5, byte 16rd6, byte 16rd7, 
		byte 16rd8, byte 16rd9, byte 16rda, byte 16rdb, byte 16rdc, byte 16rdd, byte 16rde, byte 16rdf, 
		byte 16re0, byte 16re1, byte 16re2, byte 16re3, byte 16re4, byte 16re5, byte 16re6, byte 16re7, 
		byte 16re8, byte 16re9, byte 16rea, byte 16reb, byte 16rec, byte 16red, byte 16ree, byte 16ref, 
		byte 16rf0, byte 16rf1, byte 16rf2, byte 16rf3, byte 16rf4, byte 16rf5, byte 16rf6, byte 16rf7, 
		byte 16rf8, byte 16rf9, byte 16rfa, byte 16rfb, byte 16rfc, byte 16rfd, byte 16rfe, byte 16rff, 
	},
	array[] of {
		byte 16re3, byte 16r87, byte 16raa, byte 16ra5, byte 16r8b, byte 16ra4, byte 16r83, byte 16raf, 
		byte 16ra7, byte 16re8, byte 16reb, byte 16r46, byte 16r97, byte 16r78, byte 16r31, byte 16r7e, 
		byte 16rcf, byte 16r4c, byte 16rf5, byte 16r73, byte 16raa, byte 16r9d, byte 16r4e, byte 16rac, 
		byte 16r23, byte 16rf2, byte 16rcd, byte 16rf9, byte 16r14, byte 16re4, byte 16re2, byte 16r00, 
		byte 16ra8, byte 16rb4, byte 16r90, byte 16re4, byte 16r2e, byte 16re6, byte 16r46, byte 16r80, 
		byte 16r2d, byte 16rc6, byte 16ree, byte 16r2b, byte 16r47, byte 16r1b, byte 16r27, byte 16r81, 
		byte 16r95, byte 16rd6, byte 16r09, byte 16r18, byte 16rec, byte 16rec, byte 16rb4, byte 16r4b, 
		byte 16rf7, byte 16r99, byte 16r66, byte 16rf8, byte 16r3f, byte 16rab, byte 16ra0, byte 16r49, 
		byte 16r92, byte 16r98, byte 16reb, byte 16rc6, byte 16r99, byte 16rc0, byte 16rc8, byte 16r63, 
		byte 16r47, byte 16r15, byte 16ra3, byte 16r20, byte 16rbb, byte 16r4f, byte 16r07, byte 16r5d, 
		byte 16r62, byte 16r2e, byte 16r74, byte 16rc8, byte 16rc9, byte 16r32, byte 16r00, byte 16r4f, 
		byte 16r25, byte 16rb4, byte 16r1e, byte 16r36, byte 16r10, byte 16r25, byte 16rb5, byte 16ra8, 
		byte 16r78, byte 16r15, byte 16r39, byte 16r1f, byte 16r61, byte 16r08, byte 16rfc, byte 16r4a, 
		byte 16rfa, byte 16r6a, byte 16r05, byte 16rd9, byte 16r30, byte 16r3c, byte 16r6b, byte 16ra6, 
		byte 16r8a, byte 16r12, byte 16r8a, byte 16r55, byte 16r70, byte 16r5d, byte 16r41, byte 16r59, 
		byte 16r85, byte 16r83, byte 16r2f, byte 16rde, byte 16raa, byte 16re6, byte 16rc8, byte 16re1, 
		byte 16r91, byte 16r10, byte 16re8, byte 16r4d, byte 16r1b, byte 16r1f, byte 16r19, byte 16r9a, 
		byte 16r26, byte 16r92, byte 16r11, byte 16r9e, byte 16rdc, byte 16r96, byte 16r13, byte 16r26, 
		byte 16r58, byte 16rf0, byte 16r9d, byte 16ra7, byte 16rc6, byte 16r23, byte 16ref, byte 16rce, 
		byte 16rc7, byte 16r12, byte 16r53, byte 16r7a, byte 16r3d, byte 16r94, byte 16rc0, byte 16rbf, 
		byte 16r5d, byte 16r7e, byte 16r35, byte 16r2e, byte 16rc9, byte 16r4a, byte 16re5, byte 16r79, 
		byte 16r7f, byte 16rdb, byte 16r37, byte 16r7d, byte 16rc1, byte 16r55, byte 16r11, byte 16r50, 
		byte 16r72, byte 16r1a, byte 16rdf, byte 16r15, byte 16rbd, byte 16r26, byte 16ra8, byte 16ref, 
		byte 16rc2, byte 16rfc, byte 16raa, byte 16rd5, byte 16r68, byte 16r81, byte 16rfa, byte 16r9e, 
		byte 16r62, byte 16r46, byte 16r2c, byte 16r28, byte 16rf3, byte 16r0a, byte 16re1, byte 16rce, 
		byte 16rac, byte 16ra9, byte 16r3c, byte 16r34, byte 16r5c, byte 16rf2, byte 16r43, byte 16rb7, 
		byte 16r3f, byte 16r54, byte 16r2e, byte 16r20, byte 16r74, byte 16ra7, byte 16r05, byte 16rbd, 
		byte 16r26, byte 16r43, byte 16rbb, byte 16r9f, byte 16r7c, byte 16rc7, byte 16r9b, byte 16rb6, 
		byte 16re7, byte 16r09, byte 16r1e, byte 16ra6, byte 16re2, byte 16r32, byte 16rdf, byte 16r0f, 
		byte 16r9a, byte 16rd0, byte 16rd6, byte 16rcf, byte 16r50, byte 16r23, byte 16r27, byte 16r87, 
		byte 16r6d, byte 16r82, byte 16r20, byte 16r7a, byte 16rbf, byte 16r21, byte 16r15, byte 16rcd, 
		byte 16rac, byte 16rf6, byte 16rd5, byte 16ra4, byte 16r8f, byte 16r6c, byte 16r18, byte 16r79, 
		byte 16ra6, byte 16r5b, byte 16r11, byte 16r5f, byte 16r0f, byte 16r8b, byte 16r3c, byte 16rb3, 
		byte 16rc5, byte 16r9d, byte 16r15, byte 16rdd, byte 16r8c, byte 16r76, byte 16r9b, byte 16rc0, 
		byte 16r14, byte 16r79, byte 16r5a, byte 16r18, byte 16r37, byte 16rf3, byte 16r90, byte 16r1b, 
		byte 16r58, byte 16r45, byte 16reb, byte 16r49, byte 16r1a, byte 16rdf, byte 16ref, byte 16re0, 
		byte 16r97, byte 16rb1, byte 16rfa, byte 16r30, byte 16ra1, byte 16r2f, byte 16rc1, byte 16rf6, 
		byte 16r5b, byte 16ra2, byte 16r29, byte 16r05, byte 16r03, byte 16r15, byte 16r39, byte 16r97, 
		byte 16r1a, byte 16r10, byte 16rf2, byte 16rf3, byte 16r6c, byte 16r32, byte 16r1b, byte 16rb5, 
		byte 16r13, byte 16r31, byte 16rcd, byte 16ref, byte 16rb3, byte 16r9e, byte 16r39, byte 16r64, 
		byte 16rc7, byte 16ref, byte 16r07, byte 16r99, byte 16r94, byte 16rf5, byte 16rb6, byte 16r9b, 
		byte 16r2e, byte 16rdd, byte 16r83, byte 16ra7, byte 16r1e, byte 16rf5, byte 16r49, byte 16r97, 
		byte 16r1e, byte 16re9, byte 16r3f, byte 16r44, byte 16rea, byte 16rc3, byte 16r93, byte 16r8f, 
		byte 16rcd, byte 16rd6, byte 16r1d, byte 16r01, byte 16rfa, byte 16r71, byte 16r79, byte 16r9d, 
		byte 16ra3, byte 16ra8, byte 16r09, byte 16r1c, byte 16r4c, byte 16r48, byte 16raa, byte 16r9e, 
		byte 16rd2, byte 16r63, byte 16rff, byte 16r07, byte 16r49, byte 16rdf, byte 16r95, byte 16rd4, 
		byte 16r4f, byte 16ref, byte 16r6a, byte 16r0b, byte 16rb5, byte 16r78, byte 16rec, byte 16r69, 
		byte 16r45, byte 16r6a, byte 16ra5, byte 16r40, byte 16r8a, byte 16re3, byte 16r2c, byte 16r7a, 
		byte 16rf0, byte 16r8a, byte 16rd7, byte 16rba, byte 16r89, byte 16r21, byte 16r28, byte 16r7e, 
		byte 16r3b, byte 16rbe, byte 16re3, byte 16r1b, byte 16r76, byte 16r7b, byte 16re0, byte 16r6a, 
		byte 16r0e, byte 16r70, byte 16r5c, byte 16r86, byte 16r4a, byte 16r76, byte 16r91, byte 16r37, 
		byte 16rdf, byte 16r28, byte 16r29, byte 16r22, byte 16r83, byte 16rea, byte 16r81, byte 16ra2, 
		byte 16r48, byte 16r02, byte 16r41, byte 16rb4, byte 16r4d, byte 16r99, byte 16r21, byte 16rcd, 
		byte 16rbe, byte 16rc1, byte 16rbc, byte 16r28, byte 16rdc, byte 16r1f, byte 16rda, byte 16r11, 
		byte 16r4b, byte 16rd8, byte 16re5, byte 16r21, byte 16r7a, byte 16rc9, byte 16rd8, byte 16reb, 
		byte 16raf, byte 16ra7, byte 16r20, byte 16re9, byte 16rda, byte 16r4f, byte 16r9a, byte 16rce, 
		byte 16r23, byte 16r1c, byte 16rc9, byte 16r49, byte 16re5, byte 16rb9, byte 16r6f, byte 16re7, 
		byte 16r6f, byte 16rfc, byte 16r21, byte 16r06, byte 16r3f, byte 16rdd, byte 16rc8, byte 16r3a, 
		byte 16r6b, byte 16r86, byte 16r79, byte 16rc0, byte 16r0d, byte 16r35, byte 16re0, byte 16r95, 
		byte 16r76, byte 16ra8, byte 16r75, byte 16r30, byte 16r5b, byte 16red, byte 16r5f, byte 16r36, 
		byte 16red, byte 16r24, byte 16r2c, byte 16r89, byte 16r00, byte 16rdd, byte 16r1f, byte 16ra9, 
		byte 16r65, byte 16rbc, byte 16r95, byte 16r0d, byte 16rfc, byte 16re0, byte 16r9b, byte 16r13, 
		byte 16r22, byte 16r63, byte 16ra1, byte 16ree, byte 16rf5, byte 16r2d, byte 16rd6, byte 16r88, 
		byte 16r8c, byte 16r30, byte 16r9f, byte 16r5a, byte 16r7d, byte 16r71, byte 16r28, byte 16r26, 
	}
),
Xts (
	big 16rffffffff,
	array[] of {
		byte 16r27, byte 16r18, byte 16r28, byte 16r18, byte 16r28, byte 16r45, byte 16r90, byte 16r45, 
		byte 16r23, byte 16r53, byte 16r60, byte 16r28, byte 16r74, byte 16r71, byte 16r35, byte 16r26, 
		byte 16r62, byte 16r49, byte 16r77, byte 16r57, byte 16r24, byte 16r70, byte 16r93, byte 16r69, 
		byte 16r99, byte 16r59, byte 16r57, byte 16r49, byte 16r66, byte 16r96, byte 16r76, byte 16r27, 
		byte 16r31, byte 16r41, byte 16r59, byte 16r26, byte 16r53, byte 16r58, byte 16r97, byte 16r93, 
		byte 16r23, byte 16r84, byte 16r62, byte 16r64, byte 16r33, byte 16r83, byte 16r27, byte 16r95, 
		byte 16r02, byte 16r88, byte 16r41, byte 16r97, byte 16r16, byte 16r93, byte 16r99, byte 16r37, 
		byte 16r51, byte 16r05, byte 16r82, byte 16r09, byte 16r74, byte 16r94, byte 16r45, byte 16r92, 
	},
	array[] of {
		byte 16r00, byte 16r01, byte 16r02, byte 16r03, byte 16r04, byte 16r05, byte 16r06, byte 16r07, 
		byte 16r08, byte 16r09, byte 16r0a, byte 16r0b, byte 16r0c, byte 16r0d, byte 16r0e, byte 16r0f, 
		byte 16r10, byte 16r11, byte 16r12, byte 16r13, byte 16r14, byte 16r15, byte 16r16, byte 16r17, 
		byte 16r18, byte 16r19, byte 16r1a, byte 16r1b, byte 16r1c, byte 16r1d, byte 16r1e, byte 16r1f, 
		byte 16r20, byte 16r21, byte 16r22, byte 16r23, byte 16r24, byte 16r25, byte 16r26, byte 16r27, 
		byte 16r28, byte 16r29, byte 16r2a, byte 16r2b, byte 16r2c, byte 16r2d, byte 16r2e, byte 16r2f, 
		byte 16r30, byte 16r31, byte 16r32, byte 16r33, byte 16r34, byte 16r35, byte 16r36, byte 16r37, 
		byte 16r38, byte 16r39, byte 16r3a, byte 16r3b, byte 16r3c, byte 16r3d, byte 16r3e, byte 16r3f, 
		byte 16r40, byte 16r41, byte 16r42, byte 16r43, byte 16r44, byte 16r45, byte 16r46, byte 16r47, 
		byte 16r48, byte 16r49, byte 16r4a, byte 16r4b, byte 16r4c, byte 16r4d, byte 16r4e, byte 16r4f, 
		byte 16r50, byte 16r51, byte 16r52, byte 16r53, byte 16r54, byte 16r55, byte 16r56, byte 16r57, 
		byte 16r58, byte 16r59, byte 16r5a, byte 16r5b, byte 16r5c, byte 16r5d, byte 16r5e, byte 16r5f, 
		byte 16r60, byte 16r61, byte 16r62, byte 16r63, byte 16r64, byte 16r65, byte 16r66, byte 16r67, 
		byte 16r68, byte 16r69, byte 16r6a, byte 16r6b, byte 16r6c, byte 16r6d, byte 16r6e, byte 16r6f, 
		byte 16r70, byte 16r71, byte 16r72, byte 16r73, byte 16r74, byte 16r75, byte 16r76, byte 16r77, 
		byte 16r78, byte 16r79, byte 16r7a, byte 16r7b, byte 16r7c, byte 16r7d, byte 16r7e, byte 16r7f, 
		byte 16r80, byte 16r81, byte 16r82, byte 16r83, byte 16r84, byte 16r85, byte 16r86, byte 16r87, 
		byte 16r88, byte 16r89, byte 16r8a, byte 16r8b, byte 16r8c, byte 16r8d, byte 16r8e, byte 16r8f, 
		byte 16r90, byte 16r91, byte 16r92, byte 16r93, byte 16r94, byte 16r95, byte 16r96, byte 16r97, 
		byte 16r98, byte 16r99, byte 16r9a, byte 16r9b, byte 16r9c, byte 16r9d, byte 16r9e, byte 16r9f, 
		byte 16ra0, byte 16ra1, byte 16ra2, byte 16ra3, byte 16ra4, byte 16ra5, byte 16ra6, byte 16ra7, 
		byte 16ra8, byte 16ra9, byte 16raa, byte 16rab, byte 16rac, byte 16rad, byte 16rae, byte 16raf, 
		byte 16rb0, byte 16rb1, byte 16rb2, byte 16rb3, byte 16rb4, byte 16rb5, byte 16rb6, byte 16rb7, 
		byte 16rb8, byte 16rb9, byte 16rba, byte 16rbb, byte 16rbc, byte 16rbd, byte 16rbe, byte 16rbf, 
		byte 16rc0, byte 16rc1, byte 16rc2, byte 16rc3, byte 16rc4, byte 16rc5, byte 16rc6, byte 16rc7, 
		byte 16rc8, byte 16rc9, byte 16rca, byte 16rcb, byte 16rcc, byte 16rcd, byte 16rce, byte 16rcf, 
		byte 16rd0, byte 16rd1, byte 16rd2, byte 16rd3, byte 16rd4, byte 16rd5, byte 16rd6, byte 16rd7, 
		byte 16rd8, byte 16rd9, byte 16rda, byte 16rdb, byte 16rdc, byte 16rdd, byte 16rde, byte 16rdf, 
		byte 16re0, byte 16re1, byte 16re2, byte 16re3, byte 16re4, byte 16re5, byte 16re6, byte 16re7, 
		byte 16re8, byte 16re9, byte 16rea, byte 16reb, byte 16rec, byte 16red, byte 16ree, byte 16ref, 
		byte 16rf0, byte 16rf1, byte 16rf2, byte 16rf3, byte 16rf4, byte 16rf5, byte 16rf6, byte 16rf7, 
		byte 16rf8, byte 16rf9, byte 16rfa, byte 16rfb, byte 16rfc, byte 16rfd, byte 16rfe, byte 16rff, 
		byte 16r00, byte 16r01, byte 16r02, byte 16r03, byte 16r04, byte 16r05, byte 16r06, byte 16r07, 
		byte 16r08, byte 16r09, byte 16r0a, byte 16r0b, byte 16r0c, byte 16r0d, byte 16r0e, byte 16r0f, 
		byte 16r10, byte 16r11, byte 16r12, byte 16r13, byte 16r14, byte 16r15, byte 16r16, byte 16r17, 
		byte 16r18, byte 16r19, byte 16r1a, byte 16r1b, byte 16r1c, byte 16r1d, byte 16r1e, byte 16r1f, 
		byte 16r20, byte 16r21, byte 16r22, byte 16r23, byte 16r24, byte 16r25, byte 16r26, byte 16r27, 
		byte 16r28, byte 16r29, byte 16r2a, byte 16r2b, byte 16r2c, byte 16r2d, byte 16r2e, byte 16r2f, 
		byte 16r30, byte 16r31, byte 16r32, byte 16r33, byte 16r34, byte 16r35, byte 16r36, byte 16r37, 
		byte 16r38, byte 16r39, byte 16r3a, byte 16r3b, byte 16r3c, byte 16r3d, byte 16r3e, byte 16r3f, 
		byte 16r40, byte 16r41, byte 16r42, byte 16r43, byte 16r44, byte 16r45, byte 16r46, byte 16r47, 
		byte 16r48, byte 16r49, byte 16r4a, byte 16r4b, byte 16r4c, byte 16r4d, byte 16r4e, byte 16r4f, 
		byte 16r50, byte 16r51, byte 16r52, byte 16r53, byte 16r54, byte 16r55, byte 16r56, byte 16r57, 
		byte 16r58, byte 16r59, byte 16r5a, byte 16r5b, byte 16r5c, byte 16r5d, byte 16r5e, byte 16r5f, 
		byte 16r60, byte 16r61, byte 16r62, byte 16r63, byte 16r64, byte 16r65, byte 16r66, byte 16r67, 
		byte 16r68, byte 16r69, byte 16r6a, byte 16r6b, byte 16r6c, byte 16r6d, byte 16r6e, byte 16r6f, 
		byte 16r70, byte 16r71, byte 16r72, byte 16r73, byte 16r74, byte 16r75, byte 16r76, byte 16r77, 
		byte 16r78, byte 16r79, byte 16r7a, byte 16r7b, byte 16r7c, byte 16r7d, byte 16r7e, byte 16r7f, 
		byte 16r80, byte 16r81, byte 16r82, byte 16r83, byte 16r84, byte 16r85, byte 16r86, byte 16r87, 
		byte 16r88, byte 16r89, byte 16r8a, byte 16r8b, byte 16r8c, byte 16r8d, byte 16r8e, byte 16r8f, 
		byte 16r90, byte 16r91, byte 16r92, byte 16r93, byte 16r94, byte 16r95, byte 16r96, byte 16r97, 
		byte 16r98, byte 16r99, byte 16r9a, byte 16r9b, byte 16r9c, byte 16r9d, byte 16r9e, byte 16r9f, 
		byte 16ra0, byte 16ra1, byte 16ra2, byte 16ra3, byte 16ra4, byte 16ra5, byte 16ra6, byte 16ra7, 
		byte 16ra8, byte 16ra9, byte 16raa, byte 16rab, byte 16rac, byte 16rad, byte 16rae, byte 16raf, 
		byte 16rb0, byte 16rb1, byte 16rb2, byte 16rb3, byte 16rb4, byte 16rb5, byte 16rb6, byte 16rb7, 
		byte 16rb8, byte 16rb9, byte 16rba, byte 16rbb, byte 16rbc, byte 16rbd, byte 16rbe, byte 16rbf, 
		byte 16rc0, byte 16rc1, byte 16rc2, byte 16rc3, byte 16rc4, byte 16rc5, byte 16rc6, byte 16rc7, 
		byte 16rc8, byte 16rc9, byte 16rca, byte 16rcb, byte 16rcc, byte 16rcd, byte 16rce, byte 16rcf, 
		byte 16rd0, byte 16rd1, byte 16rd2, byte 16rd3, byte 16rd4, byte 16rd5, byte 16rd6, byte 16rd7, 
		byte 16rd8, byte 16rd9, byte 16rda, byte 16rdb, byte 16rdc, byte 16rdd, byte 16rde, byte 16rdf, 
		byte 16re0, byte 16re1, byte 16re2, byte 16re3, byte 16re4, byte 16re5, byte 16re6, byte 16re7, 
		byte 16re8, byte 16re9, byte 16rea, byte 16reb, byte 16rec, byte 16red, byte 16ree, byte 16ref, 
		byte 16rf0, byte 16rf1, byte 16rf2, byte 16rf3, byte 16rf4, byte 16rf5, byte 16rf6, byte 16rf7, 
		byte 16rf8, byte 16rf9, byte 16rfa, byte 16rfb, byte 16rfc, byte 16rfd, byte 16rfe, byte 16rff, 
	},
	array[] of {
		byte 16rbf, byte 16r53, byte 16rd2, byte 16rda, byte 16rde, byte 16r78, byte 16re8, byte 16r22, 
		byte 16ra4, byte 16rd9, byte 16r49, byte 16ra9, byte 16rbc, byte 16r67, byte 16r66, byte 16rb0, 
		byte 16r1b, byte 16r06, byte 16ra8, byte 16ref, byte 16r70, byte 16rd2, byte 16r67, byte 16r48, 
		byte 16rc6, byte 16ra7, byte 16rfc, byte 16r36, byte 16rd8, byte 16r0a, byte 16re4, byte 16rc5, 
		byte 16r52, byte 16r0f, byte 16r7c, byte 16r4a, byte 16rb0, byte 16rac, byte 16r85, byte 16r44, 
		byte 16r42, byte 16r4f, byte 16ra4, byte 16r05, byte 16r16, byte 16r2f, byte 16ref, byte 16r5a, 
		byte 16r6b, byte 16r7f, byte 16r22, byte 16r94, byte 16r98, byte 16r06, byte 16r36, byte 16r18, 
		byte 16rd3, byte 16r9f, byte 16r00, byte 16r03, byte 16rcb, byte 16r5f, byte 16rb8, byte 16rd1, 
		byte 16rc8, byte 16r6b, byte 16r64, byte 16r34, byte 16r97, byte 16rda, byte 16r1f, byte 16rf9, 
		byte 16r45, byte 16rc8, byte 16rd3, byte 16rbe, byte 16rde, byte 16rca, byte 16r4f, byte 16r47, 
		byte 16r97, byte 16r02, byte 16ra7, byte 16ra7, byte 16r35, byte 16rf0, byte 16r43, byte 16rdd, 
		byte 16rb1, byte 16rd6, byte 16raa, byte 16rad, byte 16re3, byte 16rc4, byte 16ra0, byte 16rac, 
		byte 16r7c, byte 16ra7, byte 16rf3, byte 16rfa, byte 16r52, byte 16r79, byte 16rbe, byte 16rf5, 
		byte 16r6f, byte 16r82, byte 16rcd, byte 16r7a, byte 16r2f, byte 16r38, byte 16r67, byte 16r2e, 
		byte 16r82, byte 16r48, byte 16r14, byte 16re1, byte 16r07, byte 16r00, byte 16r30, byte 16r0a, 
		byte 16r05, byte 16r5e, byte 16r16, byte 16r30, byte 16rb8, byte 16rf1, byte 16rcb, byte 16r0e, 
		byte 16r91, byte 16r9f, byte 16r5e, byte 16r94, byte 16r20, byte 16r10, byte 16ra4, byte 16r16, 
		byte 16re2, byte 16rbf, byte 16r48, byte 16rcb, byte 16r46, byte 16r99, byte 16r3d, byte 16r3c, 
		byte 16rb6, byte 16ra5, byte 16r1c, byte 16r19, byte 16rba, byte 16rcf, byte 16r86, byte 16r47, 
		byte 16r85, byte 16ra0, byte 16r0b, byte 16rc2, byte 16rec, byte 16rff, byte 16r15, byte 16rd3, 
		byte 16r50, byte 16r87, byte 16r5b, byte 16r24, byte 16r6e, byte 16rd5, byte 16r3e, byte 16r68, 
		byte 16rbe, byte 16r6f, byte 16r55, byte 16rbd, byte 16r7e, byte 16r05, byte 16rcf, byte 16rc2, 
		byte 16rb2, byte 16red, byte 16r64, byte 16r32, byte 16r19, byte 16r8a, byte 16r64, byte 16r44, 
		byte 16rb6, byte 16rd8, byte 16rc2, byte 16r47, byte 16rfa, byte 16rb9, byte 16r41, byte 16rf5, 
		byte 16r69, byte 16r76, byte 16r8b, byte 16r5c, byte 16r42, byte 16r93, byte 16r66, byte 16rf1, 
		byte 16rd3, byte 16rf0, byte 16r0f, byte 16r03, byte 16r45, byte 16rb9, byte 16r61, byte 16r23, 
		byte 16rd5, byte 16r62, byte 16r04, byte 16rc0, byte 16r1c, byte 16r63, byte 16rb2, byte 16r2c, 
		byte 16re7, byte 16r8b, byte 16raf, byte 16r11, byte 16r6e, byte 16r52, byte 16r5e, byte 16rd9, 
		byte 16r0f, byte 16rde, byte 16ra3, byte 16r9f, byte 16ra4, byte 16r69, byte 16r49, byte 16r4d, 
		byte 16r38, byte 16r66, byte 16rc3, byte 16r1e, byte 16r05, byte 16rf2, byte 16r95, byte 16rff, 
		byte 16r21, byte 16rfe, byte 16ra8, byte 16rd4, byte 16re6, byte 16re1, byte 16r3d, byte 16r67, 
		byte 16re4, byte 16r7c, byte 16re7, byte 16r22, byte 16re9, byte 16r69, byte 16r8a, byte 16r1c, 
		byte 16r10, byte 16r48, byte 16rd6, byte 16r8e, byte 16rbc, byte 16rde, byte 16r76, byte 16rb8, 
		byte 16r6f, byte 16rcf, byte 16r97, byte 16r6e, byte 16rab, byte 16r8a, byte 16ra9, byte 16r79, 
		byte 16r02, byte 16r68, byte 16rb7, byte 16r06, byte 16r8e, byte 16r01, byte 16r7a, byte 16r8b, 
		byte 16r9b, byte 16r74, byte 16r94, byte 16r09, byte 16r51, byte 16r4f, byte 16r10, byte 16r53, 
		byte 16r02, byte 16r7f, byte 16rd1, byte 16r6c, byte 16r37, byte 16r86, byte 16rea, byte 16r1b, 
		byte 16rac, byte 16r5f, byte 16r15, byte 16rcb, byte 16r79, byte 16r71, byte 16r1e, byte 16re2, 
		byte 16rab, byte 16re8, byte 16r2f, byte 16r5c, byte 16rf8, byte 16rb1, byte 16r3a, byte 16re7, 
		byte 16r30, byte 16r30, byte 16ref, byte 16r5b, byte 16r9e, byte 16r44, byte 16r57, byte 16re7, 
		byte 16r5d, byte 16r13, byte 16r04, byte 16rf9, byte 16r88, byte 16rd6, byte 16r2d, byte 16rd6, 
		byte 16rfc, byte 16r4b, byte 16r94, byte 16red, byte 16r38, byte 16rba, byte 16r83, byte 16r1d, 
		byte 16ra4, byte 16rb7, byte 16r63, byte 16r49, byte 16r71, byte 16rb6, byte 16rcd, byte 16r8e, 
		byte 16rc3, byte 16r25, byte 16rd9, byte 16rc6, byte 16r1c, byte 16r00, byte 16rf1, byte 16rdf, 
		byte 16r73, byte 16r62, byte 16r7e, byte 16rd3, byte 16r74, byte 16r5a, byte 16r5e, byte 16r84, 
		byte 16r89, byte 16rf3, byte 16ra9, byte 16r5c, byte 16r69, byte 16r63, byte 16r9c, byte 16r32, 
		byte 16rcd, byte 16r6e, byte 16r1d, byte 16r53, byte 16r7a, byte 16r85, byte 16rf7, byte 16r5c, 
		byte 16rc8, byte 16r44, byte 16r72, byte 16r6e, byte 16r8a, byte 16r72, byte 16rfc, byte 16r00, 
		byte 16r77, byte 16rad, byte 16r22, byte 16r00, byte 16r0f, byte 16r1d, byte 16r50, byte 16r78, 
		byte 16rf6, byte 16rb8, byte 16r66, byte 16r31, byte 16r8c, byte 16r66, byte 16r8f, byte 16r1a, 
		byte 16rd0, byte 16r3d, byte 16r5a, byte 16r5f, byte 16rce, byte 16rd5, byte 16r21, byte 16r9f, 
		byte 16r2e, byte 16rab, byte 16rbd, byte 16r0a, byte 16ra5, byte 16rc0, byte 16rf4, byte 16r60, 
		byte 16rd1, byte 16r83, byte 16rf0, byte 16r44, byte 16r04, byte 16ra0, byte 16rd6, byte 16rf4, 
		byte 16r69, byte 16r55, byte 16r8e, byte 16r81, byte 16rfa, byte 16rb2, byte 16r4a, byte 16r16, 
		byte 16r79, byte 16r05, byte 16rab, byte 16r4c, byte 16r78, byte 16r78, byte 16r50, byte 16r2a, 
		byte 16rd3, byte 16re3, byte 16r8f, byte 16rdb, byte 16re6, byte 16r2a, byte 16r41, byte 16r55, 
		byte 16r6c, byte 16rec, byte 16r37, byte 16r32, byte 16r57, byte 16r59, byte 16r53, byte 16r3c, 
		byte 16re8, byte 16rf2, byte 16r5f, byte 16r36, byte 16r7c, byte 16r87, byte 16rbb, byte 16r55, 
		byte 16r78, byte 16rd6, byte 16r67, byte 16rae, byte 16r93, byte 16rf9, byte 16re2, byte 16rfd, 
		byte 16r99, byte 16rbc, byte 16rbc, byte 16r5f, byte 16r2f, byte 16rbb, byte 16ra8, byte 16r8c, 
		byte 16rf6, byte 16r51, byte 16r61, byte 16r39, byte 16r42, byte 16r0f, byte 16rcf, byte 16rf3, 
		byte 16rb7, byte 16r36, byte 16r1d, byte 16r86, byte 16r32, byte 16r2c, byte 16r4b, byte 16rd8, 
		byte 16r4c, byte 16r82, byte 16rf3, byte 16r35, byte 16rab, byte 16rb1, byte 16r52, byte 16rc4, 
		byte 16ra9, byte 16r34, byte 16r11, byte 16r37, byte 16r3a, byte 16raa, byte 16r82, byte 16r20, 
	}
),
Xts (
	big 16rffffffffff,
	array[] of {
		byte 16r27, byte 16r18, byte 16r28, byte 16r18, byte 16r28, byte 16r45, byte 16r90, byte 16r45, 
		byte 16r23, byte 16r53, byte 16r60, byte 16r28, byte 16r74, byte 16r71, byte 16r35, byte 16r26, 
		byte 16r62, byte 16r49, byte 16r77, byte 16r57, byte 16r24, byte 16r70, byte 16r93, byte 16r69, 
		byte 16r99, byte 16r59, byte 16r57, byte 16r49, byte 16r66, byte 16r96, byte 16r76, byte 16r27, 
		byte 16r31, byte 16r41, byte 16r59, byte 16r26, byte 16r53, byte 16r58, byte 16r97, byte 16r93, 
		byte 16r23, byte 16r84, byte 16r62, byte 16r64, byte 16r33, byte 16r83, byte 16r27, byte 16r95, 
		byte 16r02, byte 16r88, byte 16r41, byte 16r97, byte 16r16, byte 16r93, byte 16r99, byte 16r37, 
		byte 16r51, byte 16r05, byte 16r82, byte 16r09, byte 16r74, byte 16r94, byte 16r45, byte 16r92, 
	},
	array[] of {
		byte 16r00, byte 16r01, byte 16r02, byte 16r03, byte 16r04, byte 16r05, byte 16r06, byte 16r07, 
		byte 16r08, byte 16r09, byte 16r0a, byte 16r0b, byte 16r0c, byte 16r0d, byte 16r0e, byte 16r0f, 
		byte 16r10, byte 16r11, byte 16r12, byte 16r13, byte 16r14, byte 16r15, byte 16r16, byte 16r17, 
		byte 16r18, byte 16r19, byte 16r1a, byte 16r1b, byte 16r1c, byte 16r1d, byte 16r1e, byte 16r1f, 
		byte 16r20, byte 16r21, byte 16r22, byte 16r23, byte 16r24, byte 16r25, byte 16r26, byte 16r27, 
		byte 16r28, byte 16r29, byte 16r2a, byte 16r2b, byte 16r2c, byte 16r2d, byte 16r2e, byte 16r2f, 
		byte 16r30, byte 16r31, byte 16r32, byte 16r33, byte 16r34, byte 16r35, byte 16r36, byte 16r37, 
		byte 16r38, byte 16r39, byte 16r3a, byte 16r3b, byte 16r3c, byte 16r3d, byte 16r3e, byte 16r3f, 
		byte 16r40, byte 16r41, byte 16r42, byte 16r43, byte 16r44, byte 16r45, byte 16r46, byte 16r47, 
		byte 16r48, byte 16r49, byte 16r4a, byte 16r4b, byte 16r4c, byte 16r4d, byte 16r4e, byte 16r4f, 
		byte 16r50, byte 16r51, byte 16r52, byte 16r53, byte 16r54, byte 16r55, byte 16r56, byte 16r57, 
		byte 16r58, byte 16r59, byte 16r5a, byte 16r5b, byte 16r5c, byte 16r5d, byte 16r5e, byte 16r5f, 
		byte 16r60, byte 16r61, byte 16r62, byte 16r63, byte 16r64, byte 16r65, byte 16r66, byte 16r67, 
		byte 16r68, byte 16r69, byte 16r6a, byte 16r6b, byte 16r6c, byte 16r6d, byte 16r6e, byte 16r6f, 
		byte 16r70, byte 16r71, byte 16r72, byte 16r73, byte 16r74, byte 16r75, byte 16r76, byte 16r77, 
		byte 16r78, byte 16r79, byte 16r7a, byte 16r7b, byte 16r7c, byte 16r7d, byte 16r7e, byte 16r7f, 
		byte 16r80, byte 16r81, byte 16r82, byte 16r83, byte 16r84, byte 16r85, byte 16r86, byte 16r87, 
		byte 16r88, byte 16r89, byte 16r8a, byte 16r8b, byte 16r8c, byte 16r8d, byte 16r8e, byte 16r8f, 
		byte 16r90, byte 16r91, byte 16r92, byte 16r93, byte 16r94, byte 16r95, byte 16r96, byte 16r97, 
		byte 16r98, byte 16r99, byte 16r9a, byte 16r9b, byte 16r9c, byte 16r9d, byte 16r9e, byte 16r9f, 
		byte 16ra0, byte 16ra1, byte 16ra2, byte 16ra3, byte 16ra4, byte 16ra5, byte 16ra6, byte 16ra7, 
		byte 16ra8, byte 16ra9, byte 16raa, byte 16rab, byte 16rac, byte 16rad, byte 16rae, byte 16raf, 
		byte 16rb0, byte 16rb1, byte 16rb2, byte 16rb3, byte 16rb4, byte 16rb5, byte 16rb6, byte 16rb7, 
		byte 16rb8, byte 16rb9, byte 16rba, byte 16rbb, byte 16rbc, byte 16rbd, byte 16rbe, byte 16rbf, 
		byte 16rc0, byte 16rc1, byte 16rc2, byte 16rc3, byte 16rc4, byte 16rc5, byte 16rc6, byte 16rc7, 
		byte 16rc8, byte 16rc9, byte 16rca, byte 16rcb, byte 16rcc, byte 16rcd, byte 16rce, byte 16rcf, 
		byte 16rd0, byte 16rd1, byte 16rd2, byte 16rd3, byte 16rd4, byte 16rd5, byte 16rd6, byte 16rd7, 
		byte 16rd8, byte 16rd9, byte 16rda, byte 16rdb, byte 16rdc, byte 16rdd, byte 16rde, byte 16rdf, 
		byte 16re0, byte 16re1, byte 16re2, byte 16re3, byte 16re4, byte 16re5, byte 16re6, byte 16re7, 
		byte 16re8, byte 16re9, byte 16rea, byte 16reb, byte 16rec, byte 16red, byte 16ree, byte 16ref, 
		byte 16rf0, byte 16rf1, byte 16rf2, byte 16rf3, byte 16rf4, byte 16rf5, byte 16rf6, byte 16rf7, 
		byte 16rf8, byte 16rf9, byte 16rfa, byte 16rfb, byte 16rfc, byte 16rfd, byte 16rfe, byte 16rff, 
		byte 16r00, byte 16r01, byte 16r02, byte 16r03, byte 16r04, byte 16r05, byte 16r06, byte 16r07, 
		byte 16r08, byte 16r09, byte 16r0a, byte 16r0b, byte 16r0c, byte 16r0d, byte 16r0e, byte 16r0f, 
		byte 16r10, byte 16r11, byte 16r12, byte 16r13, byte 16r14, byte 16r15, byte 16r16, byte 16r17, 
		byte 16r18, byte 16r19, byte 16r1a, byte 16r1b, byte 16r1c, byte 16r1d, byte 16r1e, byte 16r1f, 
		byte 16r20, byte 16r21, byte 16r22, byte 16r23, byte 16r24, byte 16r25, byte 16r26, byte 16r27, 
		byte 16r28, byte 16r29, byte 16r2a, byte 16r2b, byte 16r2c, byte 16r2d, byte 16r2e, byte 16r2f, 
		byte 16r30, byte 16r31, byte 16r32, byte 16r33, byte 16r34, byte 16r35, byte 16r36, byte 16r37, 
		byte 16r38, byte 16r39, byte 16r3a, byte 16r3b, byte 16r3c, byte 16r3d, byte 16r3e, byte 16r3f, 
		byte 16r40, byte 16r41, byte 16r42, byte 16r43, byte 16r44, byte 16r45, byte 16r46, byte 16r47, 
		byte 16r48, byte 16r49, byte 16r4a, byte 16r4b, byte 16r4c, byte 16r4d, byte 16r4e, byte 16r4f, 
		byte 16r50, byte 16r51, byte 16r52, byte 16r53, byte 16r54, byte 16r55, byte 16r56, byte 16r57, 
		byte 16r58, byte 16r59, byte 16r5a, byte 16r5b, byte 16r5c, byte 16r5d, byte 16r5e, byte 16r5f, 
		byte 16r60, byte 16r61, byte 16r62, byte 16r63, byte 16r64, byte 16r65, byte 16r66, byte 16r67, 
		byte 16r68, byte 16r69, byte 16r6a, byte 16r6b, byte 16r6c, byte 16r6d, byte 16r6e, byte 16r6f, 
		byte 16r70, byte 16r71, byte 16r72, byte 16r73, byte 16r74, byte 16r75, byte 16r76, byte 16r77, 
		byte 16r78, byte 16r79, byte 16r7a, byte 16r7b, byte 16r7c, byte 16r7d, byte 16r7e, byte 16r7f, 
		byte 16r80, byte 16r81, byte 16r82, byte 16r83, byte 16r84, byte 16r85, byte 16r86, byte 16r87, 
		byte 16r88, byte 16r89, byte 16r8a, byte 16r8b, byte 16r8c, byte 16r8d, byte 16r8e, byte 16r8f, 
		byte 16r90, byte 16r91, byte 16r92, byte 16r93, byte 16r94, byte 16r95, byte 16r96, byte 16r97, 
		byte 16r98, byte 16r99, byte 16r9a, byte 16r9b, byte 16r9c, byte 16r9d, byte 16r9e, byte 16r9f, 
		byte 16ra0, byte 16ra1, byte 16ra2, byte 16ra3, byte 16ra4, byte 16ra5, byte 16ra6, byte 16ra7, 
		byte 16ra8, byte 16ra9, byte 16raa, byte 16rab, byte 16rac, byte 16rad, byte 16rae, byte 16raf, 
		byte 16rb0, byte 16rb1, byte 16rb2, byte 16rb3, byte 16rb4, byte 16rb5, byte 16rb6, byte 16rb7, 
		byte 16rb8, byte 16rb9, byte 16rba, byte 16rbb, byte 16rbc, byte 16rbd, byte 16rbe, byte 16rbf, 
		byte 16rc0, byte 16rc1, byte 16rc2, byte 16rc3, byte 16rc4, byte 16rc5, byte 16rc6, byte 16rc7, 
		byte 16rc8, byte 16rc9, byte 16rca, byte 16rcb, byte 16rcc, byte 16rcd, byte 16rce, byte 16rcf, 
		byte 16rd0, byte 16rd1, byte 16rd2, byte 16rd3, byte 16rd4, byte 16rd5, byte 16rd6, byte 16rd7, 
		byte 16rd8, byte 16rd9, byte 16rda, byte 16rdb, byte 16rdc, byte 16rdd, byte 16rde, byte 16rdf, 
		byte 16re0, byte 16re1, byte 16re2, byte 16re3, byte 16re4, byte 16re5, byte 16re6, byte 16re7, 
		byte 16re8, byte 16re9, byte 16rea, byte 16reb, byte 16rec, byte 16red, byte 16ree, byte 16ref, 
		byte 16rf0, byte 16rf1, byte 16rf2, byte 16rf3, byte 16rf4, byte 16rf5, byte 16rf6, byte 16rf7, 
		byte 16rf8, byte 16rf9, byte 16rfa, byte 16rfb, byte 16rfc, byte 16rfd, byte 16rfe, byte 16rff, 
	},
	array[] of {
		byte 16r64, byte 16r49, byte 16r7e, byte 16r5a, byte 16r83, byte 16r1e, byte 16r4a, byte 16r93, 
		byte 16r2c, byte 16r09, byte 16rbe, byte 16r3e, byte 16r53, byte 16r93, byte 16r37, byte 16r6d, 
		byte 16raa, byte 16r59, byte 16r95, byte 16r48, byte 16rb8, byte 16r16, byte 16r03, byte 16r1d, 
		byte 16r22, byte 16r4b, byte 16rbf, byte 16r50, byte 16ra8, byte 16r18, byte 16red, byte 16r23, 
		byte 16r50, byte 16rea, byte 16re7, byte 16re9, byte 16r60, byte 16r87, byte 16rc8, byte 16ra0, 
		byte 16rdb, byte 16r51, byte 16rad, byte 16r29, byte 16r0b, byte 16rd0, byte 16r0c, byte 16r1a, 
		byte 16rc1, byte 16r62, byte 16r08, byte 16r57, byte 16r63, byte 16r5b, byte 16rf2, byte 16r46, 
		byte 16rc1, byte 16r76, byte 16rab, byte 16r46, byte 16r3b, byte 16re3, byte 16r0b, byte 16r80, 
		byte 16r8d, byte 16ra5, byte 16r48, byte 16r08, byte 16r1a, byte 16rc8, byte 16r47, byte 16rb1, 
		byte 16r58, byte 16re1, byte 16r26, byte 16r4b, byte 16re2, byte 16r5b, byte 16rb0, byte 16r91, 
		byte 16r0b, byte 16rbc, byte 16r92, byte 16r64, byte 16r71, byte 16r08, byte 16r08, byte 16r94, 
		byte 16r15, byte 16rd4, byte 16r5f, byte 16rab, byte 16r1b, byte 16r3d, byte 16r26, byte 16r04, 
		byte 16re8, byte 16ra8, byte 16ref, byte 16rf1, byte 16rae, byte 16r40, byte 16r20, byte 16rcf, 
		byte 16ra3, byte 16r99, byte 16r36, byte 16rb6, byte 16r68, byte 16r27, byte 16rb2, byte 16r3f, 
		byte 16r37, byte 16r1b, byte 16r92, byte 16r20, byte 16r0b, byte 16re9, byte 16r02, byte 16r51, 
		byte 16re6, byte 16rd7, byte 16r3c, byte 16r5f, byte 16r86, byte 16rde, byte 16r5f, byte 16rd4, 
		byte 16ra9, byte 16r50, byte 16r78, byte 16r19, byte 16r33, byte 16rd7, byte 16r9a, byte 16r28, 
		byte 16r27, byte 16r2b, byte 16r78, byte 16r2a, byte 16r2e, byte 16rc3, byte 16r13, byte 16ref, 
		byte 16rdf, byte 16rcc, byte 16r06, byte 16r28, byte 16rf4, byte 16r3d, byte 16r74, byte 16r4c, 
		byte 16r2d, byte 16rc2, byte 16rff, byte 16r3d, byte 16rcb, byte 16r66, byte 16r99, byte 16r9b, 
		byte 16r50, byte 16rc7, byte 16rca, byte 16r89, byte 16r5b, byte 16r0c, byte 16r64, byte 16r79, 
		byte 16r1e, byte 16rea, byte 16ra5, byte 16rf2, byte 16r94, byte 16r99, byte 16rfb, byte 16r1c, 
		byte 16r02, byte 16r6f, byte 16r84, byte 16rce, byte 16r5b, byte 16r5c, byte 16r72, byte 16rba, 
		byte 16r10, byte 16r83, byte 16rcd, byte 16rdb, byte 16r5c, byte 16re4, byte 16r54, byte 16r34, 
		byte 16r63, byte 16r16, byte 16r65, byte 16rc3, byte 16r33, byte 16rb6, byte 16r0b, byte 16r11, 
		byte 16r59, byte 16r3f, byte 16rb2, byte 16r53, byte 16rc5, byte 16r17, byte 16r9a, byte 16r2c, 
		byte 16r8d, byte 16rb8, byte 16r13, byte 16r78, byte 16r2a, byte 16r00, byte 16r48, byte 16r56, 
		byte 16ra1, byte 16r65, byte 16r30, byte 16r11, byte 16re9, byte 16r3f, byte 16rb6, byte 16rd8, 
		byte 16r76, byte 16rc1, byte 16r83, byte 16r66, byte 16rdd, byte 16r86, byte 16r83, byte 16rf5, 
		byte 16r34, byte 16r12, byte 16rc0, byte 16rc1, byte 16r80, byte 16rf9, byte 16rc8, byte 16r48, 
		byte 16r59, byte 16r2d, byte 16r59, byte 16r3f, byte 16r86, byte 16r09, byte 16rca, byte 16r73, 
		byte 16r63, byte 16r17, byte 16rd3, byte 16r56, byte 16re1, byte 16r3e, byte 16r2b, byte 16rff, 
		byte 16r3a, byte 16r9f, byte 16r59, byte 16rcd, byte 16r9a, byte 16reb, byte 16r19, byte 16rcd, 
		byte 16r48, byte 16r25, byte 16r93, byte 16rd8, byte 16rc4, byte 16r61, byte 16r28, byte 16rbb, 
		byte 16r32, byte 16r42, byte 16r3b, byte 16r37, byte 16ra9, byte 16rad, byte 16rfb, byte 16r48, 
		byte 16r2b, byte 16r99, byte 16r45, byte 16r3f, byte 16rbe, byte 16r25, byte 16ra4, byte 16r1b, 
		byte 16rf6, byte 16rfe, byte 16rb4, byte 16raa, byte 16r0b, byte 16ref, byte 16r5e, byte 16rd2, 
		byte 16r4b, byte 16rf7, byte 16r3c, byte 16r76, byte 16r29, byte 16r78, byte 16r02, byte 16r54, 
		byte 16r82, byte 16rc1, byte 16r31, byte 16r15, byte 16re4, byte 16r01, byte 16r5a, byte 16rac, 
		byte 16r99, byte 16r2e, byte 16r56, byte 16r13, byte 16ra3, byte 16rb5, byte 16rc2, byte 16rf6, 
		byte 16r85, byte 16rb8, byte 16r47, byte 16r95, byte 16rcb, byte 16r6e, byte 16r9b, byte 16r26, 
		byte 16r56, byte 16rd8, byte 16rc8, byte 16r81, byte 16r57, byte 16re5, byte 16r2c, byte 16r42, 
		byte 16rf9, byte 16r78, byte 16rd8, byte 16r63, byte 16r4c, byte 16r43, byte 16rd0, byte 16r6f, 
		byte 16rea, byte 16r92, byte 16r8f, byte 16r28, byte 16r22, byte 16re4, byte 16r65, byte 16raa, 
		byte 16r65, byte 16r76, byte 16re9, byte 16rbf, byte 16r41, byte 16r93, byte 16r84, byte 16r50, 
		byte 16r6c, byte 16rc3, byte 16rce, byte 16r3c, byte 16r54, byte 16rac, byte 16r1a, byte 16r6f, 
		byte 16r67, byte 16rdc, byte 16r66, byte 16rf3, byte 16rb3, byte 16r01, byte 16r91, byte 16re6, 
		byte 16r98, byte 16r38, byte 16r0b, byte 16rc9, byte 16r99, byte 16rb0, byte 16r5a, byte 16rbc, 
		byte 16re1, byte 16r9d, byte 16rc0, byte 16rc6, byte 16rdc, byte 16rc2, byte 16rdd, byte 16r00, 
		byte 16r1e, byte 16rc5, byte 16r35, byte 16rba, byte 16r18, byte 16rde, byte 16rb2, byte 16rdf, 
		byte 16r1a, byte 16r10, byte 16r10, byte 16r23, byte 16r10, byte 16r83, byte 16r18, byte 16rc7, 
		byte 16r5d, byte 16rc9, byte 16r86, byte 16r11, byte 16ra0, byte 16r9d, byte 16rc4, byte 16r8a, 
		byte 16r0a, byte 16rcd, byte 16rec, byte 16r67, byte 16r6f, byte 16rab, byte 16rdf, byte 16r22, 
		byte 16r2f, byte 16r07, byte 16re0, byte 16r26, byte 16rf0, byte 16r59, byte 16rb6, byte 16r72, 
		byte 16rb5, byte 16r6e, byte 16r5c, byte 16rbc, byte 16r8e, byte 16r1d, byte 16r21, byte 16rbb, 
		byte 16rd8, byte 16r67, byte 16rdd, byte 16r92, byte 16r72, byte 16r12, byte 16r05, byte 16r46, 
		byte 16r81, byte 16rd7, byte 16r0e, byte 16ra7, byte 16r37, byte 16r13, byte 16r4c, byte 16rdf, 
		byte 16rce, byte 16r93, byte 16rb6, byte 16rf8, byte 16r2a, byte 16re2, byte 16r24, byte 16r23, 
		byte 16r27, byte 16r4e, byte 16r58, byte 16ra0, byte 16r82, byte 16r1c, byte 16rc5, byte 16r50, 
		byte 16r2e, byte 16r2d, byte 16r0a, byte 16rb4, byte 16r58, byte 16r5e, byte 16r94, byte 16rde, 
		byte 16r69, byte 16r75, byte 16rbe, byte 16r5e, byte 16r0b, byte 16r4e, byte 16rfc, byte 16re5, 
		byte 16r1c, byte 16rd3, byte 16re7, byte 16r0c, byte 16r25, byte 16ra1, byte 16rfb, byte 16rbb, 
		byte 16rd6, byte 16r09, byte 16rd2, byte 16r73, byte 16rad, byte 16r5b, byte 16r0d, byte 16r59, 
		byte 16r63, byte 16r1c, byte 16r53, byte 16r1f, byte 16r6a, byte 16r0a, byte 16r57, byte 16rb9, 
	}
),
};
